/* -*- c++ -*- */
/****************************************************************
 *                                                              *
 *  Project name:      Biped Locomotion                         *
 *                                                              *
 *  Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)           *
 ****************************************************************
 * $Id: biped_controller.cpp,v 1.5 2004/08/15 20:04:40 s0343981 Exp $ 
 ****************************************************************/
/**
 * \file Controller.cpp 
 * \brief class Controller methods' implementations.
 */

// Arabica (SAX) includes:
#include <SAX/InputSource.h>
#include <SAX/helpers/FeatureNames.h>
#include <SAX/XMLReader.h>
// CGAL includes:
#include <CGAL/convex_hull_2.h>
#include <CGAL/Convex_hull_projective_xz_traits_2.h>
#include <CGAL/convex_hull_constructive_traits_2.h>
#include <CGAL/Polygon_2.h>
#include <CGAL/Polygon_2_algorithms.h>
#include <CGAL/squared_distance_2.h>
// Webots includes:
#include <device/robot.h>
#include <device/servo.h>
#include <device/emitter.h>
#include <device/camera.h>
#include <device/distance_sensor.h>
#include <device/gps.h>
#include <device/touch_sensor.h>
// ROBOOP includes:
#include <newmat/newmat.h>
#include <newmat/newmatio.h>
// C++ includes:
#include <cstdlib>
#include <iostream>
#include <sstream>
#include <iomanip>
#include <string>
#include <cmath>
#include <limits>
// BIPED includes:
#include <biped_exception.hpp>
#include <biped_logging.hpp>
#include "biped_controller.hpp"

using std::cerr;
using std::cout;
using std::endl;
using std::string;

// -----------------------------------------------------------------------------
//  Function: Controller::Controller
/// \brief Controller's c'tor.
/// \param jointsNum No. of joints existing for this robot.
/// \param lt Handle to the left touch sensor device.
/// \param rt Handle to the right touch sensor device.
/// \param range Handle to the range sensor device.
/// \param gps Handle to the GPS sensor device.
/// \param joints Array of handles to all servo joint devices.
/// \param camera Handle to the camera device.
/// \param emitter Handle to the emitter device.
// -----------------------------------------------------------------------------
Controller::Controller(uint jointsNum, int lt, int rt, int range, int gps, 
					   DeviceTag * joints, DeviceTag camera, DeviceTag emitter) :
	jointsNum_(jointsNum) , leftTouch_(lt) , rightTouch_(rt) , range_(range) , 
	gps_(gps) , servos_(joints) , camera_(camera) , emitter_(emitter) , 
	emitterBuffer_(NULL) , motionSequencesMap_() 
{

}

// -----------------------------------------------------------------------------
//  Function: Controller::enableDevices
/// \brief Enables all the devices available to the robot with same time-step.
// -----------------------------------------------------------------------------
void Controller::enableDevices()
{
	TRY(Controller::enableDevices);
	cout << "Controller: Enabling devices ... " << endl;
	camera_enable(camera_, TIME_STEP);
	distance_sensor_enable(range_, TIME_STEP);
	touch_sensor_enable(leftTouch_, TIME_STEP);
	touch_sensor_enable(rightTouch_, TIME_STEP);
	gps_enable(gps_, TIME_STEP);

	// Enable feedback on servos:
	for (uint i = 0; i < jointsNum_; i++) {
		servo_enable_position(servos_[i], TIME_STEP);
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Controller::getCOM
/// \brief Computes the center-of-mass of the robot. (see Kagami et al., 2003).
/// \param m ColumnVector (1xN) containing masses of all links.
/// \param P 3xN matrix containing absolute positions of all links.
/// \return 3D point containing COM.
// -----------------------------------------------------------------------------
Point_3 Controller::getCOM(const ColumnVector & m, const Matrix & P) const
{
	TRY(Controller::getCOM);
	dReal totalMass = m.Sum();
	dReal x = DotProduct(m, P.Row(1)) / totalMass;
	dReal y = DotProduct(m, P.Row(2)) / totalMass;
	dReal z = DotProduct(m, P.Row(3)) / totalMass;
	return Point_3(x, y, z);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Controller::getZMP
/// \brief Computes the Zero-Moment Point of the robot.
/// Details: See details in Sugihara, PhD thesis, 2004.
/// \param m RowVector (Nx1) containing masses of all links.
/// \param R 3xN matrix containing absolute positions of all links.
/// \param A 3xN matrix containing absolute accelerations of all links.
/// \param y scalar of the height of the ground.
/// \param I 9xN matrix containing inertia components of all links.
/// \param W 3xN matrix containing angular velocities of all links.
/// \return 3D point containing ZMP.
/// \note The result will have 'y' as the 2nd element. 
// -----------------------------------------------------------------------------
Point_3 Controller::getZMP(const ColumnVector & m, const Matrix & R, 
						   const Matrix & A, const dReal & y, const Matrix & I,
						   const Matrix & W) const
{
	TRY(Controller::getZMP);
	const dReal g[3] = { 0, -9.81, 0};
	
	// Calculate total accelerations:
	RowVector acc_x = A.Row(1) - g[0];
	RowVector acc_y = A.Row(2) - g[1];
	RowVector acc_z = A.Row(3) - g[2];

	// Calculate the denominator:
	dReal denom = DotProduct(m, acc_y);
	
	// The offset between Y components and given y of the ZMP:
	RowVector offset_y = R.Row(2) - y; 

	// Calculate the x dimension of the ZMP:
	RowVector I_z = I.Row(9);	// Izz elements
	dReal M_z = DotProduct(I_z, W.Row(3)); // total moment in Z-axis
	const RowVector & R_x = R.Row(1); // X position components
	dReal x = DotProduct(m, (SP(R_x, acc_y) - SP(acc_x, offset_y))) - M_z;
	x /= denom;

	// Calculate the z dimension of the ZMP:
	RowVector I_x = I.Row(1);	// Ixx elements
	dReal M_x = DotProduct(I_x, W.Row(1)); // total moment in X-axis
	const RowVector & R_z = R.Row(3); // Z position components
	dReal z = DotProduct(m, (SP(R_z, acc_y) - SP(acc_z, offset_y))) - M_x;
	z /= denom;

	return Point_3(x, y, z);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Controller::getSupportPolygon
/// \brief Calculates the support polygon coordinates.
/// \param contacts the input set of contact points with the ground.
/// \param polygon the resulting support polygon vector of 3D points.
// -----------------------------------------------------------------------------
void Controller::getSupportPolygon(const Point3Vec & contacts, 
								   SupportPolygon & polygon) const
{
	TRY(Controller::getSupportPolygon);
	typedef CGAL::Convex_hull_projective_xz_traits_2< Point_3 > ConvexProjectXZ;
 	polygon.clear();
	CGAL::convex_hull_points_2(contacts.begin(), contacts.end(), 
							   back_inserter(polygon), ConvexProjectXZ());
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Controller::getGroundHeight
/// \brief Calculates the average ground height given the support polygon.
/// \param polygon the support polygon vector of 3D points.
/// \return the ground's height
// -----------------------------------------------------------------------------
dReal Controller::getGroundHeight(const SupportPolygon & polygon) const
{
	TRY(Controller::getGroundHeight);
	dReal heightSum = 0;
	uint polygonSize = polygon.size();
	BIPED_ASSERT(polygonSize > 0);
	for (uint i = 0; i < polygonSize; i++)
		heightSum += polygon[i].y();
	return (heightSum / polygonSize);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Controller::getBoundedSide
/// \brief Computes if a point lies inside a polygon.
/// \param polygon A vector of points representing the support polygon.
/// \param pt A 3D query point.
/// \return The floating point stability margin.
// -----------------------------------------------------------------------------
CGAL::Bounded_side Controller::getBoundedSide(const SupportPolygon & polygon,
											  const Point_3 & pt) const
{
	TRY(Controller::getBoundedSide);
	typedef CGAL::Polygon_2< CGAL::Cartesian< dReal > > Polygon_2;

	// Create projected objects to X-Z plane:
	Point_2 pt2D(pt.x(), pt.z());
	Polygon_2 polygon2D;
	for (uint i = 0; i < polygon.size(); i++) 
		polygon2D.push_back(Point_2(polygon[i].x(), polygon[i].z()));

	return CGAL::bounded_side_2(polygon2D.vertices_begin(), 
								polygon2D.vertices_end(), pt2D, 
								CGAL::Cartesian< dReal >());
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Controller::getStabilityMargin
/// \brief Calculates the minimum distance between the ZMP and the boundary of
/// the stable region (i.e. the support polygon).
/// Details: See Huang et al., (2001), "Planning Walking Patterns for a Biped 
/// Robot".
/// \param polygon A vector of points representing the support polygon.
/// \param zmp A 3D point representing the ZMP.
/// \param closestBoundaryPt A 3D point representing the closest support-polygon
/// boundary point to the ZMP.
/// \return The floating point stability margin.
// -----------------------------------------------------------------------------
dReal Controller::getStabilityMargin(const SupportPolygon & polygon, 
									 const Point_3 & zmp, 
									 Point_3 & closestBoundaryPt) const
{
	TRY(Controller::getStabilityMargin);
	typedef CGAL::Cartesian< dReal > Kernel; 
	typedef CGAL::Polygon_2< Kernel > Polygon_2;
	typedef Polygon_2::Edge_const_iterator EdgeConstIter;

	SupportPolygon::size_type polygonSize = polygon.size();
	BIPED_ASSERT(polygonSize > 0);

	// Project the ZMP to X-Z plane:
	Point_2 zmp2D(zmp.x(), zmp.z());
	if (polygonSize < 2) {	// There is no polygon - only a point
		closestBoundaryPt = polygon[0];
		Point_2 supportPt2D(polygon[0].x(), polygon[0].z());
		return CGAL::squared_distance(zmp2D, supportPt2D);
	}

	// Create a projected support-polygon in X-Z plane:
	Polygon_2 polygon2D;
	for (uint i = 0; i < polygonSize; i++) 
		polygon2D.push_back(Point_2(polygon[i].x(), polygon[i].z()));

	// Find the minimum distance between the 2D projected ZMP and 
	// all edges of the 2D projected support polygon:
	dReal minDist = std::numeric_limits< dReal >::max();
	for (EdgeConstIter eit = polygon2D.edges_begin(); 
		 eit != polygon2D.edges_end(); eit++) {
		dReal dist = CGAL::squared_distance(zmp2D, (*eit));
		int index;
		CGAL::CGALi::distance_index(index, zmp2D, (*eit), Kernel());
		if(dist < minDist) {	// found more minimal
            minDist = dist;
			Point_2 pt = (index == -1) ? 
				eit->supporting_line().projection(zmp2D) : eit->vertex(index);
			BIPED_ASSERT(eit->collinear_has_on(pt));
			closestBoundaryPt = Point_3(pt.x(), zmp.y(), pt.y());
		}
	}
	return std::sqrt(minDist);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Controller::getSupportPhase
/// \brief Returns the current support phase of the robot.
/// \return NO_SUPPORT, LEFT_SINGLE_SUPPORT, RIGHT_SINGLE_SUPPORT, 
/// or DOUBLE_SUPPORT.
// -----------------------------------------------------------------------------
Controller::SupportPhase Controller::getSupportPhase() const

{
	TRY(Controller::getSupportPhase);
	// FIXME: Change the sensitivity of the touch sensor (lookup table)
	unsigned short leftTouchVal = touch_sensor_get_value(leftTouch_);
	unsigned short rightTouchVal = touch_sensor_get_value(rightTouch_);
	if (leftTouchVal != 0 && rightTouchVal != 0)
		return DOUBLE_SUPPORT;
	if (leftTouchVal == 0)
		return RIGHT_SINGLE_SUPPORT;
	if (rightTouchVal == 0)
		return LEFT_SINGLE_SUPPORT;
	return NO_SUPPORT;	
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Controller::getBaseFrameTrans
/// \brief Computes the base link's (robot's torso) position and orientation.
/// \return A 4x4 homogenuous coordinates matrix representing the transform.
// -----------------------------------------------------------------------------
ReturnMatrix Controller::getBaseFrameTrans() const
{
	TRY(Controller::getBaseFrameTrans);
	Matrix res(4, 4); 
	res = dReal(0.0);
	const float * mat = gps_get_matrix(gps_);
	dReal initArray[16] = { 
		mat[0], mat[1], mat[2], mat[3], mat[4], mat[5], mat[6], mat[7], mat[8], 
		mat[9], mat[10], mat[11], mat[12], mat[13], mat[14], mat[15] };
	res << initArray;
	res.Release();
	return res;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Controller::setJointAngles
/// \brief Sets all joints angles collectively using a vector of joint angles.
/// \param q The vector of joint angles.
// -----------------------------------------------------------------------------
void Controller::setJointAngles(const ColumnVector & q)
{
	TRY(Controller::setJointAngles);
	for (uint j = 0; j < jointsNum_; j++) {
		servo_set_position(servos_[j], q(j + 1));
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Controller::getJointAngles
/// \brief Collects all joint angles from the robot.
/// \return a vector of current joint angles.
// -----------------------------------------------------------------------------
ColumnVector Controller::getJointAngles() const
{
	TRY(Controller::getJointAngles);
	ColumnVector q(jointsNum_);
	for (uint j = 0; j < jointsNum_; j++) {
		q(j + 1) = servo_get_position(servos_[j]);
	}
	return q;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Controller::emitterSend
/// \brief Sends a given message with specified size through the emitter device. 
/// \param msg 
/// \param msgSize 
/// \note Assumes the msg is already ordered correctly by the user for the 
/// receiver's expectation.
// -----------------------------------------------------------------------------
void Controller::emitterSend(const float * msg, uint msgSize)
{
	TRY(Controller::emitterSend);
	emitterBuffer_ = (float *) emitter_get_buffer(emitter_);
	memcpy(emitterBuffer_, msg, msgSize);
	emitter_send(emitter_, msgSize);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Controller::readMotionXML
/// \brief Read the motion sequences XML file and initialize internal sequences
/// data structure.
// -----------------------------------------------------------------------------
void Controller::readMotionXML(const string & filename)
{
	TRY(Controller::readMotionXML);
	cout << "Controller: Reading motion sequences XML file" << endl;
	XMLHandler myHandler(motionSequencesMap_);
	SAX::FeatureNames< std::string > fNames;
	SAX::XMLReader< std::string > parser;
	
	try {
		parser.setFeature(fNames.external_general, true);
		parser.setFeature(fNames.validation, true);
		parser.setFeature(fNames.namespaces, true);
		parser.setFeature(fNames.namespace_prefixes, true);
	} catch(SAX::SAXException & e) { 
		cerr << e.what() << endl;
	}

	parser.setContentHandler(myHandler);
	parser.setDTDHandler(myHandler); 
	parser.setErrorHandler(myHandler);
	parser.setEntityResolver(myHandler);

	SAX::InputSource inputSrc(filename);
	parser.parse(inputSrc);

	CATCH;
}
