/* -*- c++ -*- */
/****************************************************************
 *                                                              *
 *  Project name:      Biped Locomotion                         *
 *                                                              *
 *  Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)           *
 ****************************************************************
 * $Id: biped_controller.hpp,v 1.5 2004/08/15 20:04:40 s0343981 Exp $ 
 ****************************************************************/
/**
 * \file Controller.hpp 
 * \brief class Controller definition.
 */

#ifndef __CONTROLLER_HPP__
#define __CONTROLLER_HPP__

#include <CGAL/Cartesian.h>
#include <source/robot.h>
#include <ode/common.h>
#include <biped_link.hpp>
#include <biped_xml_handler.hpp>

#define TIME_STEP 32			// millisec
#define TIME_STEP_SEC (0.001 * TIME_STEP) // in sec

class RowVector;
class ColumnVector;
class Matrix;
class Trajectory;

typedef CGAL::Point_2< CGAL::Cartesian< dReal > > Point_2;
typedef CGAL::Point_3< CGAL::Cartesian< dReal > > Point_3;
typedef vector< Point_3 > Point3Vec;
typedef Point3Vec SupportPolygon;

// =============================================================================
//  Class: Controller                                                               
/// \brief C++ definition of a robot controller in Webots.
// =============================================================================
class Controller {
public:							// definitions
	typedef enum { NO_SUPPORT, LEFT_SINGLE_SUPPORT, RIGHT_SINGLE_SUPPORT, 
				   DOUBLE_SUPPORT } SupportPhase;

public:							// methods
	Controller(uint jointsNum, int leftTouch, int rightTouch, int range, 
			   int gps, DeviceTag * joints, DeviceTag camera, 
			   DeviceTag emitter);
	virtual ~Controller() { }
	
	virtual void enableDevices();
	virtual void run() = 0;
	virtual void initialStep() = 0;
	
protected:						// methods
	// Balance related utility methods:
	virtual Point_3 getCOM(const ColumnVector & m, const Matrix & P) const;
	virtual Point_3 getZMP(const ColumnVector & m, const Matrix & R, 
						   const Matrix & A, const dReal & y, const Matrix & I,
						   const Matrix & W) const;
	virtual void getSupportPolygon(const Point3Vec & contacts, 
								   SupportPolygon & polygon) const;
	virtual dReal getGroundHeight(const SupportPolygon & polygon) const;
	virtual CGAL::Bounded_side getBoundedSide(const SupportPolygon & polygon,
											  const Point_3 & pt) const;
	virtual dReal getStabilityMargin(const SupportPolygon & polygon, 
									 const Point_3 & zmp,
									 Point_3 & closestBoundPt) const;
	virtual SupportPhase getSupportPhase() const;
	virtual ReturnMatrix getBaseFrameTrans() const;
	
	virtual void setJointAngles(const ColumnVector & q);
	virtual ColumnVector getJointAngles() const;
	virtual void emitterSend(const float * msg, uint msgSize);
	virtual void readMotionXML(const string & filename);
	
protected:						// members
	uint jointsNum_;

	// Various devices:
	int leftTouch_;
	int rightTouch_;
	int range_;
	int gps_;
	DeviceTag * servos_;
	DeviceTag camera_;
	DeviceTag emitter_;
	float * emitterBuffer_;

	MotionSequencesMap motionSequencesMap_;
};
		
#endif	// __CONTROLLER_HPP__
