/* -*- C++ -*- */
/****************************************************************
 *                                                              *
 *   Project name:        Biped Locomotion                      *
 *                                                              *
 *   Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)          *
 ****************************************************************
 * $Id: biped_exception.cpp,v 1.5 2004/08/06 18:02:42 s0343981 Exp $
 ****************************************************************/
/**
 * \file: biped_exception.cpp
 * \brief Exception handling auxilary functions implementations.
 */

#include "biped_exception.hpp"

// -----------------------------------------------------------------------------
//  Function: printFileInfo
/// \brief Prints the function, file, line number to output stream.
/// \param  pFun  the function's name which raised the exception.
/// \param  pFile the file name.
/// \param  line  the line number.
/// \param  out   the output stream.
// -----------------------------------------------------------------------------
void printFileInfo(const char * pFun, const char * pFile, int line, 
				   ostream & out)
{
	out << " " << pFun << "() in " << pFile << " L:" << line << endl;
}

// -----------------------------------------------------------------------------
//  Function: printUndefException
/// \brief Prints the information about an undefined exception to the standard 
/// error. 
/// \param  pFun  the function's name which raised the exception.
/// \param  pFile the file name.
/// \param  line  the line number.
// -----------------------------------------------------------------------------
void printUndefException(const char * pFun, const char * pFile, int line)
{
	cerr << "*** Undefined Exception !!! ";
	printFileInfo(pFun, pFile, line , cerr);
}

// -----------------------------------------------------------------------------
//  Function: BIPED_Exception::beginReport
/// \brief Prints the initial type of the error report. 
// -----------------------------------------------------------------------------
void BIPED_Exception::beginReport()
{
	cerr << "*** FATAL ERROR: " ;
}

// -----------------------------------------------------------------------------
//  Function: BIPED_Exception::beginFailureReport
/// \brief Prints the initial indication of a failure in the report.
// -----------------------------------------------------------------------------
void BIPED_Exception::beginFailureReport()
{
	cerr << "*** FAILURE: ";
}

// -----------------------------------------------------------------------------
//  Function: BIPED_Exception::endReport
/// \brief Prints the end of the error report.
/// \param pFun the function in which the error occured.
/// \param pFile the file which hosts this function.
/// \param line the line in the file where the error occured.
// -----------------------------------------------------------------------------
void BIPED_Exception::endReport(const char * pFun, const char * pFile, int line)
{
	printFI(pFun, pFile, line, 0);
}

// -----------------------------------------------------------------------------
//  Function: BIPED_Exception::processing
/// \brief Process the error report including incrementing the count of nesting. 
/// \param pFun the function in which the error occured.
/// \param pFile the file which hosts this function.
/// \param line the line in the file where the error occured.
// -----------------------------------------------------------------------------
void BIPED_Exception::processing(const char * pFun, const char * pFile, 
								int line)
{
	_count++;
	printFI(pFun, pFile, line, _count);
}

// -----------------------------------------------------------------------------
//  Function: BIPED_Exception::printFI
/// \brief Prints the failure indication.
/// \param pFun the function in which the error occured.
/// \param pFile the file which hosts this function.
/// \param line the line in the file where the error occured.
/// \param count the level of error nesting.
// -----------------------------------------------------------------------------
void BIPED_Exception::printFI(const char * pFun, const char * pFile, int line,
							 int count)
{
	cerr << "### " << count;
	printFileInfo(pFun, pFile, line, cerr);
}
