/* -*- C++ -*- */
/****************************************************************
 *                                                              *
 *   Project name:        Biped Locomotion                      *
 *                                                              *
 *   Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)          *
 ****************************************************************
 * $Id: biped_exception.hpp,v 1.5 2004/08/06 18:02:42 s0343981 Exp $
 ****************************************************************/
/**
 * \file: biped_exception.hpp
 * \brief Macros for exception handling and class Exception definition.
 */

#ifndef __BIPED_EXCEPTION_HPP__
#define __BIPED_EXCEPTION_HPP__

#include <iostream>
#include <string>

using namespace std;

void printFileInfo(const char * pFun, const char * pFile, int line, 
				   ostream & out);
void printUndefException(const char * pFun, const char * pFile, int line);

// =============================================================================
//  Class: BIPED_Exception                                                           
/// \brief A general exception handler class for hoap-2.
// =============================================================================
class BIPED_Exception {
public:
    typedef enum { BIPED_ERROR , BIPED_FATAL } ExceptionType;

    BIPED_Exception(ExceptionType type): _type(type) , _count(0) { }

    static void beginReport();
    static void beginFailureReport();
    static void endReport(const char * pFun, const char * pFile, int line);

    virtual void processing(const char * pFun, const char * pFile, int line);
    int isFatalError(void) const 
	{ return ( _type == BIPED_FATAL ); }

protected:
    ExceptionType _type;		///< Fatal error or just error

private:
    unsigned _count;			///< the line no. count

    static void printFI(const char * pFun, const char * pFile, int line, 
                        int count);
};

#define  BEGIN_THROW           BIPED_Exception::beginReport()
#define  BEGIN_FAILURE_THROW   BIPED_Exception::beginFailureReport()

#define  END_CUSTOM_THROW(what)										\
	BIPED_Exception::endReport(bipedFuncName, __FILE__, __LINE__);	\
	throw what

#define  END_THROW  END_CUSTOM_THROW(BIPED_Exception(BIPED_Exception::BIPED_FATAL))

// -----------------------------------------------------------------------------
//  Macro: THROW_ERROR
/// \brief Used for throwing a regular error (accepts a message string). 
// -----------------------------------------------------------------------------
#define  THROW_ERROR(pMsg)                      \
	do {										\
		BEGIN_FAILURE_THROW;					\
		cerr << pMsg << endl;					\
		END_THROW;								\
	} while (0)

// -----------------------------------------------------------------------------
//  Macro: THROW_FATAL
/// \brief Used for throwing a fatal error (accepts a message string). 
// -----------------------------------------------------------------------------
#define  THROW_FATAL(pMsg)                      \
	do {										\
		BEGIN_THROW;							\
		cerr << pMsg << endl;					\
		END_THROW;								\
	} while (0)

inline void bipedDeclareEmptyFunc(const char *) { } 

// -----------------------------------------------------------------------------
//  Macro: BIPED_DECLARE
/// \brief When placed at the begining of a function will output trace if an
/// error happens in that function.
// -----------------------------------------------------------------------------
#define  BIPED_DECLARE(n) const char * bipedFuncName = #n;	\
	bipedDeclareEmptyFunc(bipedFuncName) 

#define  TRY(n) BIPED_DECLARE(n); try {

#define  BEGIN_CUSTOM_CATCH(type)					\
	} catch (type & e) {							\
    e.processing(bipedFuncName, __FILE__, __LINE__)

#define  END_CUSTOM_CATCH

#define  BEGIN_DEFAULT_CATCH								\
	} catch(...) {											\
    printUndefException(bipedFuncName, __FILE__, __LINE__)
#define  END_DEFAULT_CATCH						\
	END_CUSTOM_CATCH							\
	}
    
#define  BEGIN_CATCH  BEGIN_CUSTOM_CATCH(BIPED_Exception)

#define  END_CATCH(what)                        \
	what										\
	END_CUSTOM_CATCH;							\
	BEGIN_DEFAULT_CATCH;						\
	what										\
	END_DEFAULT_CATCH;
  
#define  CATCH                                  \
	BEGIN_CATCH;								\
	END_CATCH(throw;);

// no throw
#define CATCH_FINAL								\
	BEGIN_CATCH;								\
	END_CATCH(;);														

// -----------------------------------------------------------------------------
//  Macro: BIPED_ASSERT
/// \brief Used for concept checking (like assert() in C).
// -----------------------------------------------------------------------------
#ifdef NDEBUG
#  define  BIPED_ASSERT(ignore)  ((void) 0)
#else
#  define  BIPED_ASSERT(expression)						\
	if (!(expression)) {								\
		THROW_FATAL("Assertion failed: " #expression);  \
	}													\
	else
#endif	// else of ifdef NDEBUG

#endif	// __BIPED_EXCEPTION_HPP__
