/* -*- C++ -*- */
/****************************************************************
 *                                                              *
 *   Project name:      Biped Locomotion                        *
 *                                                              *
 *   Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)          *
 ****************************************************************
 * $Id: biped_gnuplot.cpp,v 1.5 2004/08/16 21:37:28 s0343981 Exp $
 ****************************************************************/
/**
 * \file: biped_gnuplot.cpp
 * \brief class GNUPlot implementation.
 */

#include <cstdarg>
#include <iostream>
#include <fstream>
#include <sstream>
#include <list>
#include <biped_exception.hpp>
#include <biped_gnuplot.hpp>
#include <biped_logging.hpp>

using namespace std;

// -----------------------------------------------------------------------------
//  Function: GNUPlot::GNUPlot
/// \brief Default C'tor.
// -----------------------------------------------------------------------------
GNUPlot::GNUPlot(void) : 
	gnucmd_(NULL) , plotStyle_("points") , tmpFilenames_() , plotsNum_(0) , 
	isValid_(true) 
{
	TRY(GNUPlot::GNUPlot);
	init();
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: GNUPlot::GNUPlot
/// \brief C'tor using a given line style.
/// \param style A string defining the line style (e.g. "points", "lines", 
/// "boxes" etc. - see gnuplot documentation).
// -----------------------------------------------------------------------------
GNUPlot::GNUPlot(const string & style) : 
	gnucmd_(NULL) , plotStyle_("") , tmpFilenames_() , plotsNum_(0) , 
	isValid_(true)
{
	TRY(GNUPlot::GNUPlot);
	init();
	setStyle(style);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: GNUPlot::GNUPlot
/// \brief C'tor which initializes the object and plots.
/// Details: Equivalent of gnuplot_plot_once, the two forms allow you to plot 
/// either (x,y) pairs or just a single vector at one go.
/// \param title the title of the plot.
/// \param style the line style.
/// \param xLabel the label for X axis.
/// \param yLabel the label for Y axis.
/// \param x A vector of X-values.
/// \param y A vector of Y-values.
// -----------------------------------------------------------------------------
GNUPlot::GNUPlot(const string & title, const string & style, 
				 const string & xLabel, const string & yLabel, 
				 const vector< double > & x, const vector< double > & y) : 
	gnucmd_(NULL) , plotStyle_("points") , tmpFilenames_() , plotsNum_(0) , 
	isValid_(true)
{
	TRY(GNUPlot::GNUPlot);
	init();
		
    if (x.size() == 0 || y.size() == 0)
        THROW_ERROR("Vectors have zero size");

	setStyle(((style == "") ? "lines" : style));
	setXLabel(((xLabel == "") ? "X" : xLabel));
	setYLabel(((yLabel == "") ? "Y" : yLabel));
    plotXYData(x, y, title);

    cout << "Press enter to continue" << endl;
    while (getchar() != '\n') { }
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: GNUPlot::GNUPlot
/// \brief C'tor which initializes the object and plots.
/// Details: Equivalent of gnuplot_plot_once, allows to plot just a single vector.
/// \param title the title of the plot.
/// \param style the line style.
/// \param xLabel the label for X axis.
/// \param yLabel the label for Y axis.
/// \param x A vector of X-values.
// -----------------------------------------------------------------------------
GNUPlot::GNUPlot(const string & title, const string & style,
				 const string & xLabel, const string & yLabel,
				 const vector< double > & x) :
	gnucmd_(NULL) , plotStyle_("points") , tmpFilenames_() , plotsNum_(0) , 
	isValid_(true)
{
	TRY(GNUPlot::GNUPlot);
	init();
    if (x.size() == 0)
        THROW_ERROR("Vector has zero size");

	setStyle(((style == "") ? "lines" : style));
	setXLabel(((xLabel == "") ? "X" : xLabel));
	setYLabel(((yLabel == "") ? "Y" : yLabel));
    plotXData(x, title);

    cout << "Press enter to continue" << endl;
    while (getchar() != '\n') { }
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: GNUPlot::~GNUPlot
/// \brief Destructor. Releases the memory and closes the connection to gnuplot.
// -----------------------------------------------------------------------------
GNUPlot::~GNUPlot()
{
	TRY(GNUPlot::~GNUPlot);
	for (list< string >::const_iterator it = tmpFilenames_.begin(); 
		 it != tmpFilenames_.end(); it++)
		remove(it->c_str());	// removes file from filesystem

	if (pclose(gnucmd_) == -1)
		THROW_ERROR("Problem closing communication to gnuplot!");
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: GNUPlot::resetPlot
/// \brief Resets the graphical plot, i.e removes all plots present.
// -----------------------------------------------------------------------------
void GNUPlot::resetPlot(void)
{
	TRY(GNUPlot::resetPlot);
	for (list< string >::const_iterator it = tmpFilenames_.begin(); 
		 it != tmpFilenames_.end(); it++)
		remove(it->c_str()); // removes file from filesystem

    plotsNum_ = 0;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: GNUPlot::setStyle
/// \brief Sets the line style of the plot.
/// \param styleStr the style supported by gnuplot (see gnuplot documentation).
// -----------------------------------------------------------------------------
void GNUPlot::setStyle(const string & styleStr)
{
	TRY(GNUPlot::setStyle);
//     if ((styleStr.compare("lines") == 0) || (styleStr.compare("points") == 0) ||
// 		(styleStr.compare("dots") == 0) || (styleStr.compare("boxes") == 0) || 
// 		(styleStr.compare("impulses") == 0) || 
// 		(styleStr.compare("steps") == 0) ||
// 		(styleStr.compare("errorbars") == 0) || 
// 		(styleStr.compare("linespoints") == 0) || 
// 		(styleStr.compare("boxerrorbars") == 0)) {
	plotStyle_ = styleStr;
//     } else {
//         plotStyle_ = string("points");
// 	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: GNUPlot::cmd
/// \brief Sends a direct command to gnuplot.
/// \param cmdStr the command string (follow gnuplot syntax).
/// \param ... Allows additional parameters.
// -----------------------------------------------------------------------------
void GNUPlot::cmd(const char * cmdStr, ...)
{
	TRY(GNUPlot::cmd);
    va_list ap;
    char localCmd[GNUPLOT_CMD_SIZE];

    va_start(ap, cmdStr);
    vsprintf(localCmd, cmdStr, ap);
    va_end(ap);
    strcat(localCmd, "\n");
    fputs(localCmd, gnucmd_);
    fflush(gnucmd_);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: GNUPlot::setLabel
/// \brief Sets the label for a given label type (i.e. "xlabel" or "ylabel").
/// \param label the desired label for the axis.
/// \param labelTypeStr the desired axis (i.e. "xlabel" or "ylabel");
// -----------------------------------------------------------------------------
void GNUPlot::setLabel(const string & label, const string & labelTypeStr)
{
	TRY(GNUPlot::setLabel);
    ostringstream cmdStr;
    cmdStr << "set " << labelTypeStr << " \"" << label << "\"";
    cmd(cmdStr.str().c_str());
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: GNUPlot::plotXData
/// \brief Plots a single vector of numeric data.
/// \param d the data vector.
/// \param title the title to set for the plot.
/// \param prefix the persistent prefix to use.
// -----------------------------------------------------------------------------
void GNUPlot::plotXData(const vector< double > & d, const string & title,
						const string & prefix)
{
	TRY(GNUPlot::plotXData);
	vector< double > fakeY;
	plotVectorData(false, d, fakeY, title, prefix);
	CATCH;
}
    
// -----------------------------------------------------------------------------
//  Function: GNUPlot::plotXYData
/// \brief Plots X/Y pairs of data.
/// \param x the data values on X axis.
/// \param y the data values on Y axis.
/// \param title the desired title for the plot.
/// \param prefix the persistent prefix to use.
// -----------------------------------------------------------------------------
void GNUPlot::plotXYData(const vector< double > & x, const vector<double> & y,
						 const string & title, const string & prefix)
{
	TRY(GNUPlot::plotXYData);
	// Check same sizes:
    if (y.size() != x.size())
        THROW_FATAL("x and y are of different sizes");

	plotVectorData(true, x, y, title, prefix);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: GNUPlot::plotVectorData
/// \brief Plots either data vector or pairs.
/// \param isPairs treat as pairs of data values or just first data vector.
/// \param x the data values on X axis.
/// \param y the data values on Y axis.
/// \param title the desired title for the plot.
/// \param prefix the persistent prefix to use.
// -----------------------------------------------------------------------------
void GNUPlot::plotVectorData(bool isPairs, const vector< double > & x, 
							 const vector< double > & y, const string & title,
							 const string & prefix)
{
	TRY(GNUPlot::plotVectorData);
    if (tmpFilenames_.size() == GNUPLOT_MAX_TMP_FILES - 1)
        THROW_FATAL("Max. no. of temporary files reached (" 
					<< GNUPLOT_MAX_TMP_FILES << "): cannot open more files");

    // Open temporary files for output:
	ostringstream nameStream;
	nameStream << "/tmp/" << prefix << "_gnuplot.XXXXXX";
	char * name = strdup(nameStream.str().c_str());
	if (mkstemp(name) == -1)
		THROW_FATAL("Cannot create temporary file: exiting plot");
	ofstream tmp;
    tmp.open(name);
    if (tmp.bad()) 
		THROW_FATAL("Cannot create temorary file: exiting plot");
	
    // Save the temporary filename:
    tmpFilenames_.push_back(name);

    // Write the data to file:
    for (uint i = 0; i < x.size(); i++) {
		tmp << x[i];
		if (isPairs) 
			tmp << " " << y[i];
		tmp << endl;
	}
    tmp.flush();
    tmp.close();

	ostringstream cmdStr;
    // Command to be sent to gnuplot:
    cmdStr << ((plotsNum_ > 0) ? "replot " : "plot ") << "\"" << name;
	if (title != "")
		cmdStr << "\" title \"" << title;
	cmdStr << "\" with " << plotStyle_;
	
    // Do the actual plot:
    cmd(cmdStr.str().c_str());
    plotsNum_++;
	free(name);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: GNUPlot::plotSlope
/// \brief Plots a linear equation y = a * x + b.
/// \param a the slope parameter.
/// \param b the intercept parameter.
/// \param title the desired title for the plot.
// -----------------------------------------------------------------------------
void GNUPlot::plotSlope(double a, double b, const string & title)
{
	TRY(GNUPlot::plotSlope);
    ostringstream stitle, cmdStr;

    stitle << ((title == "") ? "no title" : title);
    cmdStr << ((plotsNum_ > 0) ? "replot " : "plot ") << a << " * x + " << b 
		   << " title \"" << stitle.str() << "\" with " << plotStyle_;
    cmd(cmdStr.str().c_str());
    plotsNum_++;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: GNUPlot::plotEquation
/// \brief Plots an equation which is supplied as a string (i.e. y = f(x)).
/// \param equation the string describing the f(x).
/// \param title the desired title for the plot.
// -----------------------------------------------------------------------------
void GNUPlot::plotEquation(const string & equation, const string & title)
{
	TRY(GNUPlot::plotEquation);
    string titleStr, plotStr;
    ostringstream cmdStr;

    titleStr = ((title == "") ? "no title" : title);
    cmdStr << ((plotsNum_ > 0) ? "replot " : "plot ") << equation << " " 
		   << "title \"" << titleStr << "\" with " << plotStyle_;
    cmd(cmdStr.str().c_str());
    plotsNum_++;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: GNUPlot::init
/// \brief Special initialization of the GNUPlot object. Makes sure DISPLAY is 
/// set and that gnuplot is executable.
// -----------------------------------------------------------------------------
void GNUPlot::init()
{
	TRY(GNUPlot::init);
	if (getenv("DISPLAY") == NULL) {
        isValid_ = false;
        THROW_ERROR("cannot find DISPLAY environment variable");
	}
    if (!getProgramPath("gnuplot")) {
        isValid_ = false;
        THROW_ERROR("Can't find gnuplot in your PATH");
	}
    gnucmd_ = popen("gnuplot", "w");
    if (!gnucmd_) {
        isValid_ = false;
        THROW_ERROR("Could'nt open connection to gnuplot");
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: stringtok
/// \brief A template utility function to tokenize a string.
/// \param container The STL container which the tokens will be stored.
/// \param in the input string to tokenize.
/// \param delimiters A C string containing all desired delimiters for tokenizer.
// -----------------------------------------------------------------------------
template < typename Container >
void stringtok(Container & container, string const & in,
			   const char * const delimiters = " \t\n")
{
    const string::size_type len = in.length();
	typename string::size_type i = 0;

    while (i < len) {
        // Ignore leading whitespace:
        i = in.find_first_not_of(delimiters, i);
        if (i == string::npos)  // nothing left but white space
            return; 

        // Find the end of the token:
        typename string::size_type j = in.find_first_of(delimiters, i);

        // Push token
        if (j == string::npos) {
            container.push_back(in.substr(i));
            return;
        } else
            container.push_back(in.substr(i, j - i));

        // Set up for next loop:
        i = j + 1;
    }
}

// -----------------------------------------------------------------------------
//  Function: GNUPlot::getProgramPath
/// \brief Checks if the given path is accessible (i.e. if the program can be
/// run using the current defined PATH).
/// \param pname The string containing the path.
/// \return TRUE if succeeded, FALSE otherwise.
// -----------------------------------------------------------------------------
bool GNUPlot::getProgramPath(const string & pname)
{
	TRY(GNUPlot::getProgramPath);
    list< string > ls;
    char * path = getenv("PATH");
    if (!path)
        THROW_FATAL("Path is not set");
    else {
        stringtok(ls, path, ":");
        for (list< string >::const_iterator i = ls.begin(); i != ls.end(); ++i) {
            string tmp = (*i) + "/" + pname;
			if (access(tmp.c_str(), X_OK) == 0)
				return true;
		}
	}
    return false;
	CATCH;
}
