/* -*- C++ -*- */
/****************************************************************
 *                                                              *
 *  Project name:      Biped Locomotion                         *
 *                                                              *
 *  Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)           *
 ****************************************************************
 * $Id: biped_interp.cpp,v 1.4 2004/08/15 20:04:41 s0343981 Exp $
 ****************************************************************/
/**
 * \file: biped_interp.cpp
 * \brief 1D Interpolation classes implementations.
 */

#include <cmath>
#include <biped_exception.hpp>
#include <biped_interp.hpp>

#define sqr(x) ((x) * (x))

// -----------------------------------------------------------------------------
//  Function: SplineBase::findInterval
/// \brief Find index of largest element in the increasingly ordered vector x, 
/// which is smaller than u. If u is smaller than the smallest value in the 
/// vector then the lowest index minus one is returned. 
/// Details: A fast binary search is performed. Note, that the vector must be
/// strictly increasing.
/// \param u control parameter.
/// \param x the abscissa vector.
/// \return the appropriate index.
// -----------------------------------------------------------------------------
int SplineBase::findInterval(double u, const RowVector & x) const
{
	TRY(SplineBase::findInterval);
	int i, j, lo = 1, hi = x.Ncols();
	if (u < x(lo)) {
		i = lo - 1;				// returns index below smallest index
	} else if (u >= x(hi)) {
		i = hi;					// returns highest index
	} else {
		i = lo;
		j = hi;
		do {
			int k = (i + j) / 2;
			if (u < x(k))  j = k;
			if (u >= x(k)) i = k;
		} while (j > i + 1);
	}
	return i;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: SplineBase::CubicSplineHorner
/// \brief Returns the interpolation value P(u) for a piecewise cubic curve 
/// determined by the abscissa vector, the ordinate vector, the 1st derivative
/// vector, the 2nd derivative vector, and the 3rd derivative vector, using the 
/// Horner scheme. 
/// Details: All vectors must have conformant dimenions. The abscissa x(i) values
/// must be strictly increasing. In the special case of empty data vectors (x,y)
/// a value of 0.0 is returned. This subroutine evaluates the function:
///    P(u) = y(i) + dx * (y1(i) + dx * (y2(i) + dx * y3(i)))
// where  x(i) <= u < x(i+1) and dx = u - x(i), using Horner's rule:
//    lo <= i <= hi is the index range of the vectors.
//    if  u <  x(lo) then  i = lo  is used.
//    if  u >= x(hi) then  i = hi  is used.
/// \param u the abscissa at which the interpolation is to be evaluated.
/// \param x the vectors (lo,hi) of data abscissas.
/// \param y the vectors (lo,hi) of data ordinates.
/// \param y1 vector (lo,hi) of coefficients which contains the 1st derivatives.
/// \param y2 vector (lo,hi) of coefficients which contains the 2nd derivatives.
/// \param y3 vector (lo,hi) of coefficients which contains the 3rd derivatives.
/// \return the interpolation value at given input.
// -----------------------------------------------------------------------------
double SplineBase::CubicSplineHorner(
	double u, const RowVector & x, const RowVector & y, const RowVector & y1, 
	const RowVector & y2, const RowVector & y3) const
{
	TRY(SplineBase::CubicSplineHorner);
	// Special case that there is no data:
	if (x.Ncols() == 0) return 0.0;

	int i = findInterval(u, x);  
	if (i < 1) i = 1;			// extrapolate to the left
	if (i == x.Ncols()) i--;	// extrapolate to the right
	double dx = u - x(i);
	return (y(i) + dx * (y1(i) + dx * (y2(i) + dx * y3(i))));
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: SplineBase::CubicSplineCoefficients
/// \brief Calculates the spline coefficients y2(i) and y3(i) for a natural 
/// cubic spline, given the abscissa, the ordinate, and the 1st derivative.
/// Details: The spline interpolation can then be evaluated using Horner's rule:
///      P(u) = y(i) + dx * (y1(i) + dx * (y2(i) + dx * y3(i)))
/// where  x(i) <= u < x(i+1) and dx = u - x(i).
/// \param x the vectors (lo,hi) of data abscissas.
/// \param y the vectors (lo,hi) of data ordinates.
/// \param y1 vector (lo,hi) of coefficients which contains the 1st derivatives.
/// \param y2 output vector (lo,hi) of coefficients containing 2nd derivatives.
/// \param y3 output vector (lo,hi) of coefficients containing 3rd derivatives.
// -----------------------------------------------------------------------------
void SplineBase::CubicSplineCoefficients(
	const RowVector & x, const RowVector & y, const RowVector & y1, 
	RowVector & y2, RowVector & y3) const
{
	TRY(SplineBase::CubicSplineCoefficients);
	int lo = 1, hi = x.Ncols();
	
	for (int i = lo; i < hi; i++) {
		double h  = x(i + 1) - x(i);
		double mi = (y(i + 1) - y(i)) / h;
		y2(i) = (3 * mi - 2 * y1(i) - y1(i + 1)) / h;
		y3(i) = (y1(i) + y1(i + 1) - 2 * mi) / (h * h);
	}
	y2(hi) = y3(hi) = 0.0;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: SplineBase::parameterize
/// \brief Computes the spline's length parameterization and returns accumulated
/// "distances" between the points (x(i),y(i)) and (x(i+1),y(i+1)) in t(i+1) 
/// t(lo) = 0.0 always.
/// Details: The way of parametrization is given by caller: 
///   |dx| + |dy|       for Norm1
///   sqrt(dx^2+dy^2)   for Norm2
///   (dx^2+dy^2)       for SqrNorm2
/// Parametrization using Norm2 usually gives the best results.
/// \param x the vectors (lo,hi) of data abscissas.
/// \param y the vectors (lo,hi) of data ordinates.
/// \param t the accumulated distances vector.
/// \param paramType the type of parameterization.
// -----------------------------------------------------------------------------
void SplineBase::parameterize(const RowVector & x, const RowVector & y, 
							  RowVector & t, int paramType) const
{
	TRY(SplineBase::parameterize);
	int lo = 1, hi = x.Ncols();
	t(lo) = 0.0;
	switch (paramType) {
		case Norm1:  
			for (int i = lo + 1; i <= hi; i++)
				t(i) = t(i - 1) + fabs(x(i) - x(i - 1)) + fabs(y(i) - y(i - 1));
			break;
		case Norm2:  
			for (int i = lo + 1; i <= hi; i++)
				t(i) = t(i - 1) + hypot(x(i) - x(i - 1), y(i) - y(i - 1));
			break;
		case SqrNorm2:  
			for (int i = lo + 1; i <= hi; i++)
				t(i) = t(i - 1) + sqr(x(i) - x(i - 1)) + sqr(y(i) - y(i - 1));
			break;
		default: 
			THROW_FATAL("SplineBase::parametrize: Illegal parameterization type");
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: SplineBase::createNewSpline
/// \brief Create a new spline of the given type.
/// \param type The specific spline type to create.
/// \return A pointer to the newly constructed spline. 
// -----------------------------------------------------------------------------
SplineBase * SplineBase::createNewSpline(SplineType type)
{
	TRY(SplineBase::createNewSpline);
	switch (type) {
		case AKIMA:      return new AkimaCubicSpline();
		case BEZIER:     return new BezierCubicSpline();
		case CARDINAL:   return new CardinalCubicSpline();
		case FRITSCH:    return new FritschCarlsonCubicSpline();
		case POLYNOMIAL: return new PolynomialSpline();
		default:
			THROW_FATAL("Unknown spline type (" << type << ")!"); 
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: AkimaCubicSpline::interpolate
/// \brief Calculates Akima cubic spline interpolation given abscissa and 
/// ordinate vectors. All vectors must have conformant dimenions and the abscissa
/// vector must be strictly increasing.
/// \param x the vectors (lo,hi) of data abscissas.
/// \param y the vectors (lo,hi) of data ordinates.
// -----------------------------------------------------------------------------
int AkimaCubicSpline::interpolate(const RowVector & x, const RowVector & y)
{
	TRY(AkimaCubicSpline::interpolate);
	// Check input parameters:
	if (x.Ncols() != y.Ncols())
		THROW_FATAL("AkimaCubicSpline::interpolate: vectors have different size");

	// Copy original data vectors into base class:
	x_ = x;
	y_ = y;

	// Empty data vectors - make auxilliary storage empty:
	if (x_.Ncols() == 0) {
		y1_ = RowVector();
		y2_ = RowVector();
		y3_ = RowVector();
		return 0;				// ok
	}
     
	int lo = 1, lo1 = lo + 1, lo2 = lo + 2;
	int hi = x_.Ncols(), hi1 = hi - 1, hi2 = hi - 2, hi3 = hi - 3;
	
	// ReSize the auxilliary vectors:
	y1_.ReSize(x_);
	y2_.ReSize(x_);
	y3_.ReSize(x_);
	
	if (x_.Ncols() == 1) {
		// Default derivatives are zero:
		y1_(lo) = y2_(lo) = y3_(lo) = 0.0;
	} else if (x_.Ncols() == 2) {
		// set derivatives for a line
		y1_(lo) = y1_(hi) = (y_(hi) - y_(lo)) / (x_(hi) - x_(lo));
		y2_(lo) = y2_(hi) = y3_(lo) = y3_(hi) = 0.0;
	} else {					// three or more - do Akima interpolation
		double num, den, m_m1, m_m2, m_p1, m_p2, x_m1, x_m2, x_p1, x_p2,
			y_m1, y_m2, y_p1, y_p2;
		// Short form to save some typing:
#   define m(i) ((y_(i + 1) - y_(i)) / (x_(i + 1) - x_(i)))
		
        // Interpolate the missing points:
		x_m1 = x_(lo) + x_(lo1) - x_(lo2); 
		y_m1 = (x_(lo) - x_m1) * (m(lo1) - 2 * m(lo)) + y_(lo);
		m_m1 = (y_(lo) - y_m1) / (x(lo) - x_m1);
		
		x_m2 = 2 * x_(lo) - x_(lo2);
		y_m2 = (x_m1 - x_m2) * (m(lo) - 2 * m_m1) + y_m1;
		m_m2 = (y_m1 - y_m2) / (x_m1 - x_m2);
    
		x_p1 = x_(hi) + x_(hi1) - x_(hi2);
		y_p1 = (2 * m(hi1) - m(hi2)) * (x_p1 - x_(hi)) + y_(hi);
		m_p1 = (y_p1 - y_(hi)) / (x_p1 - x_(hi));
		
		x_p2 = 2 * x_(hi) - x_(hi2);
		y_p2 = (2 * m_p1 - m(hi1)) * (x_p2 - x_p1) + y_p1;
		m_p2 = (y_p2 - y_p1) / (x_p2 - x_p1);
    
		// i = 0
		num = abs(m(lo1) - m(lo)) * m_m1 + abs(m_m1 - m_m2) * m(lo);
		den = abs(m(lo1) - m(lo)) + abs(m_m1 - m_m2);  	
		y1_(lo) = (den != 0.0) ? num / den : 0.0;
    
		// i = 1 
		if (x_.Ncols() > 3) {
			num = abs(m(lo2) - m(lo1)) * m(lo) + abs(m(lo) - m_m1) * m(lo1);
			den = abs(m(lo2) - m(lo1)) + abs(m(lo) - m_m1);
			y1_(lo1) = (den != 0.0) ? num / den :  0.0;
      
			for (int i = lo2; i < hi1; i++) {
				double mip1 = m(i + 1),	mi = m(i), mim1 = m(i - 1),	mim2 = m(i - 2);
				num = abs(mip1 - mi) * mim1 + abs(mim1 - mim2) * mi;
				den = abs(mip1 - mi) + abs(mim1 - mim2);
				y1_(i) = (den != 0.0) ? num / den : 0.0;
			}
			// i = n - 2
			num = abs(m_p1 - m(hi1)) * m(hi2) + abs(m(hi2) - m(hi3)) * m(hi1);
			den = abs(m_p1 - m(hi1)) + abs(m(hi2) - m(hi3));
			y1_(hi1) = (den != 0.0) ? num / den : 0.0;
		} else {				// exactly three elements
			num = m(lo);
			den = (m(lo1) - num) / (x_(lo2) - x_(lo));
			y1_(lo1) = num + den * (x_(lo1) - x_(lo));
		}
		
		// i = n - 1
		num = abs(m_p2 - m_p1) * m(hi1) + abs(m(hi1) - m(hi2)) * m_p1;
		den = abs(m_p2 - m_p1) + abs(m(hi1) - m(hi2));
		y1_(hi) = (den != 0.0) ? num / den : 0.0;
    
		// Calculate remaining spline coefficients y2_(i) and y3_(i):
		CubicSplineCoefficients(x_, y_, y1_, y2_, y3_);
	}
	return 0;					// ok
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: BezierCubicSpline::interpolate
/// \brief Calculates the Bezier cubic spline interpolation given abscissa and 
/// ordinates vectors. All vectors must have conformant dimensions.
/// \param x the vectors (lo,hi) of data abscissas.
/// \param y the vectors (lo,hi) of data ordinates.
// -----------------------------------------------------------------------------
int BezierCubicSpline::interpolate(const RowVector & x, const RowVector & y)
{
	TRY(BezierCubicSpline::interpolate);
	// Verify index range:
	if (x.Ncols() != y.Ncols()) 
		THROW_FATAL("BezierCubicSpline::interpolate: vectors have different size");
	// Copy original data vectors into base class:
	x_ = x;
	y_ = y;
	return 0;					// ok
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: BezierCubicSpline::getX
/// \brief Return the abscissa value given the control parameter.
/// \param t the control parameter.
/// \return the abscissa value.
// -----------------------------------------------------------------------------
double BezierCubicSpline::getX(double t) const
{
	TRY(BezierCubicSpline::getX);
	int lo = 1, hi = x_.Ncols(), i = findInterval(t, x_);  
	if (i < lo || i >= hi || hi - lo == 1) {
		// Linear extrapolation and interpolation for 2 points:
		return t;
	} else if (i == lo) {
		i = lo;
		double u  = (t - x_(i)) / (x_(i + 1) - x_(i));
		double u2 = u * u;
		double c1 = 1.0 + u * (u2 / 6.0 - 1.0), c2 = u * (1.0 - u2 / 3.0), 
			c3 = u * u2 / 6.0;
		return (c1 * x_(i) + c2 * x_(i + 1) + c3 * x_(i + 2));
	} else if (i == hi - 1) {
		i = hi - 1;
		double u  = (x_(i + 1) - t) / (x_(i + 1) - x_(i));
		double u2 = u * u;
		double c1 = 1.0 + u * (u2 / 6.0 - 1.0), c2 = u * (1.0 - u2 / 3.0), 
			c3 = u * u2 / 6.0;
		return (c1 * x_(i + 1) + c2 * x_(i) + c3 * x_(i - 1));
	} else {
		double u  = (t - x_(i)) / (x_(i+1) - x_(i));
		double u2 = u * u, u3 = 1.0 - u, u4 = u3 * u3;
		double c1 = u3 * u4 / 6.0, c2 = u2 * (u / 2.0 - 1.0) + 4.0 / 6.0,
			c3 = u * (1.0 + u * u3) / 2.0 + 1.0 / 6.0, c4 = u * u2 / 6.0;
		return (c1 * x_(i - 1) + c2 * x_(i) + c3 * x_(i + 1) + c4 * x_(i + 2));
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: BezierCubicSpline::getY
/// \brief Get the interpolation value (ordinate) at given control value.
/// \param t the control parameter.
/// \return the interpolation value.
// -----------------------------------------------------------------------------
double BezierCubicSpline::getY(double t) const
{
	TRY(BezierCubicSpline::getY);
	int lo = 1, hi = x_.Ncols(), i = findInterval(t, x_);  

	if (i < lo || hi - lo == 1) {
		// Linear extrapolation and interpolation for 2 points:
		return  (y_(lo) + (t - x_(lo)) * (y_(lo + 1) - y_(lo)) / 
				 (x_(lo + 1) - x_(lo)));
	} else if (i == lo) {
		i = lo;
		double u = (t - x_(i)) / (x_(i + 1) - x_(i));
		double u2 = u * u;
		double c1 = 1.0 + u * (u2 / 6.0 - 1.0), c2 = u * (1.0 - u2 / 3.0),
			c3 = u * u2 / 6.0;
		return (c1 * y_(i) + c2 * y_(i + 1) + c3 * y_(i + 2));
	} else if (i >= hi) {
		// Linear extrapolation:
		return (y_(hi) + (t - x_(hi)) * (y_(hi) - y_(hi - 1)) / 
				(x_(hi) - x_(hi - 1)));
	} else if (i == hi - 1) {
		i = hi - 1;
		double u  = (x_(i + 1) - t) / (x_(i + 1) - x_(i));
		double u2 = u * u;
		double c1 = 1.0 + u * (u2 / 6.0 - 1.0), c2 = u * (1.0 - u2 / 3.0),
			c3 = u * u2 / 6.0;
		return (c1 * y_(i + 1) + c2 * y_(i) + c3 * y_(i - 1));
	} else {
		double u  = (t - x_(i)) / (x_(i + 1) - x_(i));
		double u2 = u * u, u3 = 1.0 - u;
		double u4 = u3 * u3;
		double c1 = u3 * u4 / 6.0, c2 = u2 * (u / 2.0 - 1.0) + 4.0 / 6.0,
			c3 = u * (1.0 + u * u3) / 2.0 + 1.0 / 6.0, c4 = u * u2 / 6.0;
		return (c1 * y_(i - 1) + c2 * y_(i) + c3 * y_(i + 1) + 
				c4 * y_(i + 2));
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: CardinalCubicSpline::interpolate
/// \brief Calculates the Cardinal cubic spline interpolation given abscissa and 
/// ordinates vectors. All vectors must have conformant dimensions.
/// \param x the vectors (lo,hi) of data abscissas.
/// \param y the vectors (lo,hi) of data ordinates.
// -----------------------------------------------------------------------------
int CardinalCubicSpline::interpolate(const RowVector & x, const RowVector & y)
{
	TRY(CardinalCubicSpline::interpolate);
	// Verify index range:
	if (x.Ncols() != y.Ncols()) 
		THROW_FATAL("CardinalCubicSpline::interpolate: vectors have different size");
	
	// Copy original data vectors into base class:
	x_ = x;
	y_ = y;
	return 0;					// ok
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: CardinalCubicSpline::getX
/// \brief Return the abscissa value given the control parameter.
/// \param t the control parameter.
/// \return the abscissa value.
// -----------------------------------------------------------------------------
double CardinalCubicSpline::getX(double t) const
{
	TRY(CardinalCubicSpline::getX);
	int lo = 1, hi = x_.Ncols(), i = findInterval(t, x_);  
	if (i < lo || i >= hi || hi - lo == 1) {
		// Linear extrapolation and interpolation for 2 points:
		return t;
	} else if (i == lo) {
		i = lo;
		double u = (t - x_(i)) / (x_(i + 1) - x_(i));
		double u2 = u * u;
		double c1 = 1.0 + u * (-1.0 + u * (u - 1.0) / 2.0), 
			c2 = u * (1.0 + u * (1.0 - u)), c3 = u2 * (u - 1.0) / 2.0;
		return (c1 * x_(i) + c2 * x_(i + 1) + c3 * x_(i + 2));
	} else if (i == hi - 1) {
		i = hi - 1;
		double u = (x_(i + 1) - t) / (x_(i + 1) - x_(i));
		double u2 = u*u;
		double c1 = 1.0 + u * (-1.0 + u * (u - 1.0) / 2.0),
           c2 = u * (1.0 + u * (1.0 - u)), c3 = u2 * (u - 1.0) / 2.0;
		return (c1 * x_(i + 1) + c2 * x_(i) + c3 * x_(i - 1));
	} else {
		double u = (t - x_(i)) / (x_(i + 1) - x_(i));
		double u2 = u * u, u3 = 1.0 - u;
		double u4 = u3 * u3;
		double c1 = -u4 * u / 2.0, c2 = 1.0 + u2 * (3.0 * u - 5.0) / 2.0,
			c3 = u * (1.0 + u * (4.0 - 3.0 * u)) / 2.0, c4 = -u2 * u3 / 2.0;
		return (c1 * x_(i - 1) + c2 * x_(i) + c3 * x_(i + 1) + c4 * x_(i + 2));
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: CardinalCubicSpline::getY
/// \brief Get the interpolation value (ordinate) at given control value.
/// \param t the control parameter.
/// \return the interpolation value.
// -----------------------------------------------------------------------------
double CardinalCubicSpline::getY(double t) const
{
	TRY(CardinalCubicSpline::getY);
	int lo = 1, hi = x_.Ncols(), i = findInterval(t, x_);
	if (i < lo || hi - lo == 1) {
		// Linear extrapolation and interpolation for 2 points:
		return (y_(lo) + (t - x_(lo)) * (y_(lo + 1) - y_(lo)) / 
				(x_(lo + 1) - x_(lo)));
	} else if (i == lo) {
		i = lo;
		double u = (t - x_(i)) / (x_(i + 1) - x_(i));
		double u2 = u * u;
		double c1 = 1.0 + u * (-1.0 + u * (u - 1.0) / 2.0),
           c2 = u * (1.0 + u * (1.0 - u)), c3 = u2 * (u - 1.0) / 2.0;
		return (c1 * y_(i) + c2 * y_(i + 1) + c3 * y_(i + 2));
	} else if (i >= hi) {
		// Linear extrapolation
		return (y_(hi) + (t - x_(hi)) * (y_(hi) - y_(hi - 1)) / 
				(x_(hi) - x_(hi - 1)));
	} else if (i == hi - 1) {
		i = hi - 1;
		double u = (x_(i + 1) - t) / (x_(i + 1) - x_(i));
		double u2 = u * u;
		double c1 = 1.0 + u * (-1.0 + u * (u - 1.0) / 2.0),
           c2 = u * (1.0 + u * (1.0 - u)), c3 = u2 * (u - 1.0) / 2.0;
		return (c1 * y_(i + 1) + c2 * y_(i) + c3 * y_(i - 1));
	} else {
		double u  = (t - x_(i)) / (x_(i + 1) - x_(i));
		double u2 = u * u, u3 = 1.0 - u;
		double u4 = u3 * u3;
		double c1 = -u4 * u / 2.0, c2 = 1.0 + u2 * (3.0 * u - 5.0) / 2.0,
			c3 = u * (1.0 + u * (4.0 - 3.0 * u)) / 2.0, c4 = -u2 * u3 / 2.0;
		return (c1 * y_(i - 1) + c2 * y_(i) + c3 * y_(i + 1) + c4 * y_(i + 2));
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: deriv1
/// \brief Initial derivatives at boundaries of data by quadratic Newton 
/// interpolation.
// -----------------------------------------------------------------------------
static double deriv1(const RowVector & x, const RowVector & y, int i, int sgn)
{
	TRY(deriv1);
	int i1 = i + 1, i2 = i - 1;
	double his = x(i1) - x(i2), di = (y(i1) - y(i)) / (x(i1) - x(i));
	double dis = (y(i1) - y(i2)) / his;
	double di2 = (di - dis) / (x(i) - x(i2));
	return dis + sgn * di2 * his;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: deriv2
/// \brief Initial 2nd derivatives at boundaries of data by quadratic Newton 
/// interpolation.
// -----------------------------------------------------------------------------
static double deriv2(const RowVector & x, const RowVector & y, int i)
{
	TRY(deriv2);
	int i1 = i + 1, i2 = i - 1;
	double hi0 = x(i) - x(i2);
	double di0 = (y(i) - y(i2)) / hi0;
	double di1 = (y(i1) - y(i)) / (x(i1) - x(i));
	double di2 = (di1 - di0) / (x(i1) - x(i2));
	return di0 + di2 * hi0;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: fritsch
/// \brief Fritsch-Carlson iteration to adjust the monotone cubic interpolant. 
/// The iteration converges with cubic order.
// -----------------------------------------------------------------------------
static void fritsch(const RowVector & x, const RowVector & y, RowVector & d)
{
	TRY(fritsch);
	int i, i1;
	volatile int stop;
	double d1, r2, t;

	// Note: should never reach this! Currently it can happen when the curve
	// is not strictly monotone. In the future this case should be handled more
	// gracefully without wasting CPU time:
	const int max_loop = 20;
	int loop = 0;
	do {
		stop = true;
		for (i = 1; i < x.Ncols(); i++) {
			i1 = i + 1;
			d1 = (y(i1) - y(i)) / (x(i1)-x(i));
			if (d1 == 0.0)
				d(i) = d(i1) = 0.0;
			else {
				t = d(i) / d1;
				r2 = t * t;
				t = d(i1) / d1;
				r2 += t * t;
				if (r2 > 9.0) {
					t = 3.0 / sqrt(r2);
					d(i) *= t;
					d(i1) *= t;
					stop = false;
				}
			}
		}
	} while (!stop && ++loop < max_loop);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: FritschCarlsonCubicSpline::interpolate
/// \brief Calculates the Fritsch-Carlson monotone cubic spline interpolation 
/// given abscissa and ordinates vectors. All vectors must have conformant 
/// dimenions. The abscissa vector must be strictly increasing.
/// \param x the vectors (lo,hi) of data abscissas.
/// \param y the vectors (lo,hi) of data ordinates.
// -----------------------------------------------------------------------------
int FritschCarlsonCubicSpline::interpolate(const RowVector & x, 
										   const RowVector & y)
{
	TRY(FritschCarlsonCubicSpline::interpolate);
	// Check input parameters:
	if (x.Ncols() != y.Ncols())
		THROW_FATAL("FritschCarlsonCubicSpline::interpolate: index range vectors" 
					<< " mismatch");
	// Copy original data vectors into base class:
	x_ = x;
	y_ = y;

	// Empty data vectors - make auxilliary storage empty:
	if (x_.Ncols() == 0) {
		y1_ = RowVector();
		y2_ = RowVector();
		y3_ = RowVector();
		return 0;				// ok
	}
	int lo = 1, hi = x_.Ncols();
	
	// ReSize the auxilliary vectors:
	y1_.ReSize(x_);
	y2_.ReSize(x_);
	y3_.ReSize(x_);

	if (x_.Ncols() == 1) {
		// Default derivative is 0.0:
		y1_(lo) = y2_(lo) = y3_(lo) = 0.0;
	} else if (x.Ncols() == 2) {
		// Set derivatives for a line:
		y1_(lo) = y1_(hi) = (y_(hi) - y_(lo)) / (x_(hi) - x_(lo));
		y2_(lo) = y2_(hi) = y3_(lo) = y3_(hi) = 0.0;
	} else { // three or more points
		// Initial guess derivative vector:
		y1_(lo) = deriv1(x_, y_, lo + 1, -1);
		y1_(hi) = deriv1(x_, y_, hi - 1, 1);
		for (int i = lo + 1; i < hi; i++)
			y1_(i) = deriv2(x_, y_, i);
		
		if (x_.Ncols() > 3) {
			// Adjust derivatives at boundaries:
			if (y1_(lo) * y1_(lo + 1) < 0) 
				y1_(lo) = 0;
			if (y1_(hi) * y1_(hi - 1) < 0) 
				y1_(hi) = 0;
			// Adjustment of cubic interpolant:
			fritsch(x_, y_, y1_);
		}
		// Calculate remaining spline coefficients y2_(i) and y3_(i):
		CubicSplineCoefficients(x_, y_, y1_, y2_, y3_);
	}
	return 0;					// ok
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: PolynomialSpline::interpolate
/// \brief Calculates the polynomial interpolation using Aitken-Neville 
/// tableaux given abscissa and ordinates vectors.
/// \param x the vectors (lo,hi) of data abscissas.
/// \param y the vectors (lo,hi) of data ordinates.
// -----------------------------------------------------------------------------
int PolynomialSpline::interpolate(const RowVector & x, const RowVector & y)
{
	TRY(PolynomialSpline::interpolate);
	// Check input parameters:
	if (x.Ncols() != y.Ncols())
		THROW_FATAL("PolynomialSpline::interpolate: vectors have different size");
	
	// Copy original data vectors into base class:
	x_ = x;
	y_ = y;  

// 	// Empty data vectors - make auxilliary storage empty:
// 	if (x_.Ncols() == 0) {
// 		C_ = RowVector();
// 		D_ = RowVector();
// 	}
	return 0;					// ok
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: PolynomialSpline::getY
/// \brief Get the interpolation value (ordinate) at given control value.
/// Details: The returned y is the value corresponding to the value of the 
/// poynomial y = P(x) of degree n = hi-lo that interpolates the data points
/// (x(i),y(i)), lo <= i <= hi.In the special case of empty data vectors (x,y) 
/// a value of 0.0 is returned.
/// \param u the control parameter.
/// \return the interpolation value.
// -----------------------------------------------------------------------------
double PolynomialSpline::getY(double u) const
{
	TRY(PolynomialSpline::getY);
	// Special case that there is no data:
	if (x_.Ncols() == 0) return 0.0;
	
	int lo = 1, hi = x_.Ncols();
	// Initialize auxilliary storage:
	RowVector C = y_, D = y_;

	int pos = lo;				// find position of closest abscissa value
	double delta = abs(u - x_(lo)), delta2;
	for (int i = lo; i <= hi; i++) {
		if ((delta2 = abs(u - x_(i))) < delta ) {
			delta = delta2;
			pos = i;
		}
	}
 
	double dy, yy = C(pos--); 	// initial approximation
	for (int m = lo; m < hi; m++) {
		for (int i = lo; i <= hi - m; i++) {
			double h1 = x_(i) - u, h2 = x_(i + m) - u;
			double CD = C(i + 1) - D(i), denom;
			if ((denom = h1 - h2) == 0.0 ) 
				THROW_FATAL("PolynomialSpline::getY: two abscissa values " 
							<< "are identical");
			denom = CD / denom;
			C(i) = denom * h1;
			D(i) = denom * h2;
		}
		yy += (dy = (2 * pos - lo + 1 < (hi - m) ? C(pos + 1) : D(pos--)));
	}
	return yy;
	CATCH;
}
