/* -*- C++ -*- */
/****************************************************************
 *                                                              *
 *  Project name:      Biped Locomotion                         *
 *                                                              *
 *  Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)           *
 ****************************************************************
 * $Id: biped_interp.hpp,v 1.4 2004/08/15 20:04:41 s0343981 Exp $
 ****************************************************************/
/**
 * \file: biped_interp.hpp
 * \brief 1D Interpolation classes definitions.
 */

#ifndef __BIPED_INTERP_HPP__
#define __BIPED_INTERP_HPP__

#include <newmat/newmat.h>

// =============================================================================
//  Class: SplineBase                                                          
/// \brief Abstract class of 1-dimensional interpolation.
// =============================================================================
class SplineBase {
public:							// definitions
    enum Parametrization {		// curve parametrization methods
		No = 0,					// don't parametrize (default)
		Norm2,					// use sqrt(dx^2+dy^2)
		SqrNorm2,				// use (dx^2+dy^2)
		Norm1					// use |dx| + |dy|
    };

	enum SplineType { AKIMA = 0, BEZIER, CARDINAL, FRITSCH, POLYNOMIAL };

public:							// methods
    SplineBase(void) : x_() , y_() { }
    virtual ~SplineBase(void) { }
    virtual int interpolate(const RowVector & x, const RowVector & y) = 0;
    virtual double getX(double x) const = 0;
    virtual double getY(double x) const = 0;
    virtual void parameterize(const RowVector & x, const RowVector & y, 
							  RowVector & t, int parametrization) const;
	virtual SplineBase * clone() const = 0;
    static SplineBase * createNewSpline(SplineType type);
	virtual const RowVector & getPointsX() const { return x_; }
	virtual const RowVector & getPointsY() const { return y_; }

protected:						// methods
    int findInterval(double u, const RowVector & x) const;
    double CubicSplineHorner(double u, const RowVector & x, const RowVector & y, 
							 const RowVector & y1, const RowVector & y2, 
							 const RowVector & y3) const;
    void CubicSplineCoefficients(const RowVector & x, const RowVector & y, 
								 const RowVector & y1, RowVector & y2, 
								 RowVector & y3) const;

protected:						// members
    RowVector x_, y_;
};

// =============================================================================
//  Class: AkimaCubicSpline                                                    
/// \brief Akima cubic spline interpolation class.
/// Details: Calclates the interpolation for the given abscissa vector x and 
/// ordinate vector y. All vectors must have conformant dimenions. The abscissa 
/// vector must be strictly increasing. See:
///   H. Akima, "A new method of interpolation and smooth curve fitting based 
///   on local procedures", J. Assoc. Comput. Mach. 17, 589-602, 1970.
// =============================================================================
class AkimaCubicSpline : public SplineBase {
public:							// methods
    virtual ~AkimaCubicSpline(void) { }
    virtual int interpolate(const RowVector & x, const RowVector & y);
    virtual double getX(double u) const { return u; }
    virtual double getY(double u) const
	{ return CubicSplineHorner(u, x_, y_, y1_, y2_, y3_); }
	virtual AkimaCubicSpline * clone() const 
	{ return new AkimaCubicSpline(*this); }

protected:						// members
    RowVector y1_, y2_, y3_;
};

// =============================================================================
//  Class: BezierCubicSpline
/// \brief Bezier cubic spline interpolation class.
/// Details: Calclates the interpolation for the given abscissa vector x and 
/// ordinate vector y. All vectors must have conformant dimenions:
/// T = [t^3 t^2 t 1] - is the polynomial basis vector.
/// M = 1/6 * [B1 B2 B3 B4]' - is the basis matrix of the Bezier spline.
/// B1 = [-1 3 -3 1] ; B2 = [3 -6 3 0] ; B3 = [-3 0 3 0] ; B4 = [1 4 1 0]
/// G = [ p1 p2 p3 p4 ]' - is the geometry vector of the control points.
/// BSpline(t) = T * M * G.
// =============================================================================
class BezierCubicSpline : public SplineBase {
public:
    BezierCubicSpline(void) { }
    virtual ~BezierCubicSpline(void) { }
    virtual int interpolate(const RowVector & x, const RowVector & y);
    virtual double getX(double u) const;
    virtual double getY(double u) const;
	virtual BezierCubicSpline * clone() const 
	{ return new BezierCubicSpline(*this); }
};

// =============================================================================
//  Class: CardinalCubicSpline
/// \brief Cardinal cubic spline interpolation class.
/// Details: Calclates the interpolation for the given abscissa vector x and 
/// ordinate vector y. All vectors must have conformant dimenions:
/// T = [t^3 t^2 t 1] - is the polynomial basis vector.
/// M = 0.5 * [C1 C2 C3 C4]' - is the basis matrix of the Cardinal spline.
/// C1 = [-1 3 -3 1] ; C2 = [2 -5 4 -1] ; C3 = [-1 0 1 0] ; C4 = [0 2 0 0]
/// G = [ p1 p2 p3 p4 ]' - is the geometry vector of the control points.
/// CSpline(t) = T * M * G.
// =============================================================================
class CardinalCubicSpline : public SplineBase {
public:							// methods
    CardinalCubicSpline(void) { }
    virtual ~CardinalCubicSpline(void) { }
    virtual int interpolate(const RowVector & x, const RowVector & y);
    virtual double getX(double u) const;
    virtual double getY(double u) const;
	virtual CardinalCubicSpline * clone() const 
	{ return new CardinalCubicSpline(*this); }
};

// =============================================================================
//  Class: FritschCarlsonCubicSpline                                         
/// \brief Calculates the Fritsch-Carlson monotone cubic spline interpolation 
/// for a given abscissa vector x and ordinate vector y. All vectors must have
/// conformant dimenions. The abscissa vector must be strictly increasing.
/// Details: The Fritsch-Carlson interpolation produces a neat monotone 
/// piecewise cubic curve, which is especially suited for the presentation of 
/// scientific data. This is state of the art to create curves that preserve 
/// monotonicity. See details in:
///   F.N.Fritsch, R.E.Carlson, "Monotone Piecewise Cubic Interpolation", SIAM 
///   Journal of Numerical Analysis, Vol 17, No. 2, April 1980.
// =============================================================================
class FritschCarlsonCubicSpline : public SplineBase {
public:
    virtual ~FritschCarlsonCubicSpline(void) { }
    virtual int interpolate(const RowVector & x, const RowVector & y);
    virtual double getX(double u) const { return u; }
    virtual double getY(double u) const 
	{ return CubicSplineHorner(u, x_, y_, y1_, y2_, y3_); }
	virtual FritschCarlsonCubicSpline * clone() const 
	{ return new FritschCarlsonCubicSpline(*this); }

protected:
    RowVector y1_, y2_, y3_;
};

// =============================================================================
//  Class: PolynomialSpline
/// \brief  Polynomial interpolation using the Aitken-Neville tableaux class.
/// Details: The returned y is the value that corresponds to the value of the 
/// poynomial y = P(x) of degree n = hi-lo that interpolates the data points
/// (x(i),y(i)), lo <= i <= hi. In the special case of empty data vectors (x,y) 
/// a value of 0.0 is returned.
// =============================================================================
class PolynomialSpline : public SplineBase {
public:							// methods
    virtual ~PolynomialSpline(void) { }
    virtual int interpolate(const RowVector & x, const RowVector & y);
    virtual double getX(double u) const { return u; }
    virtual double getY(double u) const;
	virtual PolynomialSpline * clone() const 
	{ return new PolynomialSpline(*this); }
};

#endif	// __BIPED_INTERP_HPP__
