/* -*- C++ -*- */
/****************************************************************
 *                                                              *
 *  Project name:      Biped Locomotion                         *
 *                                                              *
 *  Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)           *
 ****************************************************************
 * $Id: biped_matrix_io.cpp,v 1.7 2004/08/15 20:04:41 s0343981 Exp $
 ****************************************************************/
/**
 * \file: biped_matrix_io.cpp
 * \brief Matrix IO classes implementations.
 */

#include <sstream>
#include <fstream>
#include <newmat/newmat.h>
#include <biped_exception.hpp>
#include <biped_logging.hpp>
#include <biped_tokenizer.hpp>
#include <biped_matrix_io.hpp>

using std::ifstream;
using std::ofstream;

// -----------------------------------------------------------------------------
//  Function: MatrixFileIO::write
/// \brief Write matrix data to the file using default data names.
/// \param data a vector of matrix data.
// -----------------------------------------------------------------------------
void MatrixFileIO::write(const vector< Matrix > & data)
{
	TRY(MatrixFileIO::write);
	vector< string > titles;
	string tmp = "data#";
	for (uint i = 0; i < data.size(); i++) {
		ostringstream intStream;
		intStream << i;
		titles.push_back(tmp + intStream.str());
	}
	return MatrixFileIO::write(data, titles);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: MatrixFileIO::write
/// \brief Write matrix data to the file using specified data names.
/// Details: If the file "filename" does not exist yet, creates it. The file is 
/// organized as follows:
///    line 1: the number(n) of matrix data.
///    line 2: matrix_1 : dim= R1 x C1 : [ matrix_1 values (space separated) ]; 
///    line i+1: matrix_i : dim= Ri x Ci : [ matrix_i values (space separated) ];
///    line n+2: END
/// \param data a vector of matrix data.
/// \param titles Name of each data member (i.e. speed, time, velocity, ...)
// -----------------------------------------------------------------------------
void MatrixFileIO::write(const vector< Matrix > & data, 
						 const vector< string > & titles)
{
	TRY(MatrixFileIO::write);
	const char * filenameCPtr = filename_.c_str();
	uint dataSize = data.size();
	if (dataSize != titles.size()) {
		THROW_FATAL("MatrixFileIO::write: data size and no. of titles " 
					<< "are different!");
	} else if (dataSize) {
		ofstream outFile(filenameCPtr);
		if (!outFile) 
			THROW_FATAL("MatrixFileIO::write: couldn't open file " << filename_);			
		// Write header information:
		outFile << "data_size: " << dataSize << endl;
		for(uint i = 0; i < dataSize; i++) {
			// Start with name and dimensions:
			outFile << titles[i] << " : dim= " << data[i].Nrows() << " x "
					<< data[i].Ncols() << " : [ ";
			// Now actual data:
			for (int j = 1; j <= data[i].Nrows(); j++) {
				for (int k = 1; k <= data[i].Ncols(); k++) {
					outFile << data[i](j, k) << " ";
				}
			}
			outFile << "];" << endl;
		}
		outFile << "END" << endl;
	} else {
		THROW_ERROR("MatrixFileIO::write: vector data is empty!");
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: MatrixFileIO::read
/// \brief Read matrix data from the file and store it and its names.
/// Details: The names are given according to the file's signature.
/// \param data a vector of matrix data (output parameter).
/// \param titles Name of each data member (output parameter).
// -----------------------------------------------------------------------------
void MatrixFileIO::read(vector< Matrix > & data, vector< string > & titles)
{
	TRY(MatrixFileIO::read);
	const char * filenameCPtr = filename_.c_str();
	ifstream inFile(filenameCPtr, ios::in);
	if (!inFile)
		THROW_FATAL("MatrixFileIO::read: couldn't open file " << filename_);
		
	// Extract data size:
	string temp;
	inFile >> temp >> elementNum_;
	
	// Extract the data matrices:
	data.clear();
	titles.clear();
	int colsNum = 0, rowsNum = 0;
	for (uint i = 0; i < elementNum_; i++) {
		// Extract the matrix name:
		inFile >> temp; 
		titles.push_back(temp);
		// Extract dimensions of the matrix:
		inFile >> temp >> temp // ignore ':' and "dim="
			   >> rowsNum >> temp >> colsNum;
		// Ignore ':' and '[':
		inFile >> temp >> temp;
		// Extract matrix values:
		Matrix currMat(rowsNum, colsNum);
		for (int j = 1; j <= rowsNum; j++) {
			for (int k = 1; k <= colsNum; k++) {
				inFile >> currMat(j, k);
			}
		}
		// Ignore '];\n':
		inFile >> temp;
		// Make sure reached '];\n':
 		BIPED_ASSERT(temp.find("];") != string::npos);
		data.push_back(currMat);
	}
	
	// Make sure reached the 'END' now:
	inFile >> temp;
	BIPED_ASSERT(temp.find("END") != string::npos);
	CATCH;
}
