/* -*- C++ -*- */
/****************************************************************
 *                                                              *
 *   Project name:      Biped Locomotion                        *
 *                                                              *
 *   Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)          *
 ****************************************************************
 * $Id: biped_sem.cpp,v 1.5 2004/08/06 18:02:43 s0343981 Exp $
 ****************************************************************/
/**
 * \file: biped_sem.cpp
 * \brief
 */

#include "biped_exception.hpp"
#include "biped_sem.hpp"

// -----------------------------------------------------------------------------
//  Function: Semaphore::control
/// \brief Wrapper interface method for semctl().
// -----------------------------------------------------------------------------
int Semaphore::control(int cmd, int value, u_short semnum) const
{
	TRY(Semaphore::control);
	if (internalID_ == -1)
		return -1;
	else {
		semun semctl_arg;
		semctl_arg.val = value;
		return semctl(internalID_, semnum, cmd, semctl_arg);
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Semaphore::op
/// \brief General Semaphore operation. Increment or decrement by a specific 
/// amount (positive or negative; amount can`t be zero).
/// Details: Wrapper interface method for semop().
// -----------------------------------------------------------------------------
int Semaphore::op(int val, u_short n, int flags) const
{
	TRY(Semaphore::op);
	sembuf op_op;
	op_op.sem_num = n;
	op_op.sem_flg = flags;

	if (internalID_ == -1)
		return -1;
	else if ((op_op.sem_op = val) == 0)
		return -1;
	else
		return semop(internalID_, &op_op, 1);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Semaphore::open
/// \brief Open or create one or more Semaphores.
/// \return 0 if all is OK, else -1.
// -----------------------------------------------------------------------------
int Semaphore::open(key_t k, int flags, int initial_value, u_short n, int perms)
{
	TRY(Semaphore::open);
	union semun ivalue;

	if (k == IPC_PRIVATE || k == static_cast< key_t >(INVALID_SEM_KEY))
		return -1;

	ivalue.val = initial_value;
	key_ = k;
	semNum_ = n;

	if ((internalID_ = semget(key_, n, perms | flags)) == -1)
		return -1;

	if (BIT_ENABLED(flags, IPC_CREAT)) {
		for (int i = 0; i < n; i++) {
			if (semctl(internalID_, i, SETVAL, ivalue) == -1)
				return -1;
		}
	}
	return 0;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Semaphore::Semaphore
/// \brief C'tor with specified semaphore values.
// -----------------------------------------------------------------------------
Semaphore::Semaphore(key_t k, int flags, int initial_value, u_short n, 
					 int perms) : key_(k)
{
	TRY(Semaphore::Semaphore);
	if (open(k, flags, initial_value, n, perms) == -1) {
		THROW_ERROR("Cannot open the specified semaphore: key=" << k 
					<< " flags=" << flags << " initial_value=" << initial_value 
					<< " n=" << n << " perms=" << perms);
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Semaphore::remove
/// \brief Remove all Semaphores associated with a particular key. 
/// Details: Intended to be called from a server, e.g. when it is being shut 
/// down, as we do an IPC_RMID on the Semaphore, regardless of whether other 
/// processes may be using it or not. Most other processes should use close().
// -----------------------------------------------------------------------------
int Semaphore::remove(void) const
{
	TRY(Semaphore::remove);
	int result = control(IPC_RMID);
	((Semaphore *) this)->init();
	return result;
	CATCH;
}
