/* -*- C++ -*- */
/****************************************************************
 *                                                              *
 *   Project name:      Biped Locomotion                        *
 *                                                              *
 *   Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)          *
 ****************************************************************
 * $Id: biped_sem.hpp,v 1.5 2004/08/06 18:02:43 s0343981 Exp $
 ****************************************************************/
/**
 * \file: biped_sem.hpp
 * \brief
 */

#ifndef __BIPED_SEM_HPP__
#define __BIPED_SEM_HPP__

#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#define DEFAULT_FILE_PERMS 0644
#define INVALID_SEM_KEY    -1
#define BIT_ENABLED(word, bit) (((word) & (bit)) != 0)

union semun {
	int val;					/// value for SETVAL
	struct semid_ds * buf;		/// buffer for IPC_STAT & IPC_SET
	u_short * array;			/// array for GETALL & SETALL
};

// =============================================================================
//  Class: Semaphore                                                           
/// \brief This is a simple semaphore class that assumes there are no race 
/// conditions for initialization (i.e., the order of process startup must be 
/// well defined). 
// =============================================================================
class Semaphore {
public:
	enum { SEM_CREATE = IPC_CREAT, SEM_EXCL = IPC_EXCL, SEM_OPEN = 0 };

	Semaphore(void) { init(); }

	Semaphore(key_t key, int flags = Semaphore::SEM_CREATE, 
			  int initial_value = 1, u_short nsems = 1,
			  int perms = DEFAULT_FILE_PERMS);
	~Semaphore(void) { close(); }

	/// Open or create one or more Semaphores. Returns 0 if all OK, else -1.
	int open(key_t key, int flags = Semaphore::SEM_CREATE, 
			 int initial_value = 1, u_short nsems = 1,
			 int perms = DEFAULT_FILE_PERMS);
	
	/// Close a Semaphore, marking it as invalid for subsequent operations.
	int close(void)
	{ return init(); }

	/// Remove all Semaphores associated with a particular key. This call is
	/// intended to be called from a server, e.g. when it is being shut down,
	/// as we do an IPC_RMID on the Semaphore, regardless of whether other 
	/// processes may be using it or not. Most other processes should use 
	/// <close>.
	int remove(void) const;

	/// Wait until a Semaphore's value is greater than 0, then decrement it
	/// by 1 and return (i.e. P() operation).
	int acquire(u_short n = 0, int flags = 0) const
	{ return op(-1, n, flags); }

	/// Acquire a semaphore for reading.
	int acquire_read(u_short n = 0, int flags = 0) const
	{ return acquire(n, flags); }

	/// Acquire a semaphore for writing
	int acquire_write(u_short n = 0, int flags = 0) const
	{ return acquire(n, flags); }

	/// Non-blocking version of <acquire>.
	int tryacquire(u_short n = 0, int flags = 0) const
	{ return op(-1, n, flags | IPC_NOWAIT); }

	/// Try to acquire the semaphore for reading.
	int tryacquire_read(u_short n = 0, int flags = 0) const
	{ return tryacquire(n, flags); }

	/// Try to acquire the semaphore for writing.
	int tryacquire_write(u_short n = 0, int flags = 0) const
	{ return tryacquire(n, flags); }

	/// Increment Semaphore by one (i.e. V() operation).
	int release(u_short n = 0, int flags = 0) const
	{ return op(1, n, flags); }

	/// General Semaphore operation. Increment or decrement by a specific amount
	/// (positive or negative; amount can`t be zero).
	int op(int val, u_short semnum = 0, int flags = SEM_UNDO) const;

	/// General Semaphore operation on an array of Semaphores.
	int op(sembuf op_vec[], u_short nsems) const
	{ return (internalID_ == -1) ? -1 : semop(internalID_, op_vec, nsems); }

	// Semaphore control methods.
	int control(int cmd, semun arg, u_short n = 0) const
	{ return (internalID_ == -1) ? -1 : semctl(internalID_, n, cmd, arg); }
	int control(int cmd, int value = 0, u_short n = 0) const;

	/// Get underlying internal id.
	int getID(void) const
	{ return internalID_; }

protected:						// methods
	int init(key_t k = static_cast< key_t >(INVALID_SEM_KEY), int i = -1)
	{ key_ = k; internalID_ = i; return 0; }

protected:						// members
	/// Semaphore key.
	key_t key_;			
	
	/// Internal ID to identify the semaphore group within this process.
	int internalID_;

	/// Number of semaphores we're creating.
	int semNum_;

};

#endif	// __BIPED_SEM_HPP__
