/* -*- C++ -*- */
/****************************************************************
 *                                                              *
 *   Project name:      Biped Locomotion                        *
 *                                                              *
 *   Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)          *
 ****************************************************************
 * $Id: biped_shmem.cpp,v 1.5 2004/08/06 18:02:44 s0343981 Exp $
 ****************************************************************/
/**
 * \file: biped_shmem.cpp
 * \brief class SharedMemory implementations.
 */

#include <iostream>
#include "biped_exception.hpp"
#include "biped_shmem.hpp"

using namespace std;

// -----------------------------------------------------------------------------
//  Function: SharedMemory::SharedMemory
/// \brief Constructor interface to this->openAndAttach () member function.
/// \param externalID the external key for accessing this shared memory.
/// \param sz the required size for the shared memory.
/// \param create the creation flags to be passed to the shared memory.
/// \param perms the file permissions to be created for this shared memory.
/// \param virtualAddr a virtual address which should be attached.
/// \param flags shared memory specific flags.
// -----------------------------------------------------------------------------
SharedMemory::SharedMemory(key_t externalID, size_t sz, int create, int perms,
						   void * virtualAddr, int flags)
{
	TRY(SharedMemory::SharedMemory);
	if (openAndAttach(externalID, sz, create, perms, virtualAddr, flags) == -1)
		cerr << "SharedMemory constructor failed!" << endl;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: SharedMemory::SharedMemory
/// \brief The "do nothing" constructor. 
// -----------------------------------------------------------------------------
SharedMemory::SharedMemory(void)
	: internalID_(0), size_(0), segmentPtr_(0)
{

}

// -----------------------------------------------------------------------------
//  Function: SharedMemory::open
/// \brief Creates a shared memory segment of SIZE bytes. 
/// Details: Does *not* attach this memory segment.
/// \param externalID the external key for accessing this shared memory.
/// \param sz the required size for the shared memory.
/// \param create the creation flags to be passed to the shared memory.
/// \param perms the file permissions to be created for this shared memory.
/// \return -1 if failed to allocate a the shared memory segment, otherwise - 0.
// -----------------------------------------------------------------------------
int SharedMemory::open(key_t externalID, size_t sz, int create, int perms)
{
	TRY(SharedMemory::open);
	segmentPtr_ = 0;
	size_ = roundUp(sz);
	if (size_ != sz) {
		cout << "SharedMemory::open: Warning: Given shared memory size " 
			 << "isn't page aligned!" << endl;
	}
	internalID_ = shmget(externalID, sz, create | perms);
	return (internalID_ == -1) ? -1 : 0;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: SharedMemory::openAndAttach
/// \brief Creates a shared memory segment of SIZE bytes and *does* attach to 
/// this segment.
/// \param externalID the external key for accessing this shared memory.
/// \param sz the required size for the shared memory.
/// \param create the creation flags to be passed to the shared memory.
/// \param perms the file permissions to be created for this shared memory.
/// \param virtualAddr a virtual address which should be attached.
/// \param flags shared memory specific flags.
/// \return -1 if failed to allocate a the shared memory segment, otherwise - 0. 
// -----------------------------------------------------------------------------
int SharedMemory::openAndAttach(key_t externalID, size_t sz, int create, 
								int perms, void * virtualAddr, int flags)
{
	TRY(SharedMemory::openAndAttach);
	if (open(externalID, sz, create, perms) == -1)
		return -1;
	else if (attach(virtualAddr, flags) == -1)
		return -1;
	else
		return 0;
	CATCH;
}
