/* -*- C++ -*- */
/****************************************************************
 *                                                              *
 *   Project name:      Biped Locomotion                        *
 *                                                              *
 *   Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)          *
 ****************************************************************
 * $Id: biped_shmem.hpp,v 1.5 2004/08/06 18:02:44 s0343981 Exp $
 ****************************************************************/
/**
 * \file: biped_shmem.hpp
 * \brief class SharedMemory definition
 * definition.
 */

#ifndef __BIPED_SHMEM__
#define __BIPED_SHMEM__

#include <CGAL/Cartesian.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include "biped_sem.hpp"

#define SHM_DEFAULT_FILE_PERMS DEFAULT_FILE_PERMS

// =============================================================================
//  Class: SharedMemory                                                        
/// \brief This is a wrapper for System V shared memory.
  // =============================================================================
class SharedMemory {
public:
	enum { SHM_CREATE = IPC_CREAT, SHM_OPEN = 0 };
	
	SharedMemory(void);
	SharedMemory(key_t externalID, size_t size, int create, 
				 int perms = SHM_DEFAULT_FILE_PERMS, void * virtualAddr = 0,
				 int flags = 0);

	/// Open a shared memory segment.
	int open(key_t externalID, size_t size, 
			 int create = SharedMemory::SHM_OPEN,
             int perms = SHM_DEFAULT_FILE_PERMS);

	/// Combine open() and attach().
	int openAndAttach(key_t externalID, size_t size,
					  int create = SharedMemory::SHM_OPEN,
					  int perms = SHM_DEFAULT_FILE_PERMS,
					  void * virtualAddr = 0, int flags = 0);
	
	/// Attach this shared memory segment.
	int attach(void * virtualAddr = 0, int flags = 0)
	{  
		segmentPtr_ = shmat(internalID_, virtualAddr, flags);
		return segmentPtr_ == (void *) -1 ? -1 : 0;
	}

	/// Detach this shared memory segment.
	int detach(void)
	{ return shmdt(segmentPtr_); }

	/// Remove this shared memory segment.
	int remove(void)
	{ return shmctl(internalID_, IPC_RMID, 0); }

	/// Forward to underlying System V <shmctl>.
	int control(int cmd, void * buf) 
	{ return shmctl(internalID_, cmd, (struct shmid_ds *) buf); }

	// Segment-related info:
	void * getSegmentPtr(void) const
	{ return segmentPtr_; }

	size_t getSegmentSize(void) const
	{ return size_; }

	/// Return the ID of the shared memory segment.
	int getID(void) const 
	{ return internalID_; }

protected:						// methods
	/// Round up to an appropriate page size.
	size_t roundUp(size_t len) 
	{ return (len + ALIGN_WORDB - 1) & ~(ALIGN_WORDB - 1); }
	
protected:						// members
	enum { ALIGN_WORDB = 8 };	/// Most restrictive alignment.

	/// Internal identifier.
	int internalID_;

	/// Size of the mapped segment.
	size_t size_;

	/// Pointer to the beginning of the segment.
	void * segmentPtr_;
};

#endif	// __BIPED_SHMEM__
