/* -*- C++ -*- */
/****************************************************************
 *                                                              *
 *   Project name:      Biped Locomotion                        *
 *                                                              *
 *   Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)          *
 ****************************************************************
 * $Id: biped_supervisor.cpp,v 1.9 2004/08/29 11:09:38 s0343981 Exp $
 ****************************************************************/
/**
 * \file: biped_supervisor.cpp
 * \brief class Supervisor methods' implementations.
 */

#include <iostream>
#include <sstream>
#include <device/supervisor.h>
#include <device/receiver.h>
#include <device/robot.h>
#include <biped_exception.hpp>
#include <biped_logging.hpp>
#include <biped_supervisor.hpp>

using std::cout;
using std::cerr;
using std::endl;

// -----------------------------------------------------------------------------
//  Function: Supervisor::Supervisor
/// \brief Supervisor's c'tor. 
/// \param receiver Handle to the receiver device.
// -----------------------------------------------------------------------------
Supervisor::Supervisor(DeviceTag receiver, const string & robotDEF) : 
	receiver_(receiver) , robotNode_() , comNode_() , zmpNode_()
{
	TRY(Supervisor::Supervisor);
	// Get the indicator nodes references:
	comNode_ = supervisor_node_get_from_def(COM_IND_DEF);
	zmpNode_ = supervisor_node_get_from_def(ZMP_IND_DEF);
	robotNode_ = supervisor_node_get_from_def(robotDEF.c_str());
	robot_step(TIME_STEP);	// wait a bit 
	if (!supervisor_node_was_found(comNode_))
		cerr << "CoM indicator was not found" << endl;
	if (!supervisor_node_was_found(zmpNode_))
		cerr << "ZMP indicator node was not found" << endl;
	if (!supervisor_node_was_found(robotNode_))
		cerr << "Robot was not found" << endl;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Supervisor::initialStep
/// \brief Run the initial step of the supervisor.
/// Details: Used to initialize initial positions.
// -----------------------------------------------------------------------------
void Supervisor::initialStep() 
{
	TRY(Supervisor::initialStep);
	float currPos[7] = { 0, 0, 0, 0, 0 ,0 ,0 };
	supervisor_field_get(robotNode_, 
						 SUPERVISOR_FIELD_TRANSLATION_AND_ROTATION,
						 &currPos, TIME_STEP);
	robot_step(TIME_STEP * 2);
	memcpy(robotInitPos_, currPos, 7 * sizeof(float));
	PRINT_VECTOR7(robotInitPos_);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Supervisor::step
/// \brief Run a supervisor's single step. 
/// Details: The supervisor stays listening to any messages coming from the 
/// robot.
// -----------------------------------------------------------------------------
void Supervisor::step() 
{
	TRY(Supervisor::step);
	unsigned bufSize = receiver_get_buffer_size(receiver_);
	if (bufSize >= 8 * sizeof(float)) {
		float * msg = (float *) receiver_get_buffer(receiver_);
		if (msg[0] == 1) {		// update of CoM and ZMP
			float comTrans[3] = { msg[1], msg[2], msg[3] };
			float zmpTrans[3] = { msg[4], msg[5], msg[6] };
			float currTime[1] = { msg[7] };
			updateLabelAndIndicator(COM_ID, comTrans);
			updateLabelAndIndicator(ZMP_ID, zmpTrans);
			updateLabelAndIndicator(TIME_ID, currTime);
			float tmp = abs(sin(currTime[0] * M_PI / (SNAPSHOT_INTERVAL) + 
								M_PI_2));
			if (tmp == 1) 
				takeSnapshot(currTime[0]);
		} else if (msg[0] == 2) { // reset request
			resetRobotPosition();		   
		}
	}
	robot_step(TIME_STEP); // run one step
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Supervisor::updateLabelAndIndicator
/// \brief Update the label and transformation of the given indicator.
// -----------------------------------------------------------------------------
void Supervisor::updateLabelAndIndicator(LabelID id, float vals[])
{
	TRY(Supervisor::updateLabelAndIndicator);
	float labelX = (id == COM_ID) ? COM_LABEL_X : 
		((id == TIME_ID) ? TIME_LABEL_X : ZMP_LABEL_X);
	float labelY = (id == COM_ID) ? COM_LABEL_Y : 
		((id == TIME_ID) ? TIME_LABEL_Y : ZMP_LABEL_Y);
	uint labelColor = (id == COM_ID) ? COM_LABEL_COLOR : 
		((id == TIME_ID) ? TIME_LABEL_COLOR : ZMP_LABEL_COLOR);

	// Create the label string:
	std::stringstream posStream;
	if (id == TIME_ID) {
		posStream << "TIME: " << vals[0] << " [sec]";
	} else {
		posStream << ((id == COM_ID) ? "CoM: [" : "ZMP: [") 
				  << vals[0] << "," << vals[1] << "," << vals[2] << "]";
		// Set the indicator's new position in the scene:
		if (id == ZMP_ID) vals[1] += 0.001; // makes the ZMP more visible
		supervisor_field_set(((id == COM_ID) ? comNode_ : zmpNode_), 
							 SUPERVISOR_FIELD_TRANSLATION, vals);
	}

	// Update the label:
	supervisor_set_label(id, posStream.str().c_str(), labelX, labelY, 
						 LABEL_SIZE, labelColor); 	
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Supervisor::resetRobotPosition
/// \brief Resets the robot to starting position.
// -----------------------------------------------------------------------------
void Supervisor::resetRobotPosition()
{
	TRY(Supervisor::resetRobotPosition);
	cout << "Supervisor: Resetting the robot position!" << endl;
	PRINT_VECTOR7(robotInitPos_);
	supervisor_field_set(robotNode_, 
						 SUPERVISOR_FIELD_TRANSLATION_AND_ROTATION,
						 robotInitPos_);
	robot_step(TIME_STEP * 2);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: Supervisor::takeSnapshot
/// \brief Exports the current snapshot of the scene.
// -----------------------------------------------------------------------------
void Supervisor::takeSnapshot(float time)
{
	TRY(Supervisor::takeSnapshot);
	ostringstream filename;
	filename << "/tmp/snap" << time << ".jpg";
	supervisor_export_image(filename.str().c_str(), 100);
	CATCH;
}
