/* -*- C++ -*- */
/****************************************************************
 *                                                              *
 *    Project name:      Biped Locomotion                       *
 *                                                              *
 *   Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)          *
 ****************************************************************
 * $Id: biped_tokenizer.cpp,v 1.5 2004/08/06 18:02:44 s0343981 Exp $
 ****************************************************************/
/**
 * \file: biped_tokenizer.cpp
 * \brief class Tokenizer implementations.
 */

#include "biped_exception.hpp"
#include "biped_tokenizer.hpp"

int Tokenizer::delimiter(char d)
{
	TRY(Tokenizer::delimiter);
	if(delimiterInx_ == MAX_DELIMITERS)
		return -1;
	delimiters_[delimiterInx_].delimiter_ = d;
	delimiters_[delimiterInx_].replace_ = 0;
	delimiterInx_++;
	return 0;
	CATCH;
}

int Tokenizer::delimiterReplace(char d, char replacement)
{
	TRY(Tokenizer::delimiterReplace);
	// Make it possible to replace delimiters on-the-fly, e.g., parse string
	// until certain token count and then copy rest of the original string:
	for (int i = 0; i < delimiterInx_; i++) {
		if (delimiters_[i].delimiter_ == d) {
			delimiters_[i].replacement_ = replacement;
			delimiters_[i].replace_ = 1;
			return 0;
		}
	}

	if(delimiterInx_ >= MAX_DELIMITERS)
		return -1;

	delimiters_[delimiterInx_].delimiter_ = d;
	delimiters_[delimiterInx_].replacement_ = replacement;
	delimiters_[delimiterInx_].replace_ = 1;
	delimiterInx_++;
	return 0;
	CATCH;
}

int Tokenizer::preserveDesignators(char start, char stop, int strip)
{
	TRY(Tokenizer::preserveDesignators);
	if (preservesInx_ == MAX_PRESERVES)
		return -1;

	preserves_[preservesInx_].start_ = start;
	preserves_[preservesInx_].stop_ = stop;
	preserves_[preservesInx_].strip_ = strip;
	preservesInx_++;
	return 0;
	CATCH;
}

int Tokenizer::isDelimiter(char d, int & replace, char & r)
{
	TRY(Tokenizer::isDelimiter);
	replace = 0;
	for (int x = 0; x < delimiterInx_; x++) {
		if (delimiters_[x].delimiter_ == d)	{
			if(delimiters_[x].replace_)	{
				r = delimiters_[x].replacement_;
				replace = 1;
			}
			return 1;
		}
	}
	return 0;
	CATCH;
}

int Tokenizer::isPreserveDesignator(char start, char & stop, int & strip)
{
	TRY(Tokenizer::isPreserveDesignator);
	for (int x = 0; x < preservesInx_; x++) {
		if (preserves_[x].start_ == start) {
			stop = preserves_[x].stop_;
			strip = preserves_[x].strip_;
			return 1;
		}
	}
	return 0;
	CATCH;
}

char * Tokenizer::next(void)
{
	TRY(Tokenizer::next);
	// Check if the previous pass was the last one in the buffer.
	if (index_ == -1) {
		index_ = 0;
		return 0;
    }

	char replacement;
	int replace;
	char * next_token;

	// Skip all leading delimiters.
	for (;;) {
		// Check for end of string.
		if (buffer_[index_] == '\0') {
			// If we hit EOS at the start, return 0.
			index_ = 0;
			return 0;
        }

		if (isDelimiter(buffer_[index_], replace, replacement))
			index_++;
		else
			break;
    }
	// When we reach this point, buffer_[index_] is a non-delimiter and
	// not EOS - the start of our next_token.
	next_token = buffer_ + index_;

	// A preserved region is it's own token.
	char stop;
	int strip;
	if (isPreserveDesignator(buffer_[index_], stop, strip)) {
		while (++index_) {
			if (buffer_[index_] == '\0') {
				index_ = -1;
				goto EXIT_LABEL;
            }
			if (buffer_[index_] == stop)
				break;
        }
		
		if (strip) {
			// Skip start preserve designator.
			next_token += 1;
			// Zap the stop preserve designator.
			buffer_[index_] = '\0';
			// Increment to the next token.
			index_++;
        }
		goto EXIT_LABEL;
    }
	// Step through finding the next delimiter or EOS.
	for (;;) {
		// Advance pointer.
		index_++;

		// Check for delimiter.
		if (isDelimiter(buffer_[index_], replace, replacement)) {
			// Replace the delimiter.
			if (replace != 0)
				buffer_[index_] = replacement;
			
			// Move the pointer up and return.
			index_++;
			goto EXIT_LABEL;
        }
		
		// A preserve designator signifies the end of this token.
		if (isPreserveDesignator(buffer_[index_], stop, strip))
			goto EXIT_LABEL;
		
		// Check for end of string.
		if (buffer_[index_] == '\0') {
			index_ = -1;
			goto EXIT_LABEL;
        }
    }
  EXIT_LABEL:
	return next_token;
	CATCH;
}
