/* -*- C++ -*- */
/****************************************************************
 *                                                              *
 *    Project name:      Biped Locomotion                       *
 *                                                              *
 *   Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)          *
 ****************************************************************
 * $Id: biped_tokenizer.hpp,v 1.5 2004/08/06 18:02:44 s0343981 Exp $
 ****************************************************************/
/**
 * \file: biped_tokenizer.hpp
 * \brief class Tokenizer definition.
 */

#ifndef __BIPED_TOKENIZER_HPP__
#define __BIPED_TOKENIZER_HPP__

// =============================================================================
//  Class: Tokenizer                                                           
/// \brief Tokenizes a buffer. Allows application to set delimiters and preserve
/// designators.
// =============================================================================
class Tokenizer {
public:
	/// Notice that Tokenizer will modify buffer if you use delimiterReplace or
	/// preserveDesignators to do character substitution. NOTE: You should NOT
    /// pass a constant string or string literal to this constructor, since 
	/// Tokenizer will try to modify the string.
	Tokenizer(char * buffer) : buffer_(buffer), index_(0), preservesInx_(0),
							   delimiterInx_(0) { }

	/// Returns 0 on success, -1 if there is no memory left.
	int delimiter(char d);

	/// Replaces the delimiter d with the replacement when found.
	int delimiterReplace(char d, char replacement);

	/// Extract string between a pair of designator characters, e.g. quotes, 
	/// or '(' and ')'.
	int preserveDesignators(char start, char stop, int strip = 1);

	/// Returns the next token.
	char * next(void);
	int index() const { return index_; }

	enum { MAX_DELIMITERS = 16, MAX_PRESERVES = 16 };

protected:
	/// Returns 1 if <d> is a delimiter, 0 otherwise. If <d> should be replaced
	/// with <r>, <replace> is set to 1, otherwise 0.
	int isDelimiter(char d, int & replace, char & r);

	/// If <start> is a start preserve designator, returns 1 and sets <stop> to
	/// the stop designator. Returns 0 if <start> is not a preserve designator.
	int isPreserveDesignator(char start, char & stop, int & strip);

	char * buffer_;
	int index_;

	/// @class Preserve_Entry
    ///	@brief Defines a set of characters that designate an area that should
	/// not be parsed, but should be treated as a complete token. Example: in:
	/// (this is a preserve region), start would be a left paren -(- and stop
	/// would be a right paren -)-. The strip determines whether the designators
	/// should be removed from the token.
	class PreserveEntry {
	public:
		char start_;
		char stop_;
		int strip_;
	};

	/// The application can specify MAX_PRESERVES preserve designators.
	PreserveEntry preserves_[MAX_PRESERVES];

	/// Pointer to the next free spot in preserves_.
	int preservesInx_;

	/// @class DelimiterEntry
    ///	@brief Describes a delimiter for the tokenizer.
	class DelimiterEntry {
	public:
		/// Most commonly a space ' '. What occurrences of delimiter_ should
		/// be replaced with. Whether replacement_ should be used. This should
		/// be replaced with a technique that sets replacement_ = delimiter by
		/// default.
		char delimiter_;
		char replacement_;
		int replace_;
	};

	/// The tokenizer allows MAX_DELIMITERS number of delimiters.
	DelimiterEntry delimiters_[MAX_DELIMITERS];

	/// Pointer to the next free space in delimiters_.
	int delimiterInx_;
};

#endif	// __BIPED_TOKENIZER_HPP__
