/* -*- C++ -*- */
/****************************************************************
 *                                                              *
 *   Project name:        Biped Locomotion                      *
 *                                                              *
 *   Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)          *
 ****************************************************************
 * $Id: biped_util.hpp,v 1.4 2004/08/29 11:09:38 s0343981 Exp $
 ****************************************************************/
/**
 * \file: biped_util.hpp
 * \brief Macros and utility functions for biped robot control.
 */

#ifndef __BIPED_UTIL_HPP__
#define __BIPED_UTIL_HPP__

#include <limits>
#include <biped_exception.hpp>
#include <newmat/newmat.h>
#include <newmat/newmatap.h>

// -----------------------------------------------------------------------------
//  Function: pseudoInvImpl
/// \brief Computes the pseudo-inverse of a given matrix.
/// Details: Produces a matrix X of the same dimensions as M' so that:
/// M*X*M = M, X*M*X = X and M*X and M*A are Hermitian. The computation is based
/// on SVD(M) and any singular values less than a tolerance are treated as zero.
/// We use the following tolerance: max(M.Nrows, M.Ncols) * norm(M) * EPS.
/// \param M the matrix to invert.
/// \return The pseudo-inverse of M.
// -----------------------------------------------------------------------------
ReturnMatrix pseudoInvImpl(const Matrix & M)
{
	TRY(pseudoInvImpl);
	const double eps = std::numeric_limits< double >::epsilon();
	int m = M.Nrows();
	int n = M.Ncols();
	int maxRowsCols = std::max(m, n);
	Matrix pinvM(n, m);
	Matrix U, V;
	DiagonalMatrix D;

	// Perform Singular Value Decomposition:
	SVD(M, D, U, V);

	double tol = 0.0;
	uint r = 0; 
	if (m > 1) {
		tol = maxRowsCols * D.Maximum() * eps;
		for (int i = 1; i <= D.Nrows(); i++) {
			if (D(i, i) > tol)
				r++;
		}
	} else if (m == 1) {
		tol = maxRowsCols * D(1) * eps;
		r = (D(1) > tol) ? 1 : 0;
	} else {					// m <= 0
		r = 0;
	}

	if (r == 0) {
		pinvM = 0.0;
	} else if (r == 1) {
		pinvM = V.Columns(1, r) * (1 / D(1)) * (U.Columns(1, r)).t();
	} else {					// r > 1
		DiagonalMatrix S(r);
		for (uint i = 1; i <= r; i++)
			S(i, i) = 1.0 / D(i, i);
		pinvM = V.Columns(1, r) * S * (U.Columns(1, r)).t();
	}
	pinvM.Release();
	return pinvM;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: getPseudoInverse
/// \brief Computes the pseudo-inverse of a given matrix.
// -----------------------------------------------------------------------------
ReturnMatrix getPseudoInverse(const Matrix & M)
{
	TRY(getPseudoInverse);
	int m = M.Nrows();
	int n = M.Ncols();
	if (m < n) {					// Makes sure no. of columns is lesser equal
		Matrix res = pseudoInvImpl(M.t()).t();
		res.Release();
		return res;
	} else {
		return pseudoInvImpl(M);
	}
	CATCH;
}

#endif	// __BIPED_UTIL_HPP__
