/* -*- C++ -*- */
/****************************************************************
 *                                                              *
 *  Project name:      Biped Locomotion                         *
 *                                                              *
 *   Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)          *
 ****************************************************************
 * $Id: biped_xml_handler.cpp,v 1.8 2004/08/15 20:04:42 s0343981 Exp $
 ****************************************************************/
/**
 * \file: biped_xml_handler.cpp
 * \brief class XMLHandler implementations.
 */

#include <newmat/newmat.h>
#include <biped_exception.hpp>
#include <biped_logging.hpp>
#include <biped_tokenizer.hpp>
#include <biped_xml_handler.hpp>
#include <iostream>

using std::string;
using std::cout;
using std::cerr;
using std::endl;

// -----------------------------------------------------------------------------
//  Function: XMLHandler::unparsedEntityDecl
/// \brief Handles notification of an unparsed entity declaration.
/// \param name The entity name.
/// \param publicId The entity public ID, or empty string if unavailable.
/// \param systemId The entity system identifier.
/// \param notationName The name of the associated notation.
// -----------------------------------------------------------------------------
void XMLHandler::unparsedEntityDecl(
	const string & name, const string & publicId,
	const string & systemId, const string & notationName)
{
	TRY(XMLHandler::unparsedEntityDecl);
// 	cout << "unparsedEntityDecl: " << name << "(public=" << publicId 
// 		 << ",system=" << systemId << ",notation=" << notationName << ")\n";
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: XMLHandler::startDocument
/// \brief Handles notification of the beginning of the document. Overloaded to
/// initialize the motions sequences data structure.
// -----------------------------------------------------------------------------
void XMLHandler::startDocument()
{
	TRY(XMLHandler::startDocument);
// 	cout << "startDocument" << endl;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: XMLHandler::endDocument
/// \brief Handles notification of the end of the document.
// -----------------------------------------------------------------------------
void XMLHandler::endDocument()
{
	TRY(XMLHandler::endDocument);
// 	cout << "endDocument" << endl;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: XMLHandler::startElement
/// \brief Handles notification of the start of an element, namely, storing 
/// specific values defined for joints.
/// \param namespaceURI The Namespace URI, or empty string if the element has no
/// Namespace URI or if Namespace processing is not being performed.
/// \param localName The local name (without prefix), or empty string if 
/// Namespace processing is not being performed.
/// \param qName The qualified name (with prefix), or empty string if qualified
/// names are unavailable.
/// \param atts The attributes attached to element. If there are none, it shall
/// be an empty Attributes object. 
// -----------------------------------------------------------------------------
void XMLHandler::startElement(
	const string & namespaceURI, const string & localName,
	const string & qName, const SAX::Attributes & atts)
{
	TRY(XMLHandler::startElement);
// 	if(localName.length())
// 		cout << "Start Element: " << namespaceURI << ":" << localName << endl;
// 	else
// 		cout << "Start Element: " << qName << endl;
	currElement_ = qName;
	if (currElement_ == "JOINT_DEF") {
		jointNum_ = atoi(atts.getValue("number").c_str());
	} else if (currElement_ == "SEQ") {
		seqID_ = atts.getValue("id");
		isNewSeq_ = true;
	} else if (currElement_ == "JOINT") {
		jointID_ = atts.getValue("id");
	} else if (currElement_ == "TIME_DEF") {
		timeType_ = atts.getValue("type");
	}

	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: XMLHandler::endElement
/// \brief Handles notification of the end of an element, namely if end of a 
/// motion sequence - store it.
/// \param namespaceURI The Namespace URI, or empty string if element has no 
/// Namespace URI or if Namespace processing is not being performed.
/// \param localName The local name (without prefix), or empty string if 
/// Namespace processing is not being performed.
/// \param qName The qualified name (with prefix), or empty string if qualified
/// names are not available.
// -----------------------------------------------------------------------------
void XMLHandler::endElement(const string & namespaceURI, 
							const string & localName, const string & qName)
{
	TRY(XMLHandler::endElement);
// 	cout << "endElement: " << namespaceURI << ":" << localName << endl;
	if (qName == "SEQ") {
		motionsMap_[seqID_] = viaPoints_;
	}
	CATCH;
}

// =============================================================================
//  Struct: IsWhiteSpace                                                       
/// \brief Predicate functor which indicates a char being a whitespace char.
// =============================================================================
struct IsWhiteSpace {
	bool operator()(char c) const {
		return (c == ' ') || (c == '\t') || (c == '\n');
	}
};

// -----------------------------------------------------------------------------
//  Function: XMLHandler::characters
/// \brief Handles notification of character data inside an element.
/// Details: Overriden method to take specific actions for each chunk of 
/// character data. In our case, handling depends on the type of element.
/// \param ch The characters.
// -----------------------------------------------------------------------------
void XMLHandler::characters(const string & ch)
{
	TRY(XMLHandler::characters);
// 	cout << "Characters: " << ch << endl;
	uint whitespaceCount = count_if(ch.begin(), ch.end(), IsWhiteSpace());
	if (whitespaceCount == ch.length())
		return;					// do nothing ...
	if (ch.length() != 0) {
		// Create a vector of joint string names. this helps figure out in what
		// row of the viaPoints matrix the values for each joint belong.
		if (currElement_ == "JOINT_DEF") {
			jointStrings_.push_back(ch);
		}
		// Update the joint's sequence row in the via points matrix:
		else if (currElement_ == "JOINT") {
			handleJointSequence(ch);
		}
		// Update the time defintion in the via points matrix:
		else if (currElement_ == "TIME_DEF") {
			handleTimeDefinition(ch);
		}
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: XMLHandler::processingInstruction
/// \brief Receives notification of a processing instruction.
/// Details: By default, do nothing. We may override this method to take 
/// specific actions for each processing instruction, such as setting status 
/// variables or invoking other methods.
/// \param target The processing instruction target.
/// \param data The processing instruction data, or empty string if none.
// -----------------------------------------------------------------------------
void XMLHandler::processingInstruction(const string & target, const string & data)
{
	TRY(XMLHandler::processingInstruction);
// 	cout << "processingInstruction " << target << " - " << data << endl;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: XMLHandler::skippedEntity
/// \brief Warns if a skipped entity nitification was received.
/// \param name The name of the skipped entity.
// -----------------------------------------------------------------------------
void XMLHandler::skippedEntity(const string & name)
{
	TRY(XMLHandler::skippedEntity);
	cout << "WARNING: skippedEntity " << name << endl;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: XMLHandler::warning
/// \brief Receives notification of a parser warning. Overriden method to insert
/// the message in a log file or printing it to the console.
/// \param exception The warning information encoded as an exception.
// -----------------------------------------------------------------------------
void XMLHandler::warning(const SAX::SAXParseException & exception)
{
	TRY(XMLHandler::warning);
	cerr << "WARNING: " << exception.what() << endl;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: XMLHandler::error
/// \brief Receives notification of a recoverable parser error. Overriden method
/// to use THROW_ERROR.
/// \param exception The error information encoded as an exception.
// -----------------------------------------------------------------------------
void XMLHandler::error(const SAX::SAXParseException & exception)
{
	TRY(XMLHandler::error);
	THROW_ERROR(exception.what());
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: XMLHandler::fatalError
/// \brief Report a fatal XML parsing error. Override method to throw a qrio 
/// fatal error.
/// Details: According to the SAX interface the application must stop all
/// regular processing when this method is invoked, since the document is no 
/// longer reliable, and the parser may no longer report parsing events.
/// \param exception The error information encoded as an exception.
// -----------------------------------------------------------------------------
void XMLHandler::fatalError(const SAX::SAXParseException & exception)
{
	TRY(XMLHandler::fatalError);
	THROW_FATAL(exception.what());
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: XMLHandler::handleJointSequence
/// \brief Handles notification of character data which belongs to a JOINT 
/// sequence element
/// \param ch The characters.
// -----------------------------------------------------------------------------
void XMLHandler::handleJointSequence(const string & ch)
{
	TRY(XMLHandler::handleJointSequence);
	// String contains float position values:
	char str[ch.length()];
	strcpy(str, ch.c_str());

	// This will replace any space delimiter with '\0', thus separating
	// the tokens in the string:
	Tokenizer tok(str);
	tok.delimiterReplace(' ', '\0');	
	vector< string > tokens;
	char * nextToken = tok.next();
	for (; nextToken; nextToken = tok.next()) {
		tokens.push_back(string(nextToken));
	}

	uint posNum = tokens.size();
	// Create the float matrix on the first pass through a motion 
	// sequence. Initialize all joints to zero:
	if (isNewSeq_) {
		viaPoints_ = Matrix(jointStrings_.size() + 1, posNum);
		viaPoints_ = 0.0;
		isNewSeq_ = false;
	}

	BIPED_ASSERT(uint(viaPoints_.ncols()) == posNum);
	// Find integer index for the joint number
	int jointNum = find(jointStrings_.begin(), jointStrings_.end(), 
						jointID_) - jointStrings_.begin();

	// Enter joint positions in sequence motion matrix:
	for (uint p = 0; p < posNum; p++) {
		dReal jointPos = atof(tokens[p].c_str());
		// The XML file inverts positive/negative:
		viaPoints_(jointNum + 2, p + 1) = -jointPos; 
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: XMLHandler::handleTimeDefinition
/// \brief Handles notification of character data which belongs to a TIME_DEF 
/// sequence element
/// \param ch The characters.
// -----------------------------------------------------------------------------
void XMLHandler::handleTimeDefinition(const string & ch)
{
	TRY(XMLHandler::handleTimeDefinition);
	// String contains float position values:
	char str[ch.length()];
	strcpy(str, ch.c_str());

	// This will replace any space delimiter with '\0', thus separating
	// the tokens in the string:
	Tokenizer tok(str);
	tok.delimiterReplace(' ', '\0');	
	vector< string > tokens;
	char * nextToken = tok.next();
	for (; nextToken; nextToken = tok.next()) {
		tokens.push_back(string(nextToken));
	}
	
	uint posNum = tokens.size();
	if (timeType_ == "interval") { // defintion by supplying only interval
		BIPED_ASSERT(posNum == 1);
		dReal timeInterval = atof(tokens[0].c_str());
		for (int i = 1; i <= viaPoints_.Ncols(); i++) {
			viaPoints_(1, i) = i * timeInterval; 
		}
	} else if (timeType_ == "explicit") { // explicit sequence of times
		BIPED_ASSERT(uint(viaPoints_.ncols()) == posNum);
		// Enter joint positions in sequence motion matrix:
		for (uint p = 0; p < posNum; p++) {
			dReal time = atof(tokens[p].c_str());
			viaPoints_(1, p + 1) = time; 
		}
	} else {
		THROW_FATAL("Unknown type of time definition!");
	}
	CATCH;
}
