/* -*- c++ -*- */
/****************************************************************
 *                                                              *
 *  Project name:      Biped Locomotion                         *
 *                                                              *
 *  Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)           *
 ****************************************************************
 * $Id: qrio_joints.hpp,v 1.6 2004/08/29 11:09:43 s0343981 Exp $ 
 ****************************************************************/
/**
 * \file qrio_joints.hpp 
 */

#ifndef __QRIO_JOINTS_HPP__
#define __QRIO_JOINTS_HPP__

#include <cstring>
#include <biped_exception.hpp>
#include <biped_joints.hpp>
#include <biped_namespace.hpp>

#define JOINTS_NUM         22
#define JOINT_MAX_FORCE    9	// in Newton

#define ANKLE_PITCH_MIN_FORCE 4	// in Newton
#define ANKLE_ROLL_MIN_FORCE  4 // in Newton
#define ANKLE_PITCH_MAXVAL    0.7
#define ANKLE_PITCH_MINVAL    -0.7
#define ANKLE_ROLL_MAXVAL     0.7
#define ANKLE_ROLL_MINVAL     -0.7

BIPED_BEGIN_NAMESPACE(QRIO);

typedef enum { torso_pitch = 0, torso_roll,
			   left_hip_roll, left_hip_pitch, left_hip_yaw,
			   left_knee_pitch, 
			   left_ankle_roll, left_ankle_pitch,
			   left_shoulder_roll, left_shoulder_pitch, 
			   left_elbow_pitch,
			   neck_pitch, neck_yaw,
			   right_hip_roll, right_hip_pitch, right_hip_yaw,
			   right_knee_pitch,
			   right_ankle_roll, right_ankle_pitch,
			   right_shoulder_roll, right_shoulder_pitch,
			   right_elbow_pitch } JointID;

#define body_center (right_elbow_pitch + 1)

static string jointNames[JOINTS_NUM] = 
{ "torso_pitch", "torso_roll", "left_hip_roll", "left_hip_pitch", 
  "left_hip_yaw", "left_knee_pitch", "left_ankle_roll", "left_ankle_pitch",
  "left_shoulder_roll", "left_shoulder_pitch", "left_elbow_pitch",
  "neck_pitch", "neck_yaw", "right_hip_roll", "right_hip_pitch", 
  "right_hip_yaw", "right_knee_pitch", "right_ankle_roll", "right_ankle_pitch",
  "right_shoulder_roll", "right_shoulder_pitch", "right_elbow_pitch" };

static string jointStrings[JOINTS_NUM] = 
{ "back_1", "back_2", "left_hip_3", "left_hip_2", "left_hip_1", 
  "left_knee", "left_ankle_2", "left_ankle_1", "left_shoulder_2", 
  "left_shoulder_1", "left_elbow", "neck_tilt", "neck", "right_hip_3", 
  "right_hip_2", "right_hip_1", "right_knee", "right_ankle_2", 
  "right_ankle_1", "right_shoulder_2", "right_shoulder_1", "right_elbow" };
 
static string jointDEFs[JOINTS_NUM + 1] = 
{ "back_1_0", "back_2_0", "left_hip_3_0", "left_hip_2_0", "left_hip_1_0", 
  "left_knee_0", "left_ankle_2_0", "left_ankle_1_0", "left_shoulder_2_0", 
  "left_shoulder_1_0", "left_elbow_0", "neck_tilt_0", "neck_0", 
  "right_hip_3_0", "right_hip_2_0", "right_hip_1_0", "right_knee_0", 
  "right_ankle_2_0", "right_ankle_1_0", "right_shoulder_2_0", 
  "right_shoulder_1_0", "right_elbow_0", "Qrio_0" };

static float jointMinLimits[JOINTS_NUM] = 
{ -1, -0.5, -0.5, -0.9, -1.4, -0.2, -0.7, -0.7, 0, -3.14, 0, -0.3, -2, 
  -0.5, -0.9, -1.4, -0.2, -0.7, -0.7, 0, -3.14, 0 };

static float jointMaxLimits[JOINTS_NUM] = 
{ 2, 0.5, 1, 1.6, 1.7, 2.5, 0.7, 0.7, 3.14, 3.14, 2.8, 1.1, 2, 
  1, 1.6, 1.7, 2.5, 0.7, 0.7, 3.14, 3.14, 2.8 };

class QRIO_JointInfo : public JointInfo {
public:
	const char * toNameString(int joint) 
	{
		TRY(QRIO_JointInfo::toNameString);
		if (joint < 0 || joint >= JOINTS_NUM)
			THROW_ERROR("No such joint!");
		return jointNames[joint].c_str();
		CATCH;
	}

	const char * toString(int joint) 
	{
		TRY(QRIO_JointInfo::toString);
		if (joint < 0 || joint >= JOINTS_NUM)
			THROW_ERROR("No such joint!");
		return jointStrings[joint].c_str();
		CATCH;
	}

	const char * toDEFString(int joint) 
	{
		TRY(QRIO_JointInfo::toString);
		if (joint < 0 || joint > JOINTS_NUM)
			THROW_ERROR("No such joint!");
		return jointDEFs[joint].c_str();
		CATCH;
	}

	float getMinLimit(int joint) 
	{
		TRY(QRIO_JointInfo::getMinLimit);
		if (joint < 0 || joint >= JOINTS_NUM)
			THROW_ERROR("No such joint!");
		return jointMinLimits[joint];
		CATCH;
	}

	float getMaxLimit(int joint) 
	{
		TRY(QRIO_JointInfo::getMaxLimit);
		if (joint < 0 || joint >= JOINTS_NUM)
			THROW_ERROR("No such joint!");
		return jointMaxLimits[joint];
		CATCH;
	}

};

BIPED_END_NAMESPACE;

#endif	// __QRIO_JOINTS_HPP__
