/* -*- c++ -*- */
/****************************************************************
 *                                                              *
 *  Project name:      Biped Locomotion                         *
 *                                                              *
 *  Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)           *
 ****************************************************************
 * $Id: interp_comparison.cpp,v 1.1 2004/08/15 19:51:48 s0343981 Exp $
 ****************************************************************/
/**
 * \file interp_comparison.cpp 
 * \brief A simple program to compare the different interpolations.
 */

// C++ includes:
#include <iostream>
#include <sstream>
#include <iomanip>
#include <string>
#include <vector>
#include <newmat/newmat.h>
// BIPED includes:
#include <biped_exception.hpp>
#include <biped_gnuplot.hpp>
#include <biped_interp.hpp>
#include <biped_logging.hpp>
#include <biped_trajectory.hpp>

using std::cout;
using std::endl;
using std::string;

#define SEQ_LEN 7
#define INTERPS 6

static double Q[SEQ_LEN] = { -2.0, -1.6, -1.6, -0.8, -0.2, 0.0, -0.6 };

RowVector getTimeVector(double startTime, double interval, int size)
{
	TRY(getTimeVector);
	RowVector t(size);
	double time = 0.0;
	for (int i = 0; i < size; i++) {
		t(i + 1) = time;
		time += interval;
	}
	t += startTime;
	t.Release();
	return t;
	CATCH;
}

void getTimeSTLVector(vector< double > & t, double startTime, 
					  double interval, int size)
{
	TRY(getTimeSTLVector);
	t.clear();
	double time = startTime;
	for (int i = 0; i < size; i++) {
		t.push_back(time);
		time += interval;
	}
	CATCH;
}

Trajectory * getTrajObject(int i)
{
	TRY(getTrajObject);
	switch (i) {
		case 0: return new SimpleTrajectory();
		case 1: return new InterpTrajectory< AkimaCubicSpline >();
		case 2: return new InterpTrajectory< BezierCubicSpline >();
		case 3: return new InterpTrajectory< CardinalCubicSpline >();
		case 4: return new InterpTrajectory< FritschCarlsonCubicSpline >();
		case 5: return new InterpTrajectory< PolynomialSpline >();
	}
	return NULL;
	CATCH;
}

string getInterpName(int i)
{
	TRY(getInterpName);
	switch (i) {
		case 0: return "Simple";
		case 1: return "Akima";
		case 2: return "Bezier";
		case 3: return "Cardinal";
		case 4: return "Fritsch";
		case 5: return "Polynomial";
	}
	return "";
	CATCH;
}

int main()
{
	TRY(main);
    cout << endl << "*** Starting interpolations comparison test ***" << endl;
    
	Matrix pts(2, SEQ_LEN);
	pts.Row(1) = getTimeVector(0.032 * 32, 0.512, SEQ_LEN);
	for (int j = 1; j <= pts.Ncols(); j++) {
		pts(2, j) = Q[j - 1];
	}

	GNUPlot gplot1 = GNUPlot("points");
    cout << "*** Plotting the points ***" << endl;
	vector< double > t;
	getTimeSTLVector(t, 0.032 * 32, 0.512, SEQ_LEN);
	vector< double > q;
	for (int c = 1; c <= pts.Ncols(); c++) {
		q.push_back(pts(2, c));
	}
	gplot1.plotXYData(t, q, "via points");
	sleep(1);

    cout << "*** Plotting the interpolations ***" << endl;
	gplot1.setStyle("lines");
	vector< double > tt;
	getTimeSTLVector(tt, 0.032 * 32, 0.032, (SEQ_LEN - 1) * 16 + 1);

	Trajectory * traj;
	vector< vector< double > > qq(INTERPS);
	for (uint i = 0; i < INTERPS; i++) {
		traj = getTrajObject(i);
		traj->setPoints(pts);
		for (uint k = 0; k < tt.size(); k++) {
			ColumnVector tmp(1);
			traj->interpolate(tt[k], tmp);
			qq[i].push_back(tmp(1));
		}
		delete traj;
	}
	for (int i = 0; i < INTERPS; i++) {
		gplot1.plotXYData(tt, qq[i], getInterpName(i));
		sleep(1);
	}

    cout << endl << "*** End of interpolations comparison test ***" << endl;
	cout << "Type any key to continue ...";
	char c;
	cin >> c;
	CATCH_FINAL;
    return 0;
}
