/* -*- c++ -*- */
/****************************************************************
 *                                                              *
 *  Project name:      Biped Locomotion                         *
 *                                                              *
 *  Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)           *
 ****************************************************************
 * $Id: interp_comparison2.cpp,v 1.1 2004/08/29 11:09:48 s0343981 Exp $
 ****************************************************************/
/**
 * \file interp_comparison.cpp 
 * \brief A simple program to compare the different interpolations.
 */

// C++ includes:
#include <iostream>
#include <sstream>
#include <iomanip>
#include <string>
#include <vector>
#include <newmat/newmat.h>
// BIPED includes:
#include <biped_exception.hpp>
#include <biped_gnuplot.hpp>
#include <biped_interp.hpp>
#include <biped_logging.hpp>
#include <biped_trajectory.hpp>

using std::cout;
using std::endl;
using std::string;

#define SEQ_LEN 7
#define INTERPS 3

static double Q[SEQ_LEN] = {0, 0, 0, 0, 0, 0, 0};

RowVector getTimeVector(double startTime, double interval, int size)
{
	TRY(getTimeVector);
	RowVector t(size);
	double time = 0.0;
	for (int i = 0; i < size; i++) {
		t(i + 1) = time;
		time += interval;
	}
	t += startTime;
	t.Release();
	return t;
	CATCH;
}

void getTimeSTLVector(vector< double > & t, double startTime, 
					  double interval, int size)
{
	TRY(getTimeSTLVector);
	t.clear();
	double time = startTime;
	for (int i = 0; i < size; i++) {
		t.push_back(time);
		time += interval;
	}
	CATCH;
}

Trajectory * getTrajObject(int i)
{
	TRY(getTrajObject);
	switch (i) {
		case 0: return new SimpleTrajectory();
		case 1: return new InterpTrajectory< BezierCubicSpline >();
		case 2: return new InterpTrajectory< FritschCarlsonCubicSpline >();
		case 3: return new InterpTrajectory< AkimaCubicSpline >();
		case 4: return new InterpTrajectory< CardinalCubicSpline >();
		case 5: return new InterpTrajectory< PolynomialSpline >();
	}
	return NULL;
	CATCH;
}

string getInterpName(int i)
{
	TRY(getInterpName);
	switch (i) {
		case 0: return "Simple";
		case 1: return "Bezier";
		case 2: return "Fritsch";
		case 3: return "Akima";
		case 4: return "Cardinal";
		case 5: return "Polynomial";
	}
	return "";
	CATCH;
}

int main()
{
	TRY(main);
    cout << endl << "*** Starting interpolations comparison test ***" << endl;
    
	Matrix pts(2, SEQ_LEN);
	pts.Row(1) = getTimeVector(14.336, 0.512, SEQ_LEN);
	for (int j = 1; j <= pts.Ncols(); j++) {
		pts(2, j) = Q[j - 1];
	}

	GNUPlot gplot1 = GNUPlot("points");
    cout << "*** Plotting the points ***" << endl;
	vector< double > t;
	getTimeSTLVector(t, 14.336, 0.512, SEQ_LEN);
	vector< double > q;
	for (int c = 1; c <= pts.Ncols(); c++) {
		q.push_back(pts(2, c));
	}
	gplot1.plotXYData(t, q, "via points");
	sleep(1);

	gplot1.setStyle("lines");
	vector< double > tt;
	getTimeSTLVector(tt, 14.336, 0.032, (SEQ_LEN - 1) * 16 + 1);

	RowVector tc(17);
	tc << 14.528 << 14.56 << 14.592 << 14.624 << 14.656 << 14.688 << 14.72 << 14.752 
	   << 14.784 << 14.816 << 14.848 << 14.88 << 14.912 << 14.944 << 14.976 << 15.008 << 15.04;
	
	Matrix newQ(1, 17);
	newQ << 0 << -0.0034937 << -0.0130729 << -0.0273852 << -0.045079 << -0.0648007 << -0.0851993 
		 << -0.104921 << -0.122615 << -0.136927 << -0.146506 << -0.15 << -0.0558849 << -0.0987642
		 << 0.0676953 << 0.106648 << -0.15;

	Trajectory * traj;
	vector< vector< double > > qq(INTERPS);
	vector< vector< double > > qqc(INTERPS);
	for (uint i = 0; i < INTERPS; i++) {
		traj = getTrajObject(i);
		traj->setPoints(pts);
		for (uint k = 0; k < tt.size(); k++) {
			ColumnVector tmp(1);
			traj->interpolate(tt[k], tmp);
			qq[i].push_back(tmp(1));
		}
		traj->adjust(tc, newQ);
		for (uint k = 0; k < tt.size(); k++) {
			ColumnVector tmp(1);
			traj->interpolate(tt[k], tmp);
			qqc[i].push_back(tmp(1));
		}
		delete traj;
	}
	cout << "*** Plotting the interpolations ***" << endl;
	for (int i = 0; i < INTERPS; i++) {
		gplot1.plotXYData(tt, qq[i], getInterpName(i));
		sleep(1);
	}
	cout << "*** Plotting the adjusted interpolations ***" << endl;
	for (int i = 0; i < INTERPS; i++) {
		gplot1.plotXYData(tt, qqc[i], getInterpName(i) + "_c");
		sleep(1);
	}

    cout << endl << "*** End of interpolations comparison test ***" << endl;
	cout << "Type any key to continue ...";
	char c;
	cin >> c;
	CATCH_FINAL;
    return 0;
}
