/* -*- c++ -*- */
/****************************************************************
 *                                                              *
 *  Project name:      Biped Locomotion                         *
 *                                                              *
 *  Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)           *
 ****************************************************************
 * $Id: spline_test.cpp,v 1.2 2004/08/15 19:47:39 s0343981 Exp $ 
 ****************************************************************/
/**
 * \file spline_test.cpp 
 * \brief A simple program to test the various splines.
 */

// C++ includes:
#include <iostream>
#include <sstream>
#include <iomanip>
#include <string>
#include <vector>
// BIPED includes:
#include <biped_exception.hpp>
#include <biped_gnuplot.hpp>
#include <biped_interp.hpp>
#include <biped_logging.hpp>

using std::cout;
using std::endl;
using std::string;

#define SEQ_LEN 7
#define SPLINES_NUM 5

static double vals[SEQ_LEN] = { -2.0, -2.0, -1.7, -1.4, -0.8, -0.4, 0.0 }; 

void initTimeVector(vector< double > & t, double interval, int size)
{
	TRY(initTimeVector);
	t.clear();
	double time = 0.0;
	for (int i = 0; i < size; i++) {
		t.push_back(time);
		time += interval;
	}
	CATCH;
}

void setVector(vector< double > & q, double seq[])
{
	TRY(setVector);
	q.clear();
	for (int i = 0; i < SEQ_LEN; i++) {
		q.push_back(seq[i]);
	}
	CATCH;
}

void getInterpVals(vector< double > & qq, const vector< double > & tt,
				   const SplineBase * spline)
{
	TRY(setVector);
	qq.clear();
	for (uint i = 0; i < tt.size(); i++) {
		qq.push_back(spline->getY(tt[i]));
	}
	CATCH;
}

RowVector getRowVector(const vector< double > & v)
{
	TRY(getRowVector);
	RowVector res(v.size());
	for (int i = 1; i <= res.Ncols(); i++) {
		res(i) = v[i - 1];
	}
	res.Release();
	return res;
	CATCH;
}

string getSplineType(int i)
{
	TRY(getSplineType);
	switch (i) {
		case SplineBase::AKIMA:      return "AKIMA";
		case SplineBase::BEZIER:     return "BEZIER";
		case SplineBase::CARDINAL:   return "CARDINAL";
		case SplineBase::FRITSCH:    return "FRITSCH";
		case SplineBase::POLYNOMIAL: return "POLYNOMIAL";
		default:
			THROW_FATAL("Unknown spline type (" << i << ")!"); 
	}
	CATCH;
}

int main()
{
	TRY(main);
    cout << endl << "*** Starting splines test ***" << endl;
    
	vector< double > t;
	vector< double > q;
	initTimeVector(t, 0.512, SEQ_LEN);

	GNUPlot gplot = GNUPlot("points");
    cout << "*** Plotting the points ***" << endl;
	setVector(q, vals);
	gplot.plotXYData(t, q, "via points");
	sleep(1);

    cout << "*** Plotting the interpolations ***" << endl;

	gplot.setStyle("lines");

	vector< double > tt;
	vector< double > qq;
	initTimeVector(tt, 0.032, SEQ_LEN * 16);
	RowVector trv = getRowVector(t), qrv = getRowVector(q);
	
	SplineBase * spline = NULL;
	for (int i = 0; i < SPLINES_NUM; i++) {
		spline = SplineBase::createNewSpline((SplineBase::SplineType) i);
		spline->interpolate(trv, qrv);
		getInterpVals(qq, tt, spline);
		gplot.plotXYData(tt, qq, getSplineType(i));
		sleep(1);
		delete spline;
	}

    cout << endl << "*** End of splines test ***" << endl;
	cout << "Type any key to continue ...";
	char c;
	cin >> c;
	CATCH_FINAL;
    return 0;
}
