/* -*- c++ -*- */
/****************************************************************
 *                                                              *
 *  Project name:      Biped Locomotion                         *
 *                                                              *
 *  Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)           *
 ****************************************************************
 * $Id: traj_adjust_test.cpp,v 1.4 2004/08/29 11:09:49 s0343981 Exp $
 ****************************************************************/
/**
 * \file traj_adjust_test.cpp 
 * \brief A simple program to test the simple trajectory adjustment.
 */

// C++ includes:
#include <iostream>
#include <sstream>
#include <iomanip>
#include <string>
#include <vector>
#include <newmat/newmat.h>
// BIPED includes:
#include <biped_exception.hpp>
#include <biped_gnuplot.hpp>
#include <biped_logging.hpp>
#include <biped_trajectory.hpp>

using std::cout;
using std::endl;
using std::string;

#define SEQ_LEN 7
#define JOINTS 3

static double Q[JOINTS][SEQ_LEN] = { 
	{-2.0, -2.0, -1.7, -1.4, -0.8, -0.4, 0.0},
	{-1.6, -1.6, -1.6, -0.9, -0.5, -0.3, 0.0},
	{-0.7, -0.7, -0.5, -0.4, -0.3, -0.1, 0.0}
};

RowVector getTimeVector(double startTime, double interval, int size)
{
	TRY(getTimeVector);
	RowVector t(size);
	double time = 0.0;
	for (int i = 0; i < size; i++) {
		t(i + 1) = time;
		time += interval;
	}
	t += startTime;
	t.Release();
	return t;
	CATCH;
}

void getTimeSTLVector(vector< double > & t, double startTime, 
					  double interval, int size)
{
	TRY(getTimeSTLVector);
	t.clear();
	double time = startTime;
	for (int i = 0; i < size; i++) {
		t.push_back(time);
		time += interval;
	}
	CATCH;
}

int main()
{
	TRY(main);
    cout << endl << "*** Starting trajectory adjustment test ***" << endl;
    
	Matrix pts(JOINTS + 1, SEQ_LEN);
	pts.Row(1) = getTimeVector(7.808, 0.512, SEQ_LEN);
	for (int i = 2; i <= pts.Nrows(); i++) {
		for (int j = 1; j <= pts.Ncols(); j++) {
			pts(i, j) = Q[i - 2][j - 1];
		}
	}
	SimpleTrajectory traj;
// 	InterpTrajectory< FritschCarlsonCubicSpline > traj;
	traj.setPoints(pts);
	
	GNUPlot gplot = GNUPlot("points");
    cout << "*** Plotting the points ***" << endl;
	vector< double > t;
	getTimeSTLVector(t, 7.808, 0.512, SEQ_LEN);
	for (int i = 0; i < JOINTS; i++) {
		vector< double > q;
		for (int c = 1; c <= pts.Ncols(); c++) {
			q.push_back(pts(i + 2, c));
		}
		ostringstream jointStr;
		jointStr << "via points " << i;
		gplot.plotXYData(t, q, jointStr.str());
		sleep(1);
	}

    cout << "*** Plotting the interpolations ***" << endl;
	gplot.setStyle("lines");

	vector< double > tt;
	getTimeSTLVector(tt, 7.808, 0.032, (SEQ_LEN - 1) * 16 + 1);
	vector< vector< double > > qq(JOINTS);
	for (uint i = 0; i < tt.size(); i++) {
		ColumnVector tmp(JOINTS);
		traj.interpolate(tt[i], tmp);
		for (int j = 1; j <= tmp.Nrows(); j++) {
			qq[j - 1].push_back(tmp(j));
		}
	}
	for (int i = 0; i < JOINTS; i++) {
		ostringstream jointStr;
		jointStr << "interp " << i;
		gplot.plotXYData(tt, qq[i], jointStr.str());
		sleep(1);
	}

    cout << "*** Making corrections ***" << endl;
	RowVector tc(6);
	tc << 8.320 << 9.088 << 9.120 << 9.152 << 9.184 << 9.216; 
	Matrix newQ(JOINTS, 6);
	newQ << -0.1 << -0.3 << -0.3 << -0.1 << -0.3 << -0.1
		 << -0.1 << 0.2  << -0.1 << 0.2  << -0.1 << 0.2 
		 << 0.5  << -0.1 << 0.5  << -0.1 << 0.5  << -0.1;
	traj.adjust(tc, newQ);

	vector< vector< double > > qqc(JOINTS);
	for (uint i = 0; i < tt.size(); i++) {
		ColumnVector tmp(JOINTS);
		traj.interpolate(tt[i], tmp);
		for (int j = 1; j <= tmp.Nrows(); j++) {
			qqc[j - 1].push_back(tmp(j));
		}
	}
	for (int i = 0; i < JOINTS; i++) {
		ostringstream jointStr;
		jointStr << "corrected interp " << i;
		gplot.plotXYData(tt, qqc[i], jointStr.str());
		sleep(1);
	}

    cout << endl << "*** End of trajectory adjustment test ***" << endl;
	cout << "Type any key to continue ...";
	char c;
	cin >> c;
	CATCH_FINAL;
    return 0;
}
