/* -*- c++ -*- */
/****************************************************************
 *                                                              *
 *  Project name:      Biped Locomotion                         *
 *                                                              *
 *  Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)           *
 ****************************************************************
 * $Id: traj_adjust_test2.cpp,v 1.1 2004/08/29 11:09:49 s0343981 Exp $
 ****************************************************************/
/**
 * \file traj_adjust_test.cpp 
 * \brief A simple program to test the simple trajectory adjustment.
 */

// C++ includes:
#include <iostream>
#include <sstream>
#include <iomanip>
#include <string>
#include <vector>
#include <newmat/newmat.h>
// BIPED includes:
#include <biped_exception.hpp>
#include <biped_gnuplot.hpp>
#include <biped_logging.hpp>
#include <biped_trajectory.hpp>

using std::cout;
using std::endl;
using std::string;

#define SEQ_LEN 7
#define JOINTS 1

static double Q[JOINTS][SEQ_LEN] = { 
	{0, 0, 0, 0, 0, 0, 0}
};

RowVector getTimeVector(double startTime, double interval, int size)
{
	TRY(getTimeVector);
	RowVector t(size);
	double time = 0.0;
	for (int i = 0; i < size; i++) {
		t(i + 1) = time;
		time += interval;
	}
	t += startTime;
	t.Release();
	return t;
	CATCH;
}

void getTimeSTLVector(vector< double > & t, double startTime, 
					  double interval, int size)
{
	TRY(getTimeSTLVector);
	t.clear();
	double time = startTime;
	for (int i = 0; i < size; i++) {
		t.push_back(time);
		time += interval;
	}
	CATCH;
}

Trajectory * getTrajObject(int i)
{
	TRY(getTrajObject);
	switch (i) {
		case 0: return new SimpleTrajectory();
		case 1: return new InterpTrajectory< AkimaCubicSpline >();
		case 2: return new InterpTrajectory< BezierCubicSpline >();
		case 3: return new InterpTrajectory< CardinalCubicSpline >();
		case 4: return new InterpTrajectory< FritschCarlsonCubicSpline >();
		case 5: return new InterpTrajectory< PolynomialSpline >();
	}
	return NULL;
	CATCH;
}

string getInterpName(int i)
{
	TRY(getInterpName);
	switch (i) {
		case 0: return "Simple";
		case 1: return "Akima";
		case 2: return "Bezier";
		case 3: return "Cardinal";
		case 4: return "Fritsch";
		case 5: return "Polynomial";
	}
	return "";
	CATCH;
}

int main()
{
	TRY(main);
    cout << endl << "*** Starting trajectory adjustment test ***" << endl;
    
	Matrix pts(JOINTS + 1, SEQ_LEN);
	pts.Row(1) = getTimeVector(14.336, 0.512, SEQ_LEN);
	for (int i = 2; i <= pts.Nrows(); i++) {
		for (int j = 1; j <= pts.Ncols(); j++) {
			pts(i, j) = Q[i - 2][j - 1];
		}
	}
// 	SimpleTrajectory traj;
	InterpTrajectory< FritschCarlsonCubicSpline > traj;
	traj.setPoints(pts);
	
	GNUPlot gplot = GNUPlot("points");
    cout << "*** Plotting the points ***" << endl;
	vector< double > t;
	getTimeSTLVector(t, 14.336, 0.512, SEQ_LEN);
	for (int i = 0; i < JOINTS; i++) {
		vector< double > q;
		for (int c = 1; c <= pts.Ncols(); c++) {
			q.push_back(pts(i + 2, c));
		}
		ostringstream jointStr;
		jointStr << "via points " << i;
		gplot.plotXYData(t, q, jointStr.str());
		sleep(1);
	}

    cout << "*** Plotting the interpolations ***" << endl;
	gplot.setStyle("lines");

	vector< double > tt;
	getTimeSTLVector(tt, 14.336, 0.032, (SEQ_LEN - 1) * 16 + 1);
	vector< vector< double > > qq(JOINTS);
	for (uint i = 0; i < tt.size(); i++) {
		ColumnVector tmp(JOINTS);
		traj.interpolate(tt[i], tmp);
		for (int j = 1; j <= tmp.Nrows(); j++) {
			qq[j - 1].push_back(tmp(j));
		}
	}
	for (int i = 0; i < JOINTS; i++) {
		ostringstream jointStr;
		jointStr << "interp " << i;
		gplot.plotXYData(tt, qq[i], jointStr.str());
		sleep(1);
	}

    cout << "*** Making corrections ***" << endl;
	RowVector tc(17);
	tc << 14.528 << 14.56 << 14.592 << 14.624 << 14.656 << 14.688 << 14.72 << 14.752 
	   << 14.784 << 14.816 << 14.848 << 14.88 << 14.912 << 14.944 << 14.976 << 15.008 << 15.04;

	Matrix newQ(JOINTS, 17);
	newQ << 0 << -0.0034937 << -0.0130729 << -0.0273852 << -0.045079 << -0.0648007 << -0.0851993 
		 << -0.104921 << -0.122615 << -0.136927 << -0.146506 << -0.15 << -0.0558849 << -0.0987642
		 << 0.0676953 << 0.106648 << -0.15;
	traj.adjust(tc, newQ);

	vector< vector< double > > qqc(JOINTS);
	for (uint i = 0; i < tt.size(); i++) {
		ColumnVector tmp(JOINTS);
		traj.interpolate(tt[i], tmp);
		for (int j = 1; j <= tmp.Nrows(); j++) {
			qqc[j - 1].push_back(tmp(j));
		}
	}
	for (int i = 0; i < JOINTS; i++) {
		ostringstream jointStr;
		jointStr << "corrected interp " << i;
		gplot.plotXYData(tt, qqc[i], jointStr.str());
		sleep(1);
	}

    cout << endl << "*** End of trajectory adjustment test ***" << endl;
	cout << "Type any key to continue ...";
	char c;
	cin >> c;
	CATCH_FINAL;
    return 0;
}
