/* -*- c++ -*- */
/****************************************************************
 *                                                              *
 *  Project name:      Biped Locomotion                         *
 *                                                              *
 *  Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)           *
 ****************************************************************
 * $Id: hoap2.cpp,v 1.5 2004/08/06 18:02:59 s0343981 Exp $ 
 ****************************************************************/
/**
 * \file hoap2.cpp 
 */

#include <device/robot.h>
#include <iostream>
#include <biped_exception.hpp>
#include <biped_logging.hpp>
#include <hoap2/hoap2_joints.hpp>
#include "hoap2_controller.hpp"

using std::cout;
using std::endl;
using namespace HOAP2;

static int range_sensor;
static int left_touch;
static int right_touch;
static int gps;
static DeviceTag joints[JOINTS_NUM];
DeviceTag camera, emitter;
static HOAP2_Controller * controller = NULL;
static HOAP2_Controller::ControlType controlType = HOAP2_Controller::LEARNING;

// -----------------------------------------------------------------------------
//  Function: hoap2_reset
/// \brief The global implementation for a resetting hoap-2 controller.
/// Details: Here the device handles are collected.
// -----------------------------------------------------------------------------
static void hoap2_reset(void) 
{
	TRY(hoap2_reset);
	cout << "Reseting hoap-2 controller" << endl;
	for (int i = 0; i < JOINTS_NUM; i++) {
		joints[i] = robot_get_device(HOAP2_JointToStr().toString(i));
	}
	for (int i = 0; i < JOINTS_NUM; i++) {
		if (joints[i] == 0)
			cout << "Joint " << i << " is missing" << endl;
	}
	
	// Get the rest of the devices:
	camera = robot_get_device("camera");
	range_sensor = robot_get_device("distance");
	left_touch = robot_get_device("left touch");
	right_touch = robot_get_device("right touch");
	gps = robot_get_device("gps");
	emitter = robot_get_device("emitter");
	if (emitter == 0) 
		cout << "missing emitter" << endl;

	controller = new HOAP2_Controller(controlType, JOINTS_NUM, left_touch, 
									 right_touch, range_sensor, gps, joints, 
									 camera, emitter);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: hoap2_exit
/// \brief The global implementation for a exiting the hoap-2 controller.
/// Details: Here the device handles are collected.
// -----------------------------------------------------------------------------
static void hoap2_exit(void) 
{
	TRY(hoap2_exit);
	cout << "Exiting hoap-2 controller" << endl;
	delete controller;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: determine_run_type
/// \brief Determines which type of control to use while running the robot.
// -----------------------------------------------------------------------------
static void determine_run_type()
{
	TRY(determine_run_type);
	char * envControlType = getenv("HOAP2_RUN");
	PRINT_VALUE(envControlType);
    if (envControlType == 0 || strcmp(envControlType, "learn") == 0) {
		controlType = HOAP2_Controller::LEARNING;
	} else if (strcmp(envControlType, "compensate") == 0) {
		controlType = HOAP2_Controller::COMPENSATE;
	} else if (strcmp(envControlType, "actual") == 0) {
		controlType = HOAP2_Controller::ACTUAL;
	} else {
		THROW_FATAL("Unknown run type for Hoap-2 robot!");
	}
	cout << "\nHOAP2 ";
	switch (controlType) {
		case HOAP2_Controller::LEARNING:   cout << "learning";     break;
		case HOAP2_Controller::COMPENSATE: cout << "compensation"; break;
		case HOAP2_Controller::ACTUAL:     cout << "actual";       break;
	}
	cout << " control!" << endl;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: main
/// \brief The main program of hoap-2 robot controller.
// -----------------------------------------------------------------------------
int main() 
{
	TRY(main);
	// Determine the control type to run the robot:
	determine_run_type();
	
	// Bind the robot_live and robot_die functions:
	robot_live(hoap2_reset);
	robot_die(hoap2_exit);

	controller->enableDevices();
	controller->run();

	hoap2_exit();
	CATCH_FINAL;
	return 0;
}
