/* -*- c++ -*- */
/****************************************************************
 *                                                              *
 *  Project name:      Biped Locomotion                         *
 *                                                              *
 *  Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)           *
 ****************************************************************
 * $Id: hoap2_controller.cpp,v 1.4 2004/08/06 18:02:59 s0343981 Exp $ 
 ****************************************************************/
/**
 * \file hoap2_controller.cpp 
 * \brief class HOAP2_Controller methods' implementations.
 */

// CGAL includes:
#include <CGAL/convex_hull_2.h>
// Webots includes:
#include <device/robot.h>
#include <device/servo.h>
#include <device/emitter.h>
#include <device/camera.h>
#include <device/distance_sensor.h>
#include <device/gps.h>
#include <device/touch_sensor.h>
// ROBOOP includes:
#include <newmat/newmat.h>
#include <newmat/newmatio.h>
// C++ includes:
#include <iostream>
#include <sstream>
#include <iomanip>
#include <string>
// BIPED includes:
#include <biped_exception.hpp>
#include <biped_gene.hpp>
#include <biped_logging.hpp>
#include <biped_matrix_io.hpp>
#include <biped_sem.hpp>
#include <biped_shmem.hpp>
#include <biped_util.hpp>
// HOAP-2 includes:
#include <hoap2/hoap2_joints.hpp>
#include <hoap2/hoap2_shmem.hpp>
#include "hoap2_controller.hpp"

using std::cout;
using std::endl;
using std::string;
using namespace HOAP2;

// Shared memory stuff:
static SharedMemory shmDynamics; // dynamics shared memory manager
static SharedMemory shmFootGround; // foot-ground shared memory manager
static DynamicsInfo * dynShmPtr; // ptr to dynamics struct in shared memory
static FootGroundInfo * footShmPtr; // ptr to foot-ground struct in shared memory
static Semaphore semDynamics;	// dynamics shared memory semaphore
static Semaphore semFootGround;	// foot-ground shared memory semaphore

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::HOAP2_Controller
/// \brief HOAP2_Controller's c'tor.
/// \param ctype The control type.
/// \param jointsNum No. of joints existing for this robot.
/// \param lt Handle to the left touch sensor device.
/// \param rt Handle to the right touch sensor device.
/// \param range Handle to the range sensor device.
/// \param gps Handle to the GPS sensor device.
/// \param joints Array of handles to all servo joint devices.
/// \param camera Handle to the camera device.
/// \param emitter Handle to the emitter device.
// -----------------------------------------------------------------------------
HOAP2_Controller::HOAP2_Controller(ControlType ctype, uint jointsNum, int lt, 
								   int rt, int range, int gps, DeviceTag * joints,
								   DeviceTag camera, DeviceTag emitter) :
	Controller(jointsNum, lt, rt, range, gps, joints, camera, emitter) , 
	ctype_(ctype) , contactPts_() , links_(NUM_RIGID_BODIES) , 
	currMotionTraj_() , currMotionType_(FIRST_STEP) , qVecs_() , jacobians_() ,
	dqCorrectMap_()
{
	TRY(HOAP2_Controller::HOAP2_Controller);
	initialize();
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::run
/// \brief Perform the endless loop of the robot control.
// -----------------------------------------------------------------------------
void HOAP2_Controller::run()
{
	TRY(HOAP2_Controller::run);
	robotSimulationStep(2);		// wait a bit ...
	initialStep();
	switch (ctype_) {
		case LEARNING:   learnRun();      break;
		case COMPENSATE: compensateRun(); break;
		case ACTUAL:     actualRun();     break;
		default:
			THROW_FATAL("HOAP2: Unknown control type!");
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::learnRun
/// \brief Perform one run of learning and then do nothing to let user stop
/// the simulation.
// -----------------------------------------------------------------------------
void HOAP2_Controller::learnRun()
{
	TRY(HOAP2_Controller::learnRun);
	double t = 0;				// current time [sec]
	uint zmpOutNum = 0;
	for (uint i = 0; i < WALK_CYCLES_NUM; i++) {
		cout << "Walking cycle " << i << endl;
		bool cycleDone = false, isZMPout = false;
		while (!cycleDone) {
			if (currMotionType_ == WALK_RIGHT || currMotionType_ == WALK_LEFT) {
// 				jiggleAllStep(t);
				jiggleIndependentStep(t);
			}
			cycleDone = regularStep(t, isZMPout);
			if (isZMPout && (currMotionType_ == WALK_RIGHT 
							 || currMotionType_ == WALK_LEFT)) {
				zmpOutNum++;
				cout << "ZMP(t=" << t << ") is outside!" << endl;
			}
		}
	}
	PRINT_VALUE(zmpOutNum);
	cout << "HOAP2_Controller: Finished learning run!" << endl;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::compensateRun
/// \brief Perform multiple runs of compensation until the ZMP is no longer 
/// outside the support polygon.
// -----------------------------------------------------------------------------
void HOAP2_Controller::compensateRun()
{
	TRY(HOAP2_Controller::compensateRun);
	for (;;) {
		double t = 0;				// current time [sec]
		uint zmpOutNum = 0;
		for (uint i = 0; i < WALK_CYCLES_NUM; i++) {
			cout << "Walking cycle " << i << endl;
			bool cycleDone = false, isZMPout = false;
			while (!cycleDone) {
				if (currMotionType_ == WALK_RIGHT || 
					currMotionType_ == WALK_LEFT) {
// 					jiggleAllStep(t);
					jiggleIndependentStep(t);
					cycleDone = compensateStep(t, isZMPout);
					if (isZMPout) {
						zmpOutNum++;
						cout << "ZMP(t=" << t << ") is outside!" << endl;
					}
				} else {
					cycleDone = regularStep(t, isZMPout);
				}
			}
		}
		PRINT_VALUE(zmpOutNum);
		if (zmpOutNum == 0)
			break;
		// Reset all joints to zero:
		cout << "HOAP2_Controller: Resetting all joints to zero!" << endl;
		ColumnVector q(jointsNum_);
		q = 0.0;
		setJointAngles(q);
		robotSimulationStep(10);	// wait a bit

		// Request supervisor to reset the robot's position:
		float msg[7] = { 2, 0, 0, 0, 0, 0, 0 };
		uint msgSize = sizeof(float) * 7;
		emitterSend(msg, msgSize);
		robotSimulationStep(10); // wait a bit
	}
	cout << "HOAP2_Controller: Finished compensation run!" << endl;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::actualRun
/// \brief Perform one actual run of corrected trajectories and then do nothing
/// to let user stop the simulation.
// -----------------------------------------------------------------------------
void HOAP2_Controller::actualRun()
{
	TRY(HOAP2_Controller::actualRun);
	double t = 0;				// current time [sec]
	for (uint i = 0; i < WALK_CYCLES_NUM; i++) {
		cout << "Walking cycle " << i << endl;
		bool cycleDone = false, isZMPout = false;
		while (!cycleDone) {
			if (currMotionType_ == WALK_RIGHT 
				|| currMotionType_ == WALK_LEFT) {
				cycleDone = actualStep(t);
	} else {
				cycleDone = regularStep(t, isZMPout);
			}
	}
	}
	cout << "HOAP2_Controller: Finished learning run!" << endl;
	CATCH;
}
	
// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::initialStep
/// \brief Performs the initial step of the robot's control.
// -----------------------------------------------------------------------------
void HOAP2_Controller::initialStep() 
{
	TRY(HOAP2_Controller::initialStep);
	robotSimulationStep(1); // run one step

	// Read the constant dynamics information. No need to perform this at
	// later calls (the values remain the same):
	readLinkConstDynamics();
	
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::regularStep
/// \brief Performs one regular motion step of the robot's control.
// -----------------------------------------------------------------------------
bool HOAP2_Controller::regularStep(double & t, bool & isZMPOut)
{
	TRY(HOAP2_Controller::regularStep);
	// Read the contact points from shared memory:
	readContactPoints();
	// Get Support Polgon:
	Point3Vec supportPolygon;
	getSupportPolygon(contactPts_, supportPolygon);

	// Get the robot's current CoM and ZMP:
	Point_3 com, zmp;
	getRobotCurrCOMAndZMP(com, zmp);

	// Check if ZMP is inside the support polygon:
	CGAL::Bounded_side boundSide = getBoundedSide(supportPolygon, zmp);
	isZMPOut = (boundSide == CGAL::ON_UNBOUNDED_SIDE) ? true : false;
 
    // Get the current joint angles vector to set:
	ColumnVector q(jointsNum_); // current column vector of joint angles
	bool res = xmlWalkStep(t, q);
	setJointAngles(q);

	// Perform the robot step:
	t += robotSimulationStep(1);
	return res;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::jiggleAllStep
/// \brief Performs one jiggle step of the redundant degrees (all together) in 
/// robot's control.
// -----------------------------------------------------------------------------
void HOAP2_Controller::jiggleAllStep(double t) 
{
	TRY(HOAP2_Controller::jiggleAllStep);
	ColumnVector q = getJointAngles(); // get current joint angles

	// Compute the robot's current ZMP before jiggle:
	Point_3 zmpBefore, dummy;
	getRobotCurrCOMAndZMP(dummy, zmpBefore);
// 	PRINT_VALUE(zmpBefore);

	// Get the current offset in redundant joints:
	ColumnVector dq(jointsNum_);
	vector< int > redunds;
	getCurrRedundantJointsOffsets(t, dq, redunds);
// 	PRINT_VALUE(dq);

	ColumnVector qNew = q + dq;
	setJointAngles(qNew);
	robotSimulationStep(1);		// run one step

	// Compute the robot's current ZMP after jiggle:
	Point_3 zmpAfter;
	getRobotCurrCOMAndZMP(dummy, zmpAfter);
// 	PRINT_VALUE(zmpAfter);
	
	// Restore original posture at the beginning of this method:
	setJointAngles(q);
	robotSimulationStep(1);		// run one step
	
	// Calculate the ZMP offset due to jiggle:
	ColumnVector dZMP(2);
	dZMP << (zmpAfter.x() - zmpBefore.x()) << (zmpAfter.z() - zmpBefore.z());
// 	PRINT_VALUE(dZMP);

	// Calculate the current jacobian, store it along with joint angles vector:
	Matrix J = computeJacobian(dZMP, dq, redunds);
	qVecs_.push_back(q);
	jacobians_.push_back(J);
// 	PRINT_VALUE(J);	
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::jiggleIndependentStep
/// \brief Performs one jiggle step of the redundant degrees (each one 
/// independently) in robot's control.
// -----------------------------------------------------------------------------
void HOAP2_Controller::jiggleIndependentStep(double t) 
{
	TRY(HOAP2_Controller::jiggleIndependentStep);
	ColumnVector q = getJointAngles(); // get current joint angles

	// Get the current offset in redundant joints:
	ColumnVector dq(jointsNum_);
	vector< int > redunds;
	getCurrRedundantJointsOffsets(t, dq, redunds);

	int redundNum = redunds.size();
	Matrix J(2, redundNum);
	// Perform a jiggle of each redundant joint independently:
	for (int i = 0; i < redundNum; i++) {
		// Compute the robot's current ZMP before jiggle:
		Point_3 zmpBefore, dummy;
		getRobotCurrCOMAndZMP(dummy, zmpBefore);
		
		// Perform the jiggle on that redundant joint:
		ColumnVector qNew = q;
		int redundInx = redunds[i];
		qNew(redundInx) = q(redundInx) + dq(redundInx);
		setJointAngles(qNew);
		robotSimulationStep(1);	// run one step
		
		// Compute the robot's current ZMP after jiggle:
		Point_3 zmpAfter;
		getRobotCurrCOMAndZMP(dummy, zmpAfter);

	// Restore original posture at the beginning of this method:
		setJointAngles(q);
		robotSimulationStep(1);	// run one step
		
		// Calculate the ZMP offset due to jiggle:
		ColumnVector dZMP(2);
		dZMP << (zmpAfter.x() - zmpBefore.x()) 
			 << (zmpAfter.z() - zmpBefore.z());
		
		// Calculate the jacobian's current column:
		J.Column(i + 1) = dZMP / dq(redundInx);
	}

	// Store the whole-body joint angles with the appropriate jacobians:
	qVecs_.push_back(q);
	jacobians_.push_back(J);
// 	PRINT_VALUE(J);	
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::compensateStep
/// \brief Performs one compensate step of the robot's control.
// -----------------------------------------------------------------------------
bool HOAP2_Controller::compensateStep(double & t, bool & isZMPOut) 
{
	TRY(HOAP2_Controller::compensateStep);
	// Read the contact points from shared memory:
	readContactPoints();
	// Get Support Polgon:
	Point3Vec supportPolygon;
	getSupportPolygon(contactPts_, supportPolygon);

	// Get the robot's current CoM and ZMP:
	Point_3 com, zmp;
	getRobotCurrCOMAndZMP(com, zmp);
// 	PRINT_VALUE(com); PRINT_VALUE(zmp);

	// Send supervisor CoM and ZMP:
	float msg[7] = { 1, com[0], com[1], com[2], zmp[0], zmp[1], zmp[2] };
	uint msgSize = sizeof(float) * 7;
	emitterSend(msg, msgSize);

	// Check if ZMP is inside the support polygon:
	CGAL::Bounded_side boundSide = getBoundedSide(supportPolygon, zmp);
	
	// Get the current joint angles vector to set:
	ColumnVector q(jointsNum_); // current column vector of joint angles
	bool res = xmlWalkStep(t, q);

	// Perform compensation using redundant joints only if we're in periodic
	// walking phases and if ZMP is outside support polygon:
	ColumnVector dqCorrected;
	dqCorrected = 0.0;
	if (boundSide == CGAL::ON_UNBOUNDED_SIDE) {
	// Calculate the stability margin:
		Point_3 closestBoundPt;
		getStabilityMargin(supportPolygon, zmp, closestBoundPt);
		
		// Compute required corrections to the ZMP:
		ColumnVector dCorrectZMP(2);
		dCorrectZMP << (closestBoundPt.x() - zmp.x()) 
					<< (closestBoundPt.z() - zmp.z());
		PRINT_VALUE(dCorrectZMP.AsRow());
		
		// Lookup the appropriate jacobian for correction:
		Matrix J = getNearestNeighbourJacobian(q);
		PRINT_VALUE(J);

		// Compute redundant dq for correction using jacobian's pseudo-inverse:
		ColumnVector dq = getCorrectionJointAngles(J, dCorrectZMP);
		PRINT_VALUE(dq.AsRow());
		
		// Set the redundant joints accordingly:
		dqCorrected = dq * COMPENSATE_FACTOR;
		PRINT_VALUE(dqCorrected.AsRow());
		isZMPOut = true;
	} else {
		isZMPOut = false;
	}

	// Determine the correction and store if needed:
	ColumnVector qNew = q;
	CorrectionMap::iterator findIter = dqCorrectMap_.find(t);
	if (findIter != dqCorrectMap_.end()) { // an entry for t exists
		if (!dqCorrected.IsZero()) { // accumulate the correction
			findIter->second += dqCorrected;
		}
		qNew += findIter->second;
	} else if (!dqCorrected.IsZero()) {	// no entry for t exists
		dqCorrectMap_[t] = dqCorrected;
		qNew += dqCorrected;
	}

	setJointAngles(qNew);
	t += robotSimulationStep(1); // run one step
	return res;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::actualStep
/// \brief Performs one step of the robot's actual control.
// -----------------------------------------------------------------------------
bool HOAP2_Controller::actualStep(double & t) 
{
	TRY(HOAP2_Controller::actualStep);
	// Read the contact points from shared memory:
	readContactPoints();
	// Get Support Polgon:
	Point3Vec supportPolygon;
	getSupportPolygon(contactPts_, supportPolygon);

	// Get the robot's current CoM and ZMP:
	Point_3 com, zmp;
	getRobotCurrCOMAndZMP(com, zmp);
// 	PRINT_VALUE(com); PRINT_VALUE(zmp);

	// Send supervisor CoM and ZMP:
	float msg[7] = { 1, com[0], com[1], com[2], zmp[0], zmp[1], zmp[2] };
	uint msgSize = sizeof(float) * 7;
	emitterSend(msg, msgSize);
	
	// Check if ZMP is inside the support polygon:
	CGAL::Bounded_side boundSide = getBoundedSide(supportPolygon, zmp);
	if (boundSide == CGAL::ON_UNBOUNDED_SIDE)
		cout << "HOAP2: The ZMP is outside despite corrections!" << endl;

	// Get the current joint angles vector to set:
	ColumnVector q(jointsNum_); // current column vector of joint angles
	bool res = xmlWalkStep(t, q);

	// Perform correction using redundant joints only if we're in periodic
	// walking phases: 
	ColumnVector qNew = q;
	CorrectionMap::iterator findIter = dqCorrectMap_.find(t);
	if (findIter != dqCorrectMap_.end()) { // an entry for t exists
		qNew += findIter->second;
	}
	
	setJointAngles(qNew);
	t += robotSimulationStep(1); // run one step
	return res;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::updateAnklesCompliance
/// \brief Updates the force of the ankle joints as a function of their 
/// position.
/// Details: For now, we create a positive parabolic function:
///    F(q) = a*q^2 + b*q + c
/// When the ankles are at their min/max positions we set the strongest force 
/// (9). When they are at 0 position we give F = 2.
/// Min. point ---> (q_min, F_min) = (0, ANKLE_{PITCH,ROLL}_MIN_FORCE)
/// Max. point ---> (q_max, F_max) = (q_max, JOINT_MAX_FORCE)
// -----------------------------------------------------------------------------
void HOAP2_Controller::updateAnklesCompliance()
{
	TRY(HOAP2_Controller::updateAnklesCompliance);
	float leftAnklePitchPos = servo_get_position(servos_[left_ankle_pitch]);
	float rightAnklePitchPos = servo_get_position(servos_[right_ankle_pitch]);
	float leftAnkleRollPos = servo_get_position(servos_[left_ankle_roll]);
	float rightAnkleRollPos = servo_get_position(servos_[right_ankle_roll]);

	float cPitch = ANKLE_PITCH_MIN_FORCE;
	float aPitch = 
		(JOINT_MAX_FORCE - cPitch) / (ANKLE_PITCH_MAXVAL * ANKLE_PITCH_MAXVAL);
	float cRoll = ANKLE_ROLL_MIN_FORCE;
	float aRoll = 
		(JOINT_MAX_FORCE - cRoll) / (ANKLE_ROLL_MAXVAL * ANKLE_ROLL_MAXVAL);

	float leftAnklePitchForce = 
		aPitch * leftAnklePitchPos * leftAnklePitchPos + cPitch;
	float rightAnklePitchForce = 
		aPitch * rightAnklePitchPos * rightAnklePitchPos + cPitch;
	float leftAnkleRollForce = 
		aRoll * leftAnkleRollPos * leftAnkleRollPos + cRoll;
	float rightAnkleRollForce = 
		aRoll * rightAnkleRollPos * rightAnkleRollPos + cRoll;
	
	servo_set_force(servos_[left_ankle_pitch], leftAnklePitchForce);
	servo_set_force(servos_[right_ankle_pitch], rightAnklePitchForce);
	servo_set_force(servos_[left_ankle_roll], leftAnkleRollForce);
	servo_set_force(servos_[right_ankle_roll], rightAnkleRollForce);
	CATCH;
}

// =============================================================================
//  Class: SqrDistLessThan                                                     
/// \brief A binary predicate whose result represents the truth or falsehood of
/// "less-than" of square distance of two vectors to a specified vector.
// =============================================================================
class SqrDistLessThan : public binary_function< double, double, bool > {
public:
	SqrDistLessThan(const ColumnVector & v) : v_(v) { }
	bool operator()(const ColumnVector & x, const ColumnVector & y) 
	{ return ((x - v_).SumSquare() < (y - v_).SumSquare()); }

private:
	ColumnVector v_;
};

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::getNearestNeighbourJacobian
/// \brief Finds the nearest neighbour vector in the joint angles vectors to the
/// the one given and returns the corresponding jacobian of that nearest 
/// neighbor.
/// \param q the current joint angles vector.
/// \return the nearest neighbour jacobian matrix.
// -----------------------------------------------------------------------------
ReturnMatrix HOAP2_Controller::getNearestNeighbourJacobian(
	const ColumnVector & q) const
{
	TRY(HOAP2_Controller::getNearestNeighbourJacobian);
	vector< ColumnVector >::const_iterator it = 
		min_element(qVecs_.begin(), qVecs_.end(), SqrDistLessThan(q));
	uint minSqrDistPos = it - qVecs_.begin();
	return jacobians_[minSqrDistPos];
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::updateNearestNeighbourJacobian
/// \brief Finds the nearest neighbour vector in the joint angles vectors to the
/// the one given and updates the corresponding jacobian of that nearest 
/// neighbor.
/// \param q the current joint angles vector.
/// \param J the Jacobian to update with.
/// \return the nearest neighbour jacobian matrix.
// -----------------------------------------------------------------------------
void HOAP2_Controller::updateNearestNeighbourJacobian(const ColumnVector & q,
													 const Matrix & J)
{
	TRY(HOAP2_Controller::updateNearestNeighbourJacobian);
	vector< ColumnVector >::const_iterator it = 
		min_element(qVecs_.begin(), qVecs_.end(), SqrDistLessThan(q));
	uint minSqrDistPos = it - qVecs_.begin();
	jacobians_[minSqrDistPos] = J;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::computeJacobian
/// \brief Computes the Jacobian relating joint angles offsets to ZMP offsets.
/// \param dZMP the current ZMP offset vector.
/// \param dq the current joint angles offset vector.
/// \param redundants the current effective redundant joints indices.
/// \return the jacobian matrix.
// -----------------------------------------------------------------------------
ReturnMatrix HOAP2_Controller::computeJacobian(const ColumnVector & dZMP, 
											  const ColumnVector & dq,
											  const vector< int > & redundants) const
{
	TRY(HOAP2_Controller::computeJacobian);
	int redundantNum = redundants.size();
	Matrix J(2, redundantNum);
	for (int k = 0; k < redundantNum; k++) {
		J.Column(k + 1) = dZMP / dq(redundants[k]);
	}
	J.Release();
	return J;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::getDynamicsParameters
/// \brief Prepare all dynamics parameters in Matrix notation for easy 
/// computation of CoM or ZMP.
/// \param masses Output column vector of all robot's links masses.
/// \param pos    Output 3xN matrix of all robot's links absolute positions.
/// \param accel  Output 3xN matrix of all robot's links absolute accelerations.
/// \param angV   Output 3xN matrix of all robot's links angular velocities.
/// \param inertias Output 9xN matrix of all robot's links inertia tensors.
// -----------------------------------------------------------------------------
void HOAP2_Controller::getDynamicsParameters(ColumnVector & masses, Matrix & pos, 
											 Matrix & accel, Matrix & angV, 
											 Matrix & inertias) const
{
	TRY(HOAP2_Controller::getDynamicsParameters);
	uint linksNum = uint(masses.Nrows());
	for (uint i = 0; i < linksNum; i++) {
		masses(i + 1) = links_[i].m();
 		pos.Column(i + 1) = links_[i].p();
		angV.Column(i + 1) = links_[i].w();
		accel.Column(i + 1) = links_[i].a();
		inertias.Column(i + 1) = links_[i].I().AsColumn();
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::getRobotCurrCOMAndZMP
/// \brief Computes the current CoM and ZMP of the robot by retreiving the 
/// current dynamics and then computing ZMP based on the current dynamics.
/// \param com The robot's current CoM (in 3D).
/// \param zmp The robot's current ZMP (in 3D).
// -----------------------------------------------------------------------------
void HOAP2_Controller::getRobotCurrCOMAndZMP(Point_3 & com, Point_3 & zmp)
{
	TRY(HOAP2_Controller::getRobotCurrCOMAndZMP);
	// Read the dynamics information from shared memory:
	readLinkDynamics();
	
	// Masses, positions, accelerations, angular velocities and inertias:
	uint linkNo = links_.size();
	ColumnVector M(linkNo);
	Matrix P(3, linkNo), A(3, linkNo), W(3, linkNo), I(9, linkNo);

	// Prepare the parameters in matrix form:
	getDynamicsParameters(M, P, A, W, I);

	// Compute ZMP:
	com = getCOM(M, P);
	zmp = getZMP(M, P, A, GROUND_HEIGHT, I, W);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::getCurrRedundantJointsOffsets
/// \brief Gets the current redundant joint angles offsets.
/// \param t The current simulation step (in sec).
/// \param dq The output redundant joint angles offsets.
/// \param redundants The output effective redundant joints indices.
// -----------------------------------------------------------------------------
void HOAP2_Controller::getCurrRedundantJointsOffsets(double t, ColumnVector & dq,
													vector< int > & redundants) const
{
	TRY(HOAP2_Controller::getCurrRedundantJointsOffsets);
	const double w = M_PI / TIME_STEP_SEC; // frequency
	const double A = 0.1;		// amplitude
	const double f = -0.5 * M_PI; // phase
	static Gene torsoPitchGene(torso_pitch, A, w, f, 0.0);
	static Gene leftHipRollGene(left_hip_roll, A, w, f, 0.0);
	static Gene rightHipRollGene(right_hip_roll, A, w, -f, 0.0);
	
	dq = 0.0;
	// Update only the redundant joint offsets:
	dq(torso_pitch + 1) = torsoPitchGene.val(t); // for torso_pitch joint
	dq(left_hip_roll + 1) = leftHipRollGene.val(t); // for left_hip_roll joint
	dq(right_hip_roll + 1) = rightHipRollGene.val(t); // for right_hip_roll joint
	
	redundants.clear();
	redundants.push_back(torso_pitch + 1);
	redundants.push_back(left_hip_roll + 1);
	redundants.push_back(right_hip_roll + 1);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::getCorrectionJointAngles
/// \brief Gets the joint correction angles offsets given the ZMP correction 
/// and the Jacobian.
/// \param J The Jacobian.
/// \param dZMP The ZMP desired correction.
/// \return The joints' correction angles offsets vector.
// -----------------------------------------------------------------------------
ColumnVector HOAP2_Controller::getCorrectionJointAngles(
	const Matrix & J, const ColumnVector & dZMP) const
{
	TRY(HOAP2_Controller::getCorrectionJointAngles);
	ColumnVector dq(jointsNum_);
	dq = 0.0;
	
	// Compute the jacobian's pseudo-inverse:
	Matrix pinvJ = getPseudoInverse(J);
	PRINT_VALUE(pinvJ);
	ColumnVector dqRed = pinvJ * dZMP; // only redundant joints

    // Update only the redundant joint offsets:
	dq(torso_pitch + 1) = dqRed(1);	// for torso_pitch joint
	dq(left_hip_roll + 1) = dqRed(2); // for left_hip_roll joint
	dq(right_hip_roll + 1) = dqRed(3); // for right_hip_roll joint

	dq.Release();
	return dq;
	CATCH;
}

void calcLeg(DeviceTag * joints, double timeStep, bool isLeft)
{
	TRY(calcLeg);
	const double AH = 0.5;
	const double AK = 0.5;
	const double phi = 0.5 * M_PI;

	double ah = AH * sin(2 * M_PI * timeStep);
    double ak = sin(2 * M_PI * timeStep + phi);
    ak = (ak < 0) ? ak = 0 : ak * AK;
	int hip = right_hip_pitch;
	int knee = right_knee_pitch;
	int ankle = right_ankle_pitch;
	if (isLeft) {
		hip = left_hip_pitch;
		knee = left_knee_pitch;
		ankle = left_ankle_pitch;
	}
	double hip_q = -(ah + ak);
	double knee_q = 2 * ak;
	double ankle_q = ah - ak;
    servo_set_position(joints[hip], hip_q);
    servo_set_position(joints[knee], knee_q);
    servo_set_position(joints[ankle], ankle_q);
	CATCH;
}

void HOAP2_Controller::simpleWalkStep()
{
	TRY(HOAP2_Controller::simpleWalkStep);
	static double t_R = 0.0;
	static double t_L = 0.0;

	calcLeg(servos_, t_R, false);
    t_L = t_L + 0.1;

	calcLeg(servos_, t_L, true);
    if (t_L > 0.5)
		t_R = t_R + 0.1;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::xmlWalkStep
/// \brief Get one control step of the XML motions (based on the XML motion
/// file) represented as the joint angles vector to set for the robot.
/// \param t The current simulation step (in sec).
/// \param q column vector containing all joint angles for the robot.
/// \return TRUE if planned motion is finished.
// -----------------------------------------------------------------------------
bool HOAP2_Controller::xmlWalkStep(double t, ColumnVector & q)
{
	TRY(HOAP2_Controller::xmlWalkStep);
	if (t == 0) currMotionTraj_.reset();

	bool res = false;
	if (t >= currMotionTraj_.getEndTime()) {
			if (t <= 0) {
				currMotionType_ = FIRST_STEP;
		} else if (currMotionType_ == FIRST_STEP || 
				   currMotionType_ == WALK_RIGHT) {
					currMotionType_ = WALK_LEFT;
		} else {
					currMotionType_ = WALK_RIGHT;
			res = true;
			}
		updateMotionTraj(currMotionType_, t);
	}
	
	currMotionTraj_.interpolate(t, q);
	return res;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::getJointTrajectories
/// \brief Creates a motion trajectory object based on the given motion type 
/// (see biped_trajectory).
/// \param motionSeq the type of motion represented as a string.
/// \param currTime current simulation time step.
/// \param motionTimeInterval the time interval between each control point.
// -----------------------------------------------------------------------------
void HOAP2_Controller::getJointTrajectories(string motionSeq, double currTime, 
										   double motionTimeInterval)
{
	TRY(HOAP2_Controller::getJointTrajectories);
	cout << "HOAP2: Retrieving motion sequence " << motionSeq << endl;
	MotionSequencesMap::const_iterator motionIter = 
		motionSequencesMap_.find(motionSeq);
	if (motionIter == motionSequencesMap_.end()) 
		THROW_FATAL("The given motion does not exist: " <<  motionSeq);
	
	Matrix positions = motionIter->second;
	uint posNum = uint(positions.ncols());

	Matrix pts(jointsNum_ + 1, posNum + 1);
	pts = 0.0;
	pts.SubMatrix(2, jointsNum_ + 1, 2, posNum + 1) = positions;
	for (uint k = 0; k <= posNum; k++) {
		pts(1, k + 1) = currTime + k * motionTimeInterval;
	}
	pts.SubMatrix(2, jointsNum_ + 1, 1, 1) = getJointAngles();
	
	currMotionTraj_.setPoints(pts, Trajectory::SMOOTH);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::updateMotionTraj
/// \brief Updates the motion trajectory object to a given motion type.
/// \param type Type of motion.
/// \param trajectories motion trajectory object.
// -----------------------------------------------------------------------------
void HOAP2_Controller::updateMotionTraj(MotionType type, double currTime)
{
	TRY(HOAP2_Controller::updateMotionTraj);
	string motionStr;
	switch (type) {
		case STAND_FACEUP:    motionStr = "STAND_FACEUP";    break;
		case STAND_FACEDOWN:  motionStr = "STAND_FACEDOWN";  break;
		case TURN_RIGHT:      motionStr = "TURN_RIGHT";      break;
		case TURN_LEFT:       motionStr = "TURN_LEFT";       break;
		case SIDESTEP_RIGHT:  motionStr = "SIDESTEP_RIGHT";  break;
		case SIDESTEP_LEFT:   motionStr = "SIDESTEP_LEFT";   break;
		case WALK_LEFT:       motionStr = "WALK_LEFT";       break;
		case WALK_RIGHT:      motionStr = "WALK_RIGHT";      break;
		case FIRST_STEP:      motionStr = "FIRST_STEP";	     break;
		case LAST_STEP_LEFT:  motionStr = "LAST_STEP_LEFT";  break;
		case LAST_STEP_RIGHT: motionStr = "LAST_STEP_RIGHT"; break;
		default:
			THROW_FATAL("Undefined type of motion: " << type);
	}
	getJointTrajectories(motionStr, currTime, 16 * TIME_STEP_SEC);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::initSharedMem
/// \brief Initialize the IPC to the physics library.
/// Details: Opens a shared memory segment (already created by hoap-2 physics) 
/// and attaches it to a local pointer. In addition, opens the mutual semaphore
/// for safe access to the shared memory.
// -----------------------------------------------------------------------------
void HOAP2_Controller::initSharedMem()
{
	TRY(HOAP2_Controller::initSharedMem);
	cout << "HOAP2: Initializing shared memory and semaphores" << endl;
	// 	Get the shared memory segments and attach them:
	if (shmDynamics.openAndAttach(DYN_SHM_KEY, sizeof(DynamicsInfo)) == -1)
		THROW_FATAL("Couldn't open the dynamics shared memory segment!");
	if (shmFootGround.openAndAttach(FOOT_SHM_KEY, sizeof(FootGroundInfo)) == -1)
		THROW_FATAL("Couldn't open the foot-ground shared memory segment!");

	dynShmPtr = (DynamicsInfo *) shmDynamics.getSegmentPtr();
	footShmPtr = (FootGroundInfo *) shmFootGround.getSegmentPtr();
	
	// Open the shared memory semaphores. 
	// The hoap2 physics must have created them already:
	if (semDynamics.open(DYN_SEM_KEY, Semaphore::SEM_OPEN) < 0)
		THROW_FATAL("HOAP2: Can't open dynamics shared memory semaphore");
	if (semFootGround.open(FOOT_SEM_KEY, Semaphore::SEM_OPEN) < 0)
		THROW_FATAL("HOAP2: Can't open foot-ground shared memory semaphore");
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::finalizeSharedMem
/// \brief Detaches the shared memory segment and removes it. Also, closes the 
/// mutual semaphore for shared memory access.
// -----------------------------------------------------------------------------
void HOAP2_Controller::finalizeSharedMem()
{
	TRY(HOAP2_Controller::finalizeSharedMem);
	cout << "HOAP2: Finalizing shared memory and semaphores" << endl;
	// Detach the shared memory segments:
	if (shmDynamics.detach() < 0)
		THROW_FATAL("HOAP2: Can't detach dynamics shared memory segment");
	if (shmFootGround.detach() < 0)
		THROW_FATAL("HOAP2: Can't detach foot-ground shared memory segment");

	// Close shared memory semaphores:
	semDynamics.close();
	semFootGround.close();
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::readLinkDynamics
/// \brief Read the robot's links current dynamics from the shared memory and
/// update in the internal links vector.
// -----------------------------------------------------------------------------
void HOAP2_Controller::readLinkDynamics()
{
	TRY(HOAP2_Controller::readLinkDynamics);
	if (semDynamics.acquire() < 0) // Lock the shared memory semaphore
		THROW_ERROR("Cannot acquire the foot-ground semaphore!");

	for (uint i = 0; i < NUM_RIGID_BODIES; i++) {
		links_[i].w() << dynShmPtr->angVel[i];
		links_[i].p() << dynShmPtr->bodyPos[i];
	}

	if (semDynamics.release() < 0) // Unlock the shared memory semaphore 
		THROW_ERROR("Cannot release foot-ground semaphore!");
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::readLinkConstDynamics
/// \brief Read the robot's links constant dynamics from the shared memory and
/// update in the internal links vector.
// -----------------------------------------------------------------------------
void HOAP2_Controller::readLinkConstDynamics()
{
	TRY(HOAP2_Controller::readLinkConstDynamics);
	if (semDynamics.acquire() < 0) // Lock the shared memory semaphore
		THROW_ERROR("Cannot acquire the foot-ground semaphore!");

	for (uint i = 0; i < NUM_RIGID_BODIES; i++) {
		links_[i].m() = dynShmPtr->mass[i];
		links_[i].r() << dynShmPtr->massPos[i];
		const dMatrix3 & shmInertia = dynShmPtr->inertia[i];
		// NOTE: Apparently the inertia comes out unsymmetrical, so here
		// we fix the order:
		links_[i].I() << shmInertia[0] << shmInertia[1] << shmInertia[2]
					  << shmInertia[4] << shmInertia[5] << shmInertia[3]
					  << shmInertia[8] << shmInertia[7] << shmInertia[6];
	}

	if (semDynamics.release() < 0) // Unlock the shared memory semaphore 
		THROW_ERROR("Cannot release foot-ground semaphore!");
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::readContactPoints
/// \brief Read the feet-gound contact points from the shared memory and
/// update in the internal contact points vector.
// -----------------------------------------------------------------------------
void HOAP2_Controller::readContactPoints()
{
	TRY(HOAP2_Controller::readContactPoints);
	if (semFootGround.acquire() < 0) // Lock the shared memory semaphore
		THROW_ERROR("Cannot acquire the foot-ground semaphore!");

	contactPts_.clear();
	for (uint k = 0; k < 2 * CONTACT_NUM; k++) {
		dReal x = footShmPtr->footGroundContacts[k][0];
		dReal y = footShmPtr->footGroundContacts[k][1];
		dReal z = footShmPtr->footGroundContacts[k][2];
		if (x != 0 || y != 0 || z != 0)
			contactPts_.push_back(Point_3(x, y, z));
	}

	if (semFootGround.release() < 0) // Unlock the shared memory semaphore 
		THROW_ERROR("Cannot release foot-ground semaphore!");
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::readJacobianData
/// \brief Read the jacobians input file and initialize the internal jacobians 
/// data structure.
/// \param filename The input filename to read from.
// -----------------------------------------------------------------------------
void HOAP2_Controller::readJacobianData(const string & filename)
{
	TRY(HOAP2_Controller::readJacobianData);
	cout << "HOAP2: Reading jacobians input file\n";
	MatrixFileIO jacobsInputFile(filename);
	vector< Matrix > data;
	vector< string > titles;
	jacobsInputFile.read(data, titles);
	qVecs_.clear();
	jacobians_.clear();
	for (uint i = 0; i < data.size(); i++) {
		if (titles[i].find('J') != string::npos) 
			jacobians_.push_back(data[i]);
		else if (titles[i].find('q') != string::npos) 
			qVecs_.push_back(data[i]);
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::writeJacobianData
/// \brief Write the joint angles vectors and their corresponding jacobians to 
/// the jacobians output file.
/// \param filename The output filename to write to.
// -----------------------------------------------------------------------------
void HOAP2_Controller::writeJacobianData(const string & filename) const 
{
	TRY(HOAP2_Controller::writeJacobianData);
	cout << "HOAP2: Writing to jacobians output file" << endl;
	MatrixFileIO jacobsOutputFile(filename);
	uint qVecSize = qVecs_.size(), jacobiansSize = jacobians_.size();
	if (qVecSize == 0 || jacobiansSize == 0) {
		cout << "HOAP2: WARNING: No q's or jacobians to write!" << endl;
		return;
	}
	BIPED_ASSERT(qVecSize == jacobiansSize);

	// Prepare data and its corresponding titles:
	vector< string > titles;
	vector< Matrix > data;
	for (uint i = 0; i < qVecSize; i++) {
		ostringstream intStream;
		intStream << i;
		titles.push_back("q" + intStream.str());
		data.push_back(qVecs_[i]);
		titles.push_back("J" + intStream.str());
		data.push_back(jacobians_[i]);
	}
	// Write the data to the file:
	jacobsOutputFile.write(data, titles);
	CATCH;
	}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::readCorrectionMap
/// \brief Read the correction map input file and initialize the internal data
/// structure.
/// \param filename The input filename to read from.
// -----------------------------------------------------------------------------
void HOAP2_Controller::readCorrectionMap(const string & filename)
{
	TRY(HOAP2_Controller::readCorrectionMap);
	cout << "HOAP2: Reading correction map input file" << endl;
	MatrixFileIO correctionInputFile(filename);

	vector< Matrix > data;
	vector< string > titles;
	correctionInputFile.read(data, titles);
	dqCorrectMap_.clear();
	for (uint i = 0; i < data.size(); i++) {
		double t = atof(titles[i].substr(2).c_str());
		dqCorrectMap_[t] = data[i];
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::writeCorrectionMap
/// \brief Write the correction map to the correction map output file.
/// \param filename The output filename to write to.
// -----------------------------------------------------------------------------
void HOAP2_Controller::writeCorrectionMap(const string & filename) const
{
	TRY(HOAP2_Controller::writeCorrectionMap);
	cout << "HOAP2: Writing to correction map output file" << endl;
	MatrixFileIO correctionOutputFile(filename);

	if (dqCorrectMap_.size() == 0) {
		cout << "HOAP2: WARNING: Correction map is empty - not writing!" << endl;
		return;
	}
	
	vector< string > titles;
	vector< Matrix > data;
	CorrectionMap::const_iterator it = dqCorrectMap_.begin();
	for (; it != dqCorrectMap_.end(); it++) {
		ostringstream timeStream;
		timeStream << it->first;
		titles.push_back("dq" + timeStream.str());
		data.push_back(it->second);
	}

	correctionOutputFile.write(data, titles);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::readPersistentData
/// \brief Read any persistent data required for the controller. 
// -----------------------------------------------------------------------------
void HOAP2_Controller::readPersistentData()
{
	TRY(HOAP2_Controller::readPersistentData);
	// Read motion sequences XML file:
	readMotionXML(MOTIONS_XML_FILENAME);

	if (ctype_ == COMPENSATE) { 
		// Read input jacobians file:
		readJacobianData(INPUT_JACOBIANS_FILENAME);
	}
	if (ctype_ == ACTUAL) {
		// Read input correction map file:
		readCorrectionMap(INPUT_CMAP_FILENAME);
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::writePersistentData
/// \brief Write any persistent data required for saving. 
// -----------------------------------------------------------------------------
void HOAP2_Controller::writePersistentData()
{
	TRY(HOAP2_Controller::writePersistentData);
	if (ctype_ != ACTUAL) {
		// Write output jacobians file:
		writeJacobianData(OUTPUT_JACOBIANS_FILENAME);
	}
	if (ctype_ == COMPENSATE) {
		// Write output correction map file:
		writeCorrectionMap(OUTPUT_CMAP_FILENAME);
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::initialize
/// \brief Handle any initialization for the HOAP2_Controller object, i.e. allocate 
/// memory, open shared memory, open semaphores etc.
// -----------------------------------------------------------------------------
void HOAP2_Controller::initialize()
{
	TRY(HOAP2_Controller::initialize);
	// Initialize shared memory stuff (segments, semaphores etc):
	initSharedMem();

	// Read any persistent data:
	readPersistentData();
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: HOAP2_Controller::finalize
/// \brief Handle any cleanup for the HOAP2_Controller object, i.e. cleanup memory, 
/// close shared memory, close semaphores etc.
// -----------------------------------------------------------------------------
void HOAP2_Controller::finalize()
{
	TRY(HOAP2_Controller::finalize);
	// Finalize shared memory stuff (segments, semaphores etc):
	finalizeSharedMem();

	// Write any persistent data:
	writePersistentData();
	CATCH;
}
