/* -*- c++ -*- */
/****************************************************************
 *                                                              *
 *  Project name:      Biped Locomotion                         *
 *                                                              *
 *  Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)           *
 ****************************************************************
 * $Id: hoap2_controller.hpp,v 1.4 2004/08/06 18:02:59 s0343981 Exp $ 
 ****************************************************************/
/**
 * \file hoap2_controller.hpp 
 * \brief class HOAP2_Controller definition.
 */

#ifndef __HOAP2_CONTROLLER_HPP__
#define __HOAP2_CONTROLLER_HPP__

#include <biped_namespace.hpp>
#include <biped_trajectory.hpp>
#include <biped_controller.hpp>

#define GROUND_HEIGHT             0.2
#define COMPENSATE_FACTOR         0.01
#define WALK_CYCLES_NUM           5
#define MOTIONS_XML_FILENAME      "Motions.xml"
#define INPUT_JACOBIANS_FILENAME  "Jacob_input.dat"
#define OUTPUT_JACOBIANS_FILENAME "Jacob_output.dat"
#define INPUT_CMAP_FILENAME       "CMap_input.dat"
#define OUTPUT_CMAP_FILENAME      "CMap_output.dat"

BIPED_BEGIN_NAMESPACE(HOAP2);

typedef map< double, ColumnVector > CorrectionMap;

// =============================================================================
//  Class: HOAP2_Controller
/// \brief C++ definition of a HOAP-2 specific controller which abides with 
/// the generic biped controller interface.
// =============================================================================
class HOAP2_Controller : public Controller {
public:						// definitions
	typedef enum { LEARNING, COMPENSATE, ACTUAL } ControlType;

	typedef enum { STAND_FACEUP, STAND_FACEDOWN, TURN_RIGHT, TURN_LEFT,
				   SIDESTEP_RIGHT, SIDESTEP_LEFT, WALK_LEFT, WALK_RIGHT,
				   FIRST_STEP, LAST_STEP_LEFT, LAST_STEP_RIGHT } MotionType;

public:						// methods
	HOAP2_Controller(ControlType ctype, uint jNum, int lt, int rt, int range, 
					 int gps, DeviceTag * joints, DeviceTag camera, 
					 DeviceTag emitter);
	~HOAP2_Controller() { finalize(); }
	
	void run();
	void learnRun(); 
	void compensateRun(); 
	void actualRun();
	
private:						// methods
	// Balance control related:
	void updateAnklesCompliance();
	ReturnMatrix getNearestNeighbourJacobian(const ColumnVector & q) const;
	void updateNearestNeighbourJacobian(const ColumnVector & q,
										const Matrix & J);
	ReturnMatrix computeJacobian(const ColumnVector & dZMP, 
								 const ColumnVector & dq,
								 const vector< int > & redundants) const;
	void getDynamicsParameters(ColumnVector & masses, Matrix & pos, 
							   Matrix & accel, Matrix & angV, 
							   Matrix & inertias) const;
	void getRobotCurrCOMAndZMP(Point_3 & com, Point_3 & zmp);
	
	// Redundant joints management:
	void getCurrRedundantJointsOffsets(double t, ColumnVector & dq, 
									   vector< int > & redundants) const;
	ColumnVector getCorrectionJointAngles(const Matrix & J, 
										  const ColumnVector & dZMP) const;

	// Motion control:
	double robotSimulationStep(uint stepNum) 
	{ 
		robot_step(TIME_STEP);
		return stepNum * TIME_STEP_SEC;	// the accumulated real-time performed 
	}
	void initialStep();
	bool compensateStep(double & t, bool & isZMPOut);
	bool actualStep(double & t);
	bool regularStep(double & t, bool & isZMPOut);
	void jiggleAllStep(double t);
	void jiggleIndependentStep(double t);
	bool xmlWalkStep(double t, ColumnVector & q);
	void simpleWalkStep();

	void getJointTrajectories(string motionSeq, double currTime, 
							  double motionTimeInterval);
	void updateMotionTraj(MotionType type, double currTime);

	// Update the physical information from the physics:
	void readLinkDynamics();
	void readLinkConstDynamics();
	void readContactPoints();
	
	// Shared memory management:
	void initSharedMem();
	void finalizeSharedMem();

	// Read/Write data management:
	void readJacobianData(const string & filename);
	void writeJacobianData(const string & filename) const;
	void readCorrectionMap(const string & filename);
	void writeCorrectionMap(const string & filename) const;
	void readPersistentData();
	void writePersistentData();

	// Initialization and finalization methods:
	void initialize();
	void finalize();
	
private:						// members
	ControlType ctype_;
	Point3Vec contactPts_;
	vector< Link > links_;

	Trajectory currMotionTraj_; ///< the current trajectories object
	MotionType currMotionType_; ///< the current motion type being used

	vector< ColumnVector > qVecs_; ///< storage for joint angles vectors
	vector< Matrix > jacobians_; ///< storage for jacobian matrices
	CorrectionMap dqCorrectMap_;
};
		
BIPED_END_NAMESPACE;

#endif	// __HOAP2_CONTROLLER_HPP__
