/* -*- c++ -*- */
/****************************************************************
 *                                                              *
 *  Project name:      Biped Locomotion                         *
 *                                                              *
 *  Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)           *
 ****************************************************************
 * $Id: qrio.cpp,v 1.8 2004/08/29 11:09:54 s0343981 Exp $ 
 ****************************************************************/
/**
 * \file qrio.cpp 
 */

#include <device/robot.h>
#include <iostream>
#include <biped_exception.hpp>
#include <biped_logging.hpp>
#include <qrio/qrio_joints.hpp>
#include "qrio_controller.hpp"

using std::cout;
using std::endl;
using namespace QRIO;

static int range_sensor;
static int left_touch;
static int right_touch;
static int gps;
static DeviceTag joints[JOINTS_NUM];
DeviceTag camera, emitter;
static QRIO_Controller * controller = NULL;
static QRIO_Controller::ControlType controlType = QRIO_Controller::LEARNING;

// -----------------------------------------------------------------------------
//  Function: qrio_reset
/// \brief The global implementation for a resetting Qrio controller.
/// Details: Here the device handles are collected.
// -----------------------------------------------------------------------------
static void qrio_reset(void) 
{
	TRY(qrio_reset);
	cout << "Reseting Qrio controller" << endl;
	QRIO_JointInfo jInfo;
	for (int i = 0; i < JOINTS_NUM; i++) {
		joints[i] = robot_get_device(jInfo.toString(i));
	}
	for (int i = 0; i < JOINTS_NUM; i++) {
		if (joints[i] == 0)
			cout << "Joint " << i << " is missing" << endl;
	}
	
	// Get the rest of the devices:
	camera = robot_get_device("camera");
	range_sensor = robot_get_device("distance");
	left_touch = robot_get_device("left touch");
	right_touch = robot_get_device("right touch");
	gps = robot_get_device("gps");
	emitter = robot_get_device("emitter");
	if (emitter == 0) 
		cout << "missing emitter" << endl;

	controller = new QRIO_Controller(controlType, JOINTS_NUM, left_touch, 
									 right_touch, range_sensor, gps, joints, 
									 camera, emitter);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: qrio_exit
/// \brief The global implementation for a exiting the Qrio controller.
/// Details: Here the device handles are collected.
// -----------------------------------------------------------------------------
static void qrio_exit(void) 
{
	TRY(qrio_exit);
	cout << "Exiting Qrio controller" << endl;
	delete controller;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: determine_run_type
/// \brief Determines which type of control to use while running the robot.
// -----------------------------------------------------------------------------
static void determine_run_type()
{
	TRY(determine_run_type);
	char * envControlType = getenv("QRIO_RUN");
    if (envControlType == 0 || strcmp(envControlType, "learn") == 0) {
		controlType = QRIO_Controller::LEARNING;
	} else if (strcmp(envControlType, "regular") == 0) {
		controlType = QRIO_Controller::REGULAR;
	} else if (strcmp(envControlType, "compensate") == 0) {
		controlType = QRIO_Controller::COMPENSATE;
	} else if (strcmp(envControlType, "actual") == 0) {
		controlType = QRIO_Controller::ACTUAL;
	} else {
		THROW_FATAL("Unknown run type for Qrio robot!");
	}
	cout << "\nQRIO ";
	switch (controlType) {
		case QRIO_Controller::REGULAR:    cout << "regular";      break;
		case QRIO_Controller::LEARNING:   cout << "learning";     break;
		case QRIO_Controller::COMPENSATE: cout << "compensation"; break;
		case QRIO_Controller::ACTUAL:     cout << "actual";       break;
	}
	cout << " control!" << endl;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: main
/// \brief The main program of Qrio robot controller.
// -----------------------------------------------------------------------------
int main() 
{
	TRY(main);
	// Determine the control type to run the robot:
	determine_run_type();

	// Bind the robot_live and robot_die functions:
	robot_live(qrio_reset);
	robot_die(qrio_exit);

	controller->enableDevices();
	try {						// Special treatment for newmat exceptions
		controller->run();
	} catch (Exception) { cout << Exception::what() << endl; }	

	qrio_exit();
	CATCH_FINAL;
	return 0;
}
