/* -*- c++ -*- */
/****************************************************************
 *                                                              *
 *  Project name:      Biped Locomotion                         *
 *                                                              *
 *  Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)           *
 ****************************************************************
 * $Id: qrio_controller.cpp,v 1.9 2004/08/29 11:09:54 s0343981 Exp $ 
 ****************************************************************/
/**
 * \file qrio_controller.cpp 
 * \brief class QRIO_Controller methods' implementations.
 */

// CGAL includes:
#include <CGAL/convex_hull_2.h>
// Webots includes:
#include <device/robot.h>
#include <device/servo.h>
#include <device/emitter.h>
#include <device/camera.h>
#include <device/distance_sensor.h>
#include <device/gps.h>
#include <device/touch_sensor.h>
// ROBOOP includes:
#include <newmat/newmat.h>
#include <newmat/newmatio.h>
// C++ includes:
#include <iostream>
#include <sstream>
#include <fstream>
#include <iomanip>
#include <string>
// BIPED includes:
#include <biped_exception.hpp>
#include <biped_gene.hpp>
#include <biped_logging.hpp>
#include <biped_matrix_io.hpp>
#include <biped_sem.hpp>
#include <biped_shmem.hpp>
#include <biped_util.hpp>
// QRIO includes:
#include <qrio/qrio_joints.hpp>
#include <qrio/qrio_shmem.hpp>
#include "qrio_controller.hpp"

using std::cout;
using std::endl;
using std::string;
using namespace QRIO;

// Shared memory stuff:
static SharedMemory shmDynamics; // dynamics shared memory manager
static SharedMemory shmFootGround; // foot-ground shared memory manager
static DynamicsInfo * dynShmPtr; // ptr to dynamics struct in shared memory
static FootGroundInfo * footShmPtr; // ptr to foot-ground struct in shared memory
static Semaphore semDynamics;	// dynamics shared memory semaphore
static Semaphore semFootGround;	// foot-ground shared memory semaphore
static dReal groundHeight = GROUND_HEIGHT;
static vector< dReal > q_dots;
static vector< ColumnVector > zmp_dots;

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::QRIO_Controller
/// \brief QRIO_Controller's c'tor.
/// \param ctype The control type.
/// \param jointsNum No. of joints existing for this robot.
/// \param lt Handle to the left touch sensor device.
/// \param rt Handle to the right touch sensor device.
/// \param range Handle to the range sensor device.
/// \param gps Handle to the GPS sensor device.
/// \param joints Array of handles to all servo joint devices.
/// \param camera Handle to the camera device.
/// \param emitter Handle to the emitter device.
// -----------------------------------------------------------------------------
QRIO_Controller::QRIO_Controller(ControlType ctype, uint jointsNum, int lt, 
								 int rt, int range, int gps, DeviceTag * joints,
								 DeviceTag camera, DeviceTag emitter) :
	Controller(jointsNum, lt, rt, range, gps, joints, camera, emitter) , 
	ctype_(ctype) , contactPts_() , links_(NUM_RIGID_BODIES) , currMotionTraj_() , 
	currMotionType_(FIRST_STEP) , qVecs_() , jacobians_() , dqCorrectMap_() , 
	qMap_() , zmpMap_() , comMap_() , fzmpTimes_()
{
	TRY(QRIO_Controller::QRIO_Controller);
	initialize();
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::run
/// \brief Perform the endless loop of the robot control.
// -----------------------------------------------------------------------------
void QRIO_Controller::run()
{
	TRY(QRIO_Controller::run);
	robotSimulationStep(1);		// wait a bit ...
	initialStep();
	string prefix;
	switch (ctype_) {
		case REGULAR:    regularRun();     prefix = "r_"; break; 
		case LEARNING:   learnRun();       prefix = "l_"; break;
		case COMPENSATE: compensateRun2(); prefix = "c_"; break;
		case ACTUAL:     actualRun();      prefix = "a_"; break;
		default:
			THROW_FATAL("QRIO: Unknown control type!");
	}
	
	// Plot the redundant degrees q's:
	vector< int > inx; 
	inx.push_back(torso_pitch + 1); inx.push_back(torso_roll + 1);
// 	inx.push_back(left_hip_roll + 1); inx.push_back(right_hip_roll + 1); 
	GNUPlot gpJoints("lines"); 
	gpJoints.setYLabel("q [rad]"); gpJoints.setXLabel("t [sec]");
	plotData(gpJoints, prefix, qMap_, inx);

	// Plot the ZMP and CoM:
	GNUPlot gpZMP("lines"), gpCoM("lines"); 
	gpZMP.setYLabel("ZMP_x_y_z [m]"); gpZMP.setXLabel("t [sec]");
	gpCoM.setYLabel("CoM_x_y_z [m]"); gpCoM.setXLabel("t [sec]");
	plotData(gpZMP, prefix + "zmp", zmpMap_);
	plotData(gpCoM, prefix + "com", comMap_);
	
	// Write the FZMP occurrences to a file:
	writeFzmpData(OUTPUT_FZMP_FILENAME);

	// Wait for user decision to quit:
	pollQuit();
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::regularRun
/// \brief Perform one regular (no corrections or learning) run.
// -----------------------------------------------------------------------------
void QRIO_Controller::regularRun()
{
	TRY(QRIO_Controller::regularRun);
	cout << "QRIO: Starting regular run ..." << endl;
	double t = 0;				// current time [sec]
	fzmpTimes_.clear();
	for (uint i = 0; i < WALK_CYCLES_NUM; i++) {
		cout << "QRIO: Walking cycle " << i << endl;
		bool cycleDone = false, isZMPout = false;
		while (!cycleDone) {
			double tmpTime = t;
			cycleDone = regularStep(false, t, isZMPout);
			if (isZMPout && (currMotionType_ == WALK_RIGHT || 
							 currMotionType_ == WALK_LEFT)) {
				fzmpTimes_.push_back(tmpTime);
				cout << "QRIO: ZMP(t=" << tmpTime << ") is outside!" << endl;
			}
		}
	}
	uint zmpOutNum = fzmpTimes_.size();
	PRINT_VALUE(zmpOutNum);

	cout << "QRIO: Finished regular run!" << endl;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::correctionRun
/// \brief Perform one corrections run.
// -----------------------------------------------------------------------------
uint QRIO_Controller::correctionRun()
{
	TRY(QRIO_Controller::correctionRun);
	cout << "QRIO: Starting correction run ..." << endl;
	double t = 0;				// current time [sec]
	fzmpTimes_.clear();
	for (uint i = 0; i < WALK_CYCLES_NUM; i++) {
		cout << "QRIO: Walking cycle " << i << endl;
		bool cycleDone = false, isZMPout = false;
		while (!cycleDone) {
			if (currMotionType_ == WALK_RIGHT || 
				currMotionType_ == WALK_LEFT) {
				double tmpTime = t;
				cycleDone = compensateStep(t, isZMPout);
				if (isZMPout) {
					fzmpTimes_.push_back(tmpTime);
					cout << "ZMP(t=" << tmpTime << ") is outside!" << endl;
				}
			} else {
				cycleDone = regularStep(false, t, isZMPout);
			}
		}
	}
	cout << "QRIO: Finished correction run!" << endl;
	return fzmpTimes_.size();
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::jiggleRun
/// \brief Perform one run of jiggles to update the jacobians.
// -----------------------------------------------------------------------------
void QRIO_Controller::jiggleRun()
{
	TRY(QRIO_Controller::jiggleRun);
	cout << "QRIO: Starting jiggle run ..." << endl;
	double t = 0;				// current time [sec]
	for (uint i = 0; i < WALK_CYCLES_NUM; i++) {
		cout << "QRIO: Walking cycle " << i << endl;
		bool cycleDone = false, isZMPout = false;
		while (!cycleDone) {
			if (currMotionType_ == WALK_RIGHT || 
				currMotionType_ == WALK_LEFT) {
				jiggleIndependentStep(t);
			}
			cycleDone = regularStep(true, t, isZMPout);
		}
	}
	cout << "QRIO: Finished jiggle run!" << endl;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::learnRun
/// \brief Perform one run of learning.
// -----------------------------------------------------------------------------
void QRIO_Controller::learnRun()
{
	TRY(QRIO_Controller::learnRun);
	cout << "QRIO: Starting learning run ..." << endl;
	double t = 0;				// current time [sec]
	fzmpTimes_.clear();
	for (uint i = 0; i < WALK_CYCLES_NUM; i++) {
		cout << "QRIO: Walking cycle " << i << endl;
		bool cycleDone = false, isZMPout = false;
		while (!cycleDone) {
			if (currMotionType_ == WALK_RIGHT || 
				currMotionType_ == WALK_LEFT) {
				jiggleIndependentStep(t);
			}
			double tmpTime = t;
			cycleDone = regularStep(false, t, isZMPout);
			if (isZMPout && (currMotionType_ == WALK_RIGHT || 
							 currMotionType_ == WALK_LEFT)) {
				fzmpTimes_.push_back(tmpTime);
				cout << "QRIO: ZMP(t=" << tmpTime << ") is outside!" << endl;
			}
		}
	}
	uint zmpOutNum = fzmpTimes_.size();
	PRINT_VALUE(zmpOutNum);
	cout << "QRIO: Finished learning run!" << endl;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::compensateRun
/// \brief Perform multiple runs of compensation until the ZMP is no longer 
/// outside the support polygon.
// -----------------------------------------------------------------------------
void QRIO_Controller::compensateRun()
{
	TRY(QRIO_Controller::compensateRun);
	cout << "QRIO: Starting compensation run ..." << endl;
	GNUPlot gpQ("lines");
	gpQ.setYLabel("q [rad]"); gpQ.setXLabel("t [sec]");
	for (uint iterNum = 0;; iterNum++) {
		currMotionType_ = FIRST_STEP; // reset back to the FIRST_STEP motion
		double t = 0;				// current time [sec]
		uint zmpOutNum = 0;
		// Perform one iteration of walking cycles:
		for (uint i = 0; i < WALK_CYCLES_NUM; i++) {
			cout << "Walking cycle " << i << endl;
			bool cycleDone = false, isZMPout = false;
			while (!cycleDone) {
				if (currMotionType_ == WALK_RIGHT || 
					currMotionType_ == WALK_LEFT) {
					jiggleIndependentStep(t);
					cycleDone = compensateStep(t, isZMPout);
					if (isZMPout) {
						zmpOutNum++;
						cout << "ZMP(t=" << t << ") is outside!" << endl;
					}
				} else {
					cycleDone = regularStep(false, t, isZMPout);
				}
			}
		}
		PRINT_VALUE(zmpOutNum);
		if (zmpOutNum == 0)
			break;
		// Reset all joints to zero:
		cout << "QRIO: Resetting all joints to zero!" << endl;
		ColumnVector q(jointsNum_);
		q = 0.0;
		setJointAngles(q);
		robotSimulationStep(3);	// wait a bit

		// Request supervisor to reset the robot's position:
		cout << "QRIO: Requesting supervisor to reset my position!" << endl;
		float msg[8] = { 2, 0, 0, 0, 0, 0, 0, 0 };
		uint msgSize = sizeof(float) * 8;
		emitterSend(msg, msgSize);
		robotSimulationStep(3); // wait a bit

		// Plot the redundant q's and ZMP offsets:
		vector< int > inx; 
		inx.push_back(torso_pitch + 1); inx.push_back(torso_roll + 1); 
// 		inx.push_back(left_hip_roll + 1); inx.push_back(right_hip_roll + 1); 
		ostringstream prefix;
		prefix << "c_" << iterNum;
		plotData(gpQ, prefix.str(), qMap_, inx);
	}
	cout << "QRIO: Finished compensation run!" << endl;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::compensateRun2
/// \brief Perform multiple runs of compensation until the ZMP is no longer 
/// outside the support polygon.
// -----------------------------------------------------------------------------
void QRIO_Controller::compensateRun2()
{
	TRY(QRIO_Controller::compensateRun2);
	cout << "QRIO: Starting compensation run 2 ..." << endl;
	GNUPlot gpQ("lines");
	gpQ.setYLabel("q [rad]"); gpQ.setXLabel("t [sec]");
	for (uint iterNum = 0;; iterNum++) {
		currMotionType_ = FIRST_STEP; // reset back to the FIRST_STEP motion
		if (iterNum % 2 == 0) { // perform one correction run
			uint zmpOutNum = correctionRun();
			PRINT_VALUE(zmpOutNum);
			if (zmpOutNum == 0)	
				break;
		} else {				// perform one run to update jacobians
			jiggleRun();
		}

		// Reset all joints to zero:
		cout << "QRIO: Resetting all joints to zero!" << endl;
		ColumnVector q(jointsNum_);
		q = 0.0;
		setJointAngles(q);
		robotSimulationStep(3);	// wait a bit

		// Request supervisor to reset the robot's position:
		cout << "QRIO: Requesting supervisor to reset my position!" << endl;
		float msg[8] = { 2, 0, 0, 0, 0, 0, 0, 0 };
		uint msgSize = sizeof(float) * 8;
		emitterSend(msg, msgSize);
		robotSimulationStep(3); // wait a bit

		if (iterNum % 2 == 0) {
			// Plot the redundant q's and ZMP offsets:
			vector< int > inx; 
			inx.push_back(torso_pitch + 1); inx.push_back(torso_roll + 1);
// 			inx.push_back(left_hip_roll + 1); inx.push_back(right_hip_roll + 1);
			ostringstream prefix;
			prefix << "c_" << iterNum << "_";
			plotData(gpQ, prefix.str(), qMap_, inx);
		}
	}
	cout << "QRIO: Finished compensation run!" << endl;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::actualRun
/// \brief Perform one actual run of corrected trajectories.
// -----------------------------------------------------------------------------
void QRIO_Controller::actualRun()
{
	TRY(QRIO_Controller::actualRun);
	cout << "QRIO: Starting actual run ..." << endl;
	double t = 0;				// current time [sec]
	fzmpTimes_.clear();
	for (uint i = 0; i < WALK_CYCLES_NUM; i++) {
		cout << "Walking cycle " << i << endl;
		bool cycleDone = false, isZMPout = false;
		while (!cycleDone) {
			if (currMotionType_ == WALK_RIGHT || 
				currMotionType_ == WALK_LEFT) {
				double tmpTime = t;
				cycleDone = regularStep(true, t, isZMPout);
				if (isZMPout) {
					fzmpTimes_.push_back(tmpTime);
					cout << "ZMP(t=" << tmpTime << ") is outside!\n";
				}
			} else {
				cycleDone = regularStep(false, t, isZMPout);
			}
		}
	}
	uint zmpOutNum = fzmpTimes_.size();
	PRINT_VALUE(zmpOutNum);
	cout << "QRIO: Finished actual run!" << endl;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::initialStep
/// \brief Performs the initial step of the robot's control.
// -----------------------------------------------------------------------------
void QRIO_Controller::initialStep() 
{
	TRY(QRIO_Controller::initialStep);
	cout << "QRIO: Performing initial step ..." << endl;
	robotSimulationStep(1); // run one step

	// Read the constant dynamics information. No need to perform this at
	// later calls (the values remain the same):
	readLinkConstDynamics();
	
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::regularStep
/// \brief Performs one regular motion step of the robot's control.
/// \param doCorrect Should motion trajectories be corrected or not ?!
/// \param t the current time in floating point precision.
/// \param isZMPout indicates if ZMP was outside support polygon in this step.
/// \return TRUE if walking cycle has finished during this step.
// -----------------------------------------------------------------------------
bool QRIO_Controller::regularStep(bool doCorrect, double & t, bool & isZMPOut)
{
	TRY(QRIO_Controller::regularStep);
	// Read the contact points from shared memory:
	readContactPoints();
	// Get Support Polgon:
	Point3Vec supportPolygon;
	getSupportPolygon(contactPts_, supportPolygon);
	groundHeight = getGroundHeight(supportPolygon);

	// Get the robot's current CoM and ZMP:
	Point_3 com, zmp;
	getRobotCurrCOMAndZMP(com, zmp);
	zmpMap_[t] = zmp;
	comMap_[t] = com;

	// Send supervisor CoM and ZMP:
	float msg[8] = { 1, com[0], com[1], com[2], zmp[0], zmp[1], zmp[2], t };
	uint msgSize = sizeof(float) * 8;
	emitterSend(msg, msgSize);
	
	// Check if ZMP is inside the support polygon:
	CGAL::Bounded_side boundSide = getBoundedSide(supportPolygon, zmp);
	isZMPOut = (boundSide == CGAL::ON_UNBOUNDED_SIDE);
 
    // Get the current joint angles vector to set:
	ColumnVector q(jointsNum_); // current column vector of joint angles
	bool wasUpdated = updateMotionTrajectories(t);
	if (doCorrect && wasUpdated) 
		correctMotionTrajectories();
	currMotionTraj_.interpolate(t, q);

	setJointAngles(q);
	qMap_[t] = q;
	t += robotSimulationStep(1); // run one step
	return (wasUpdated && (currMotionType_ == WALK_RIGHT));
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::jiggleAllStep
/// \brief Performs one jiggle step of the redundant degrees (all together) in 
/// robot's control.
/// \param t the current time in floating point precision.
// -----------------------------------------------------------------------------
void QRIO_Controller::jiggleAllStep(double t) 
{
	TRY(QRIO_Controller::jiggleAllStep);
	ColumnVector q = getJointAngles(); // get current joint angles

	// Get the current offset in redundant joints:
	ColumnVector dq(jointsNum_);
	vector< int > redunds;
	getCurrRedundantJointsOffsets(t, dq, redunds);

	// Compute the robot's current ZMP before jiggle:
	Point_3 zmpBefore, dummy;
	getRobotCurrCOMAndZMP(dummy, zmpBefore);

	// Perform the jiggle on redundant joints:
	ColumnVector qNew = q + dq;
	setJointAngles(qNew);
	robotSimulationStep(2);	// run two steps
	ColumnVector qActual = getJointAngles();

	// Compute the robot's current ZMP after jiggle:
	Point_3 zmpAfter;
	getRobotCurrCOMAndZMP(dummy, zmpAfter);

	// Restore original posture at the beginning of this method:
	setJointAngles(q);
	robotSimulationStep(2);		// run two steps
	qActual = getJointAngles();

	// Calculate the ZMP offset due to jiggle:
	ColumnVector dZMP(2);
	dZMP << (zmpAfter.x() - zmpBefore.x()) 
		 << (zmpAfter.z() - zmpBefore.z());

	// Calculate the current jacobian, store it along with joint angles vector:
	Matrix J = computeJacobian(dZMP, dq, redunds);
	storeJacobianWithJointAngles(q, J);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::jiggleIndependentStep
/// \brief Performs one jiggle step of the redundant degrees (each one 
/// independently) in robot's control.
/// \param t the current time in floating point precision.
// -----------------------------------------------------------------------------
void QRIO_Controller::jiggleIndependentStep(double t) 
{
	TRY(QRIO_Controller::jiggleIndependentStep);
	ColumnVector q = getJointAngles(); // get current joint angles

	// Get the current offset in redundant joints:
	ColumnVector dq(jointsNum_);
	vector< int > redunds;
	getCurrRedundantJointsOffsets(t, dq, redunds);

	int redundNum = redunds.size();
	Matrix J(2, redundNum);
	// Perform a jiggle of each redundant joint independently:
	for (int i = 0; i < redundNum; i++) {
		// Compute the robot's current ZMP before jiggle:
		Point_3 zmpBefore, dummy;
		getRobotCurrCOMAndZMP(dummy, zmpBefore);
		
		// Perform the jiggle on that redundant joint:
		ColumnVector qNew = q;
		int redundInx = redunds[i];
		dReal dqRed = dq(redundInx);
		qNew(redundInx) = q(redundInx) + dqRed;
// 		PRINT_VALUE(qNew.AsRow());
		setJointAngles(qNew);
		robotSimulationStep(1);	// run two steps
		ColumnVector qActual = getJointAngles();
// 		PRINT_VALUE(qActual.AsRow());
// 		if (qActual(redundInx) - qNew(redundInx) > 0.0001) {
		dqRed = qActual(redundInx) - q(redundInx);
// 		}
// 		if (i == 0 && ctype_ == LEARNING) q_dots.push_back(dqRed);

        // Compute the robot's current ZMP after jiggle:
		Point_3 zmpAfter;
		getRobotCurrCOMAndZMP(dummy, zmpAfter);
		
		// Restore original posture at the beginning of this method:
// 		PRINT_VALUE(q.AsRow());
		setJointAngles(q);
		robotSimulationStep(2); // run two steps
		qActual = getJointAngles();
// 		PRINT_VALUE(qActual.AsRow());
		if (qActual(redundInx) - q(redundInx) > 0.0001)
			PRINT_VALUE(qActual(redundInx) - q(redundInx));

		// Calculate the ZMP offset due to jiggle:
		ColumnVector dZMP(2);
		dZMP << (zmpAfter.x() - zmpBefore.x()) 
			 << (zmpAfter.z() - zmpBefore.z());
// 		if (i == 0 && ctype_ == LEARNING) zmp_dots.push_back(dZMP);
		
		// Calculate the jacobian's current column:
		J.Column(i + 1) = dZMP / dqRed;
	}

	// Store the whole-body joint angles with the appropriate jacobians:
	storeJacobianWithJointAngles(q, J);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::compensateStep
/// \brief Performs one compensate step of the robot's control.
/// \param t the current time in floating point precision.
/// \param isZMPout indicates if ZMP was outside support polygon in this step.
/// \return TRUE if walking cycle has finished during this step.
// -----------------------------------------------------------------------------
bool QRIO_Controller::compensateStep(double & t, bool & isZMPOut) 
{
	TRY(QRIO_Controller::compensateStep);
	// Read the contact points from shared memory:
	readContactPoints();
	// Get Support Polgon:
	Point3Vec supportPolygon;
	getSupportPolygon(contactPts_, supportPolygon);
	groundHeight = getGroundHeight(supportPolygon);

	// Get the robot's current CoM and ZMP:
	Point_3 com, zmp;
	getRobotCurrCOMAndZMP(com, zmp);
	zmpMap_[t] = zmp;
	comMap_[t] = com;

	// Send supervisor CoM and ZMP:
	float msg[8] = { 1, com[0], com[1], com[2], zmp[0], zmp[1], zmp[2], t };
	uint msgSize = sizeof(float) * 8;
	emitterSend(msg, msgSize);
	
	// Check if ZMP is inside the support polygon:
	CGAL::Bounded_side boundSide = getBoundedSide(supportPolygon, zmp);

	// Get the current joint angles vector to set:
	ColumnVector q(jointsNum_); // current column vector of joint angles
	bool wasUpdated = updateMotionTrajectories(t);
	if (wasUpdated) 
		correctMotionTrajectories();
	currMotionTraj_.interpolate(t, q);

	// Perform compensation using redundant joints only if we're in periodic
	// walking phases and if ZMP is outside support polygon:
	ColumnVector dqCorrected;
	dqCorrected = 0.0;
	isZMPOut = (boundSide == CGAL::ON_UNBOUNDED_SIDE);
	if (isZMPOut) {
		// Calculate the stability margin:
		Point_3 closestBoundPt;
		getStabilityMargin(supportPolygon, zmp, closestBoundPt);
		
		// Compute required corrections to the ZMP:
		ColumnVector dCorrectZMP(2);
		dCorrectZMP << (closestBoundPt.x() - zmp.x()) 
					<< (closestBoundPt.z() - zmp.z());
		PRINT_VALUE(dCorrectZMP.AsRow());
		
		// Lookup the appropriate jacobian for correction:
		Matrix J = getNearestNeighbourJacobian(q);
		PRINT_VALUE(J.AsRow());
		
		// Compute redundant dq for correction using jacobian's pseudo-inverse:
		ColumnVector dq = getCorrectionJointAngles(J, dCorrectZMP);
		
		// Set the redundant joints accordingly:
		if (dq.MaximumAbsoluteValue() < CORRECTION_OFFSET_LIMIT)
			dqCorrected = dq;
		else 
			dqCorrected = dq * COMPENSATE_FACTOR;
		PRINT_VALUE(dqCorrected.AsRow())
	}

	// Determine the correction and store if needed:
	ColumnVector qNew = q;
	TimeVectorMap::iterator findIter = dqCorrectMap_.find(t);
	if (!dqCorrected.IsZero()) {
		if (findIter != dqCorrectMap_.end()) { // an entry for t exists
			findIter->second += dqCorrected; // accumulate the correction
		} else {				// no entry for t exists
			dqCorrectMap_[t] = dqCorrected;	// insert the correction
		}
		qNew += dqCorrected;	// add the new correction to the interpolation
	}

	setJointAngles(qNew);
	qMap_[t] = qNew;
	t += robotSimulationStep(1); // run one step
	return (wasUpdated && (currMotionType_ == WALK_RIGHT));
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::updateAnklesCompliance
/// \brief Updates the force of the ankle joints as a function of their 
/// position.
/// Details: For now, we create a positive parabolic function:
///    F(q) = a*q^2 + b*q + c
/// When the ankles are at their min/max positions we set the strongest force 
/// (9). When they are at 0 position we give F = 2.
/// Min. point ---> (q_min, F_min) = (0, ANKLE_{PITCH,ROLL}_MIN_FORCE)
/// Max. point ---> (q_max, F_max) = (q_max, JOINT_MAX_FORCE)
// -----------------------------------------------------------------------------
void QRIO_Controller::updateAnklesCompliance()
{
	TRY(QRIO_Controller::updateAnklesCompliance);
	float leftAnklePitchPos = servo_get_position(servos_[left_ankle_pitch]);
	float rightAnklePitchPos = servo_get_position(servos_[right_ankle_pitch]);
	float leftAnkleRollPos = servo_get_position(servos_[left_ankle_roll]);
	float rightAnkleRollPos = servo_get_position(servos_[right_ankle_roll]);

	float cPitch = ANKLE_PITCH_MIN_FORCE;
	float aPitch = 
		(JOINT_MAX_FORCE - cPitch) / (ANKLE_PITCH_MAXVAL * ANKLE_PITCH_MAXVAL);
	float cRoll = ANKLE_ROLL_MIN_FORCE;
	float aRoll = 
		(JOINT_MAX_FORCE - cRoll) / (ANKLE_ROLL_MAXVAL * ANKLE_ROLL_MAXVAL);

	float leftAnklePitchForce = 
		aPitch * leftAnklePitchPos * leftAnklePitchPos + cPitch;
	float rightAnklePitchForce = 
		aPitch * rightAnklePitchPos * rightAnklePitchPos + cPitch;
	float leftAnkleRollForce = 
		aRoll * leftAnkleRollPos * leftAnkleRollPos + cRoll;
	float rightAnkleRollForce = 
		aRoll * rightAnkleRollPos * rightAnkleRollPos + cRoll;
	
	servo_set_force(servos_[left_ankle_pitch], leftAnklePitchForce);
	servo_set_force(servos_[right_ankle_pitch], rightAnklePitchForce);
	servo_set_force(servos_[left_ankle_roll], leftAnkleRollForce);
	servo_set_force(servos_[right_ankle_roll], rightAnkleRollForce);
	CATCH;
}

// =============================================================================
//  Class: SqrDistLessThan                                                     
/// \brief A binary predicate whose result represents the truth or falsehood of
/// "less-than" of square distance of two vectors to a specified vector.
// =============================================================================
class SqrDistLessThan : public binary_function< double, double, bool > {
public:
	SqrDistLessThan(const ColumnVector & v) : v_(v) { }
	bool operator()(const ColumnVector & x, const ColumnVector & y) 
	{ return ((x - v_).SumSquare() < (y - v_).SumSquare()); }

private:
	ColumnVector v_;
};

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::getNearestNeighbourJacobian
/// \brief Finds the nearest neighbour vector in the joint angles vectors to the
/// the one given and returns the corresponding jacobian of that nearest 
/// neighbor.
/// \param q the current joint angles vector.
/// \return the nearest neighbour jacobian matrix.
// -----------------------------------------------------------------------------
ReturnMatrix QRIO_Controller::getNearestNeighbourJacobian(
	const ColumnVector & q) const
{
	TRY(QRIO_Controller::getNearestNeighbourJacobian);
	vector< ColumnVector >::const_iterator it = 
		min_element(qVecs_.begin(), qVecs_.end(), SqrDistLessThan(q));
	uint minSqrDistPos = it - qVecs_.begin();
	return jacobians_[minSqrDistPos];
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::storeJacobianWithJointAngles
/// \brief Stores the given joint angles vector and its corresponding jacobian.
/// Details: If an the given joint angles vector already exists, replaces its
/// Jacobian with the given one.
/// \param q the current joint angles vector.
/// \param J the Jacobian to update with.
/// \return the nearest neighbour jacobian matrix.
// -----------------------------------------------------------------------------
void QRIO_Controller::storeJacobianWithJointAngles(const ColumnVector & q,
												   const Matrix & J)
{
	TRY(QRIO_Controller::updateNearestNeighbourJacobian);
	vector< ColumnVector >::const_iterator it = 
		find(qVecs_.begin(), qVecs_.end(), q);
	if (it == qVecs_.end())	{	// couldn't find given q
		qVecs_.push_back(q);
		jacobians_.push_back(J);
	} else {					// given q already exists
		uint pos = it - qVecs_.begin();
		jacobians_[pos] = J;
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::computeJacobian
/// \brief Computes the Jacobian relating joint angles offsets to ZMP offsets.
/// \param dZMP the current ZMP offset vector.
/// \param dq the current joint angles offset vector.
/// \param redundants the current effective redundant joints indices.
/// \return the jacobian matrix.
// -----------------------------------------------------------------------------
ReturnMatrix QRIO_Controller::computeJacobian(const ColumnVector & dZMP, 
											  const ColumnVector & dq,
											  const vector< int > & redundants) const
{
	TRY(QRIO_Controller::computeJacobian);
	int redundantNum = redundants.size();
	Matrix J(2, redundantNum);
	for (int k = 0; k < redundantNum; k++) {
		J.Column(k + 1) = dZMP / dq(redundants[k]);
	}
	J.Release();
	return J;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::getDynamicsParameters
/// \brief Prepare all dynamics parameters in Matrix notation for easy 
/// computation of CoM or ZMP.
/// \param masses Output column vector of all robot's links masses.
/// \param pos    Output 3xN matrix of all robot's links absolute positions.
/// \param accel  Output 3xN matrix of all robot's links absolute accelerations.
/// \param angV   Output 3xN matrix of all robot's links angular velocities.
/// \param inertias Output 9xN matrix of all robot's links inertia tensors.
// -----------------------------------------------------------------------------
void QRIO_Controller::getDynamicsParameters(ColumnVector & masses, Matrix & pos, 
											Matrix & accel, Matrix & angV, 
											Matrix & inertias) const
{
	TRY(QRIO_Controller::getDynamicsParameters);
	uint linksNum = uint(masses.Nrows());
	for (uint i = 0; i < linksNum; i++) {
		masses(i + 1) = links_[i].m();
 		pos.Column(i + 1) = links_[i].p();
		angV.Column(i + 1) = links_[i].w();
		accel.Column(i + 1) = links_[i].a();
		inertias.Column(i + 1) = links_[i].I().AsColumn();
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::getRobotCurrCOMAndZMP
/// \brief Computes the current CoM and ZMP of the robot by retreiving the 
/// current dynamics and then computing ZMP based on the current dynamics.
/// \param com The robot's current CoM (in 3D).
/// \param zmp The robot's current ZMP (in 3D).
// -----------------------------------------------------------------------------
void QRIO_Controller::getRobotCurrCOMAndZMP(Point_3 & com, Point_3 & zmp)
{
	TRY(QRIO_Controller::getRobotCurrCOMAndZMP);
	// Read the dynamics information from shared memory:
	readLinkDynamics();
	
	// Masses, positions, accelerations, angular velocities and inertias:
	uint linkNo = links_.size();
	ColumnVector M(linkNo);
	Matrix P(3, linkNo), A(3, linkNo), W(3, linkNo), I(9, linkNo);

	// Prepare the parameters in matrix form:
	getDynamicsParameters(M, P, A, W, I);

	// Compute ZMP:
	com = getCOM(M, P);
	zmp = getZMP(M, P, A, groundHeight, I, W);
// 	zmp = getZMP(M, P, A, GROUND_HEIGHT, I, W);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::getCurrRedundantJointsOffsets
/// \brief Gets the current redundant joint angles offsets.
/// \param t The current simulation step (in sec).
/// \param dq The output redundant joint angles offsets.
/// \param redundants The output effective redundant joints indices.
// -----------------------------------------------------------------------------
void QRIO_Controller::getCurrRedundantJointsOffsets(double t, ColumnVector & dq,
													vector< int > & redundants) const
{
	TRY(QRIO_Controller::getCurrRedundantJointsOffsets);
	const double A = 0.005;		// amplitude
	
	dq = 0.0;
	// Update only the redundant joint offsets:
	dq(torso_pitch + 1) = A;	// for torso_pitch joint
	dq(torso_roll + 1) = A;	// for torso_roll joint
// 	dq(left_hip_roll + 1) = A;	// for left_hip_roll joint
// 	dq(right_hip_roll + 1) = A;	// for right_hip_roll joint
	
	redundants.clear();
	redundants.push_back(torso_pitch + 1);
	redundants.push_back(torso_roll + 1);
// 	redundants.push_back(left_hip_roll + 1);
// 	redundants.push_back(right_hip_roll + 1);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::getCorrectionJointAngles
/// \brief Gets the joint correction angles offsets given the ZMP correction 
/// and the Jacobian.
/// \param J The Jacobian.
/// \param dZMP The ZMP desired correction.
/// \return The joints' correction angles offsets vector.
// -----------------------------------------------------------------------------
ColumnVector QRIO_Controller::getCorrectionJointAngles(
	const Matrix & J, const ColumnVector & dZMP) const
{
	TRY(QRIO_Controller::getCorrectionJointAngles);
	ColumnVector dq(jointsNum_);
	dq = 0.0;

	// Compute the jacobian's pseudo-inverse:
	Matrix pinvJ = getPseudoInverse(J);
	ColumnVector dqRed = pinvJ * dZMP; // only redundant joints
	
	// Update only the redundant joint offsets:
	dq(torso_pitch + 1) = dqRed(1); // for torso_pitch joint
	dq(torso_roll + 1) = dqRed(2);	// for torso_roll joint
// 	dq(left_hip_roll + 1) = dqRed(3); // for left_hip_roll joint
// 	dq(right_hip_roll + 1) = dqRed(4);	// for right_hip_roll joint

    // Adjust the corrections not to exceed joint limits:
	adjustCorrectionJointAngles(dq);

	dq.Release();
	return dq;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::adjustCorrectionJointAngles
/// \brief Adjusts the joint correction angles offsets so that each will not 
/// be over the joint's limits.
/// \param dq the given joint angles offset vector to adjust. 
// -----------------------------------------------------------------------------
void QRIO_Controller::adjustCorrectionJointAngles(ColumnVector & dq) const
{
	TRY(QRIO_Controller::adjustCorrectionJointAngles);
	QRIO_JointInfo jInfo;
	for (int i = 1; i <= dq.Nrows(); i++) {
		float maxLimit = jInfo.getMaxLimit(i - 1);
		float minLimit = jInfo.getMinLimit(i - 1);
		if (dq(i) > maxLimit) {
			dq(i) = maxLimit;
		} else if (dq(i) < minLimit) {
			dq(i) = minLimit;
		}
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::getJointTrajectories
/// \brief Creates a motion trajectory object based on the given motion type 
/// (see biped_trajectory).
/// \param motionSeq the type of motion represented as a string.
/// \param currTime current simulation time step.
/// \param motionTimeInterval the time interval between each control point.
// -----------------------------------------------------------------------------
void QRIO_Controller::getJointTrajectories(string motionSeq, double currTime)
{
	TRY(QRIO_Controller::getJointTrajectories);
	cout << "QRIO: Retrieving motion sequence " << motionSeq << endl;
	MotionSequencesMap::const_iterator motionIter = 
		motionSequencesMap_.find(motionSeq);
	if (motionIter == motionSequencesMap_.end()) 
		THROW_FATAL("The given motion does not exist: " <<  motionSeq);
	
	Matrix positions = motionIter->second;
	uint posNum = uint(positions.ncols());

	Matrix pts(jointsNum_ + 1, posNum + 1); pts = 0.0;
	pts.SubMatrix(1, jointsNum_ + 1, 2, posNum + 1) = positions;
	pts.Row(1) += currTime;
	pts.SubMatrix(2, jointsNum_ + 1, 1, 1) = getJointAngles();
	
	currMotionTraj_.setPoints(pts);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::setMotionTrajectories
/// \brief Sets the motion trajectory object to a given motion type.
/// \param type Type of motion.
/// \param trajectories motion trajectory object.
// -----------------------------------------------------------------------------
void QRIO_Controller::setMotionTrajectories(MotionType type, double currTime)
{
	TRY(QRIO_Controller::setMotionTrajectories);
	string motionStr;
	switch (type) {
		case STAND_FACEUP:    motionStr = "STAND_FACEUP";    break;
		case STAND_FACEDOWN:  motionStr = "STAND_FACEDOWN";  break;
		case TURN_RIGHT:      motionStr = "TURN_RIGHT";      break;
		case TURN_LEFT:       motionStr = "TURN_LEFT";       break;
		case SIDESTEP_RIGHT:  motionStr = "SIDESTEP_RIGHT";  break;
		case SIDESTEP_LEFT:   motionStr = "SIDESTEP_LEFT";   break;
		case WALK_LEFT:       motionStr = "WALK_LEFT";       break;
		case WALK_RIGHT:      motionStr = "WALK_RIGHT";      break;
		case FIRST_STEP:      motionStr = "FIRST_STEP";	     break;
		case LAST_STEP_LEFT:  motionStr = "LAST_STEP_LEFT";  break;
		case LAST_STEP_RIGHT: motionStr = "LAST_STEP_RIGHT"; break;
		default:
			THROW_FATAL("Undefined type of motion: " << type);
	}
	getJointTrajectories(motionStr, currTime);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::updateMotionTrajectories
/// \brief Checks if current motion has finished, if so, sets the motion 
/// trajectories object with the next motion. 
/// \param t The current simulation step (in sec).
/// \param q column vector containing all joint angles for the robot.
/// \return TRUE if a new motion was updated.
// -----------------------------------------------------------------------------
bool QRIO_Controller::updateMotionTrajectories(double t)
{
	TRY(QRIO_Controller::updateMotionTrajectories);
	if (t == 0) 
		currMotionTraj_.reset();

	double endTime = currMotionTraj_.getEndTime();
	// If the current motion has ended so set the next one:
	if (endTime <= t) {			
		if (t <= 0) {
			currMotionType_ = FIRST_STEP;
		} else if (currMotionType_ == FIRST_STEP || 
				   currMotionType_ == WALK_RIGHT) {
			currMotionType_ = WALK_LEFT;
		} else {
			currMotionType_ = WALK_RIGHT;
		}
		cout << "QRIO: Updating motion trajectories object ... " << endl;
		setMotionTrajectories(currMotionType_, t);
		return true;
	}
	return false;
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::correctMotionTrajectories
/// \brief Corrects the motion's trajectories object according to the internal
/// correction map.
// -----------------------------------------------------------------------------
void QRIO_Controller::correctMotionTrajectories()
{
	TRY(QRIO_Controller::correctMotionTrajectories);
	cout << "QRIO: Correcting motion trajectories object ..." << endl;
	double beginTime = currMotionTraj_.getBeginTime();
	double endTime = currMotionTraj_.getEndTime();
	TimeVectorMap::const_iterator lowerIt = dqCorrectMap_.lower_bound(beginTime);
	TimeVectorMap::const_iterator upperIt = dqCorrectMap_.upper_bound(endTime);
	if (lowerIt != dqCorrectMap_.end() && lowerIt != upperIt) {
		int N = dqCorrectMap_.size();
		Matrix correctionQ(jointsNum_, N); correctionQ = 0.0;
		RowVector correctionT(N); correctionT = 0.0;
		int i = 1;
		for (TimeVectorMap::const_iterator it = lowerIt; it != upperIt; it++) {
			correctionT(i) = it->first;
			correctionQ.Column(i) = it->second;
			i++;
		}
		currMotionTraj_.adjust(correctionT.Columns(1, i - 1), 
							   correctionQ.Columns(1, i - 1));
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::initSharedMem
/// \brief Initialize the IPC to the physics library.
/// Details: Opens a shared memory segment (already created by qrio's physics) 
/// and attaches it to a local pointer. In addition, opens the mutual semaphore
/// for safe access to the shared memory.
// -----------------------------------------------------------------------------
void QRIO_Controller::initSharedMem()
{
	TRY(QRIO_Controller::initSharedMem);
	cout << "QRIO: Initializing shared memory and semaphores" << endl;
	// 	Get the shared memory segments and attach them:
	if (shmDynamics.openAndAttach(DYN_SHM_KEY, sizeof(DynamicsInfo)) == -1)
		THROW_FATAL("Couldn't open the dynamics shared memory segment!");
	if (shmFootGround.openAndAttach(FOOT_SHM_KEY, sizeof(FootGroundInfo)) == -1)
		THROW_FATAL("Couldn't open the foot-ground shared memory segment!");

	dynShmPtr = (DynamicsInfo *) shmDynamics.getSegmentPtr();
	footShmPtr = (FootGroundInfo *) shmFootGround.getSegmentPtr();
	
	// Open the shared memory semaphores. 
	// The qrio physics must have created them already:
	if (semDynamics.open(DYN_SEM_KEY, Semaphore::SEM_OPEN) < 0)
		THROW_FATAL("QRIO: Can't open dynamics shared memory semaphore");
	if (semFootGround.open(FOOT_SEM_KEY, Semaphore::SEM_OPEN) < 0)
		THROW_FATAL("QRIO: Can't open foot-ground shared memory semaphore");
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::finalizeSharedMem
/// \brief Detaches the shared memory segment and removes it. Also, closes the 
/// mutual semaphore for shared memory access.
// -----------------------------------------------------------------------------
void QRIO_Controller::finalizeSharedMem()
{
	TRY(QRIO_Controller::finalizeSharedMem);
	cout << "QRIO: Finalizing shared memory and semaphores" << endl;
	// Detach the shared memory segments:
	if (shmDynamics.detach() < 0)
		THROW_FATAL("QRIO: Can't detach dynamics shared memory segment");
	if (shmFootGround.detach() < 0)
		THROW_FATAL("QRIO: Can't detach foot-ground shared memory segment");

	// Close shared memory semaphores:
	semDynamics.close();
	semFootGround.close();
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::plotData
/// \brief Plots a given data in a time-vector format.
// -----------------------------------------------------------------------------
void QRIO_Controller::plotData(
	GNUPlot & gplot, const string & prefix, const TimeVectorMap & map, 
	vector< int > inx) const
{
	TRY(QRIO_Controller::plotData);
	QRIO_JointInfo jInfo;
	vector< double > times;
	uint plotsNum = inx.size();
	vector< vector< double > > vals(plotsNum);
	for (TimeVectorMap::const_iterator it = map.begin(); it != map.end(); it++) {
		times.push_back(it->first);
		for (uint i = 0; i < plotsNum; i++) 
			vals[i].push_back((it->second)(inx[i]));
	}
	for (uint i = 0; i < plotsNum; i++) {
		ostringstream title;
		title << prefix << jInfo.toNameString(inx[i] - 1);
		gplot.plotXYData(times, vals[i], title.str(), title.str());
		sleep(1);
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::plotData
/// \brief Plots a given data in a time-vector format.
// -----------------------------------------------------------------------------
void QRIO_Controller::plotData(
	GNUPlot & gplot, const string & prefix, const TimePointMap & map) const
{
	TRY(QRIO_Controller::plotData);
	vector< double > times;
	uint plotsNum = 3;
	vector< vector< double > > vals(plotsNum);
	for (TimePointMap::const_iterator it = map.begin(); it != map.end(); it++) {
		times.push_back(it->first);
		vals[0].push_back((it->second).x());
		vals[1].push_back((it->second).y());
		vals[2].push_back((it->second).z());
	}
	gplot.plotXYData(times, vals[0], prefix + "x", prefix + "x");
	sleep(1);
	gplot.plotXYData(times, vals[1], prefix + "y", prefix + "y");
	sleep(1);
	gplot.plotXYData(times, vals[2], prefix + "z", prefix + "z");
	sleep(1);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::pollQuit
/// \brief Waits for the user to signal quit.
// -----------------------------------------------------------------------------
void QRIO_Controller::pollQuit()
{
	TRY(QRIO_Controller::pollQuit);
	robot_keyboard_enable(TIME_STEP / 4);
	cout << "Press any key to quit ... " << endl;
	do {
		robotSimulationStep(1);
	} while (robot_keyboard_get_key() == 0);
	robot_keyboard_disable();
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::readLinkDynamics
/// \brief Read the robot's links current dynamics from the shared memory and
/// update in the internal links vector.
// -----------------------------------------------------------------------------
void QRIO_Controller::readLinkDynamics()
{
	TRY(QRIO_Controller::readLinkDynamics);
	if (semDynamics.acquire() < 0) // Lock the shared memory semaphore
		THROW_ERROR("Cannot acquire the foot-ground semaphore!");

	for (uint i = 0; i < NUM_RIGID_BODIES; i++) {
		links_[i].w() << dynShmPtr->angVel[i];
		links_[i].p() << dynShmPtr->bodyPos[i];
	}

	if (semDynamics.release() < 0) // Unlock the shared memory semaphore 
		THROW_ERROR("Cannot release foot-ground semaphore!");
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::readLinkConstDynamics
/// \brief Read the robot's links constant dynamics from the shared memory and
/// update in the internal links vector.
// -----------------------------------------------------------------------------
void QRIO_Controller::readLinkConstDynamics()
{
	TRY(QRIO_Controller::readLinkConstDynamics);
	if (semDynamics.acquire() < 0) // Lock the shared memory semaphore
		THROW_ERROR("Cannot acquire the foot-ground semaphore!");

	for (uint i = 0; i < NUM_RIGID_BODIES; i++) {
		links_[i].m() = dynShmPtr->mass[i];
		links_[i].r() << dynShmPtr->massPos[i];
		const dMatrix3 & shmInertia = dynShmPtr->inertia[i];
		// NOTE: Apparently the inertia comes out unsymmetrical, so here
		// we fix the order:
		links_[i].I() << shmInertia[0] << shmInertia[1] << shmInertia[2]
					  << shmInertia[4] << shmInertia[5] << shmInertia[3]
					  << shmInertia[8] << shmInertia[7] << shmInertia[6];
	}

	if (semDynamics.release() < 0) // Unlock the shared memory semaphore 
		THROW_ERROR("Cannot release foot-ground semaphore!");
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::readContactPoints
/// \brief Read the feet-gound contact points from the shared memory and
/// update in the internal contact points vector.
// -----------------------------------------------------------------------------
void QRIO_Controller::readContactPoints()
{
	TRY(QRIO_Controller::readContactPoints);
	if (semFootGround.acquire() < 0) // Lock the shared memory semaphore
		THROW_ERROR("Cannot acquire the foot-ground semaphore!");

	contactPts_.clear();
	for (uint k = 0; k < 2 * CONTACT_NUM; k++) {
		dReal x = footShmPtr->footGroundContacts[k][0];
		dReal y = footShmPtr->footGroundContacts[k][1];
		dReal z = footShmPtr->footGroundContacts[k][2];
		if (x != 0 || y != 0 || z != 0)
			contactPts_.push_back(Point_3(x, y, z));
	}

	if (semFootGround.release() < 0) // Unlock the shared memory semaphore 
		THROW_ERROR("Cannot release foot-ground semaphore!");
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::readJacobianData
/// \brief Read the jacobians input file and initialize the internal jacobians 
/// data structure.
/// \param filename The input filename to read from.
// -----------------------------------------------------------------------------
void QRIO_Controller::readJacobianData(const string & filename)
{
	TRY(QRIO_Controller::readJacobianData);
	cout << "QRIO: Reading jacobians input file" << endl;
	MatrixFileIO jacobsInputFile(filename);
	vector< Matrix > data;
	vector< string > titles;
	jacobsInputFile.read(data, titles);
	qVecs_.clear();
	jacobians_.clear();
	for (uint i = 0; i < data.size(); i++) {
		if (titles[i].find('J') != string::npos) 
			jacobians_.push_back(data[i]);
		else if (titles[i].find('q') != string::npos) 
			qVecs_.push_back(data[i]);
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::writeJacobianData
/// \brief Write the joint angles vectors and their corresponding jacobians to 
/// the jacobians output file.
/// \param filename The output filename to write to.
// -----------------------------------------------------------------------------
void QRIO_Controller::writeJacobianData(const string & filename) const 
{
	TRY(QRIO_Controller::writeJacobianData);
	cout << "QRIO: Writing to jacobians output file" << endl;
	MatrixFileIO jacobsOutputFile(filename);
	uint qVecSize = qVecs_.size(), jacobiansSize = jacobians_.size();
	if (qVecSize == 0 || jacobiansSize == 0) {
		cout << "QRIO: WARNING: No q's or jacobians to write!" << endl;
		return;
	}
	BIPED_ASSERT(qVecSize == jacobiansSize);

	// Prepare data and its corresponding titles:
	vector< string > titles;
	vector< Matrix > data;
	for (uint i = 0; i < qVecSize; i++) {
		ostringstream intStream;
		intStream << i;
		titles.push_back("q" + intStream.str());
		data.push_back(qVecs_[i]);
		titles.push_back("J" + intStream.str());
		data.push_back(jacobians_[i]);
	}
	// Write the data to the file:
	jacobsOutputFile.write(data, titles);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::readCorrectionMap
/// \brief Read the correction map input file and initialize the internal data
/// structure.
/// \param filename The input filename to read from.
// -----------------------------------------------------------------------------
void QRIO_Controller::readCorrectionMap(const string & filename)
{
	TRY(QRIO_Controller::readCorrectionMap);
	cout << "QRIO: Reading correction map input file" << endl;
	MatrixFileIO correctionInputFile(filename);

	vector< Matrix > data;
	vector< string > titles;
	correctionInputFile.read(data, titles);
	dqCorrectMap_.clear();
	for (uint i = 0; i < data.size(); i++) {
		double t = atof(titles[i].substr(2).c_str());
		dqCorrectMap_[t] = data[i];
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::writeCorrectionMap
/// \brief Write the correction map to the correction map output file.
/// \param filename The output filename to write to.
// -----------------------------------------------------------------------------
void QRIO_Controller::writeCorrectionMap(const string & filename) const
{
	TRY(QRIO_Controller::writeCorrectionMap);
	cout << "QRIO: Writing to correction map output file" << endl;
	MatrixFileIO correctionOutputFile(filename);

	if (dqCorrectMap_.size() == 0) {
		cout << "QRIO: WARNING: Correction map is empty - not writing!" << endl;
		return;
	}

	vector< string > titles;
	vector< Matrix > data;
	TimeVectorMap::const_iterator it = dqCorrectMap_.begin();
	for (; it != dqCorrectMap_.end(); it++) {
		ostringstream timeStream;
		timeStream << it->first;
		titles.push_back("dq" + timeStream.str());
		data.push_back(it->second);
	}
	
	correctionOutputFile.write(data, titles);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::writeFzmpData
/// \brief Writes the occurences of the ZMP being outside the support-region
/// to a file.
/// \param filename The output filename to write to.
// -----------------------------------------------------------------------------
void QRIO_Controller::writeFzmpData(const string & filename) const
{
	TRY(QRIO_Controller::writeFzmpData);
	cout << "QRIO: Writing to FZMP occurences output file" << endl;
	uint fzmpNum = fzmpTimes_.size();
	if (fzmpNum == 0) {
		cout << "QRIO: WARNING: FZMP vector is empty - not writing!" << endl;
		return;
	}

	ofstream fzmpOutFile(filename.c_str());
	if (!fzmpOutFile) 
		THROW_FATAL("Couldn't open file " << filename);

	for (uint i = 0; i < fzmpNum; i++) {
		double t = fzmpTimes_[i];
		TimePointMap::const_iterator findIt = zmpMap_.find(t);
		BIPED_ASSERT(findIt != zmpMap_.end());
		fzmpOutFile << "t= " << t << " : " << findIt->second << endl;
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::writeZMPDotsAndQDots
/// \param filename The output filename to write to.
// -----------------------------------------------------------------------------
void QRIO_Controller::writeZMPDotsAndQDots(const string & filename) const
{
	TRY(QRIO_Controller::writeZMPDotsAndQDots);
	cout << "QRIO: Writing to ZMP dot and q dot output file" << endl;
	uint size = q_dots.size();
	BIPED_ASSERT(size == zmp_dots.size());
	if (size == 0) {
		cout << "QRIO: WARNING: q_dots vector is empty - not writing!" << endl;
		return;
	}

	ofstream outFile(filename.c_str());
	if (!outFile) 
		THROW_FATAL("Couldn't open file " << filename);

	for (uint i = 0; i < size; i++) {
		double q_dot = q_dots[i];
		ColumnVector zmp_dot = zmp_dots[i];
		outFile << "q_dot= " << q_dot << " ; zmp_dot=" << zmp_dot.AsRow() << endl;
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::readPersistentData
/// \brief Read any persistent data required for the controller. 
// -----------------------------------------------------------------------------
void QRIO_Controller::readPersistentData()
{
	TRY(QRIO_Controller::readPersistentData);
	// Read motion sequences XML file:
	readMotionXML(MOTIONS_XML_FILENAME);

	if (ctype_ == COMPENSATE) { 
		// Read input jacobians file:
		readJacobianData(INPUT_JACOBIANS_FILENAME);
	}
	if (ctype_ == ACTUAL) {
		// Read input correction map file:
		readCorrectionMap(INPUT_CMAP_FILENAME);
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::writePersistentData
/// \brief Write any persistent data required for saving. 
// -----------------------------------------------------------------------------
void QRIO_Controller::writePersistentData()
{
	TRY(QRIO_Controller::writePersistentData);
	if (ctype_ != ACTUAL) {
		// Write output jacobians file:
		writeJacobianData(OUTPUT_JACOBIANS_FILENAME);
	}
	if (ctype_ == COMPENSATE) {
		// Write output correction map file:
		writeCorrectionMap(OUTPUT_CMAP_FILENAME);
	}
// 	if (ctype_ == LEARNING) {
// 		writeZMPDotsAndQDots("data/ZMP_and_Q_dots.dat");
// 	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::initialize
/// \brief Handle any initialization for the QRIO_Controller object, i.e. allocate 
/// memory, open shared memory, open semaphores etc.
// -----------------------------------------------------------------------------
void QRIO_Controller::initialize()
{
	TRY(QRIO_Controller::initialize);
	// Initialize shared memory stuff (segments, semaphores etc):
	initSharedMem();

	// Read any persistent data:
	readPersistentData();
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: QRIO_Controller::finalize
/// \brief Handle any cleanup for the QRIO_Controller object, i.e. cleanup memory, 
/// close shared memory, close semaphores etc.
// -----------------------------------------------------------------------------
void QRIO_Controller::finalize()
{
	TRY(QRIO_Controller::finalize);
	// Finalize shared memory stuff (segments, semaphores etc):
	finalizeSharedMem();

	// Write any persistent data:
	writePersistentData();
	CATCH;
}
