/* -*- c++ -*- */
/****************************************************************
 *                                                              *
 *  Project name:      Biped Locomotion                         *
 *                                                              *
 *  Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)           *
 ****************************************************************
 * $Id: qrio_controller.hpp,v 1.9 2004/08/29 11:09:54 s0343981 Exp $ 
 ****************************************************************/
/**
 * \file qrio_controller.hpp 
 * \brief class QRIO_Controller definition.
 */

#ifndef __QRIO_CONTROLLER_HPP__
#define __QRIO_CONTROLLER_HPP__

#include <biped_controller.hpp>
#include <biped_gnuplot.hpp>
#include <biped_namespace.hpp>
#include <biped_trajectory.hpp>

#define GROUND_HEIGHT             0.2
#define COMPENSATE_FACTOR         0.05
#define CORRECTION_OFFSET_LIMIT   0.01
#define WALK_CYCLES_NUM           5
#define MOTIONS_XML_FILENAME      "data/Motions.curr.xml"
#define INPUT_JACOBIANS_FILENAME  "data/Jacob_input.dat"
#define OUTPUT_JACOBIANS_FILENAME "data/Jacob_output.dat"
#define INPUT_CMAP_FILENAME       "data/CMap_input.dat"
#define OUTPUT_CMAP_FILENAME      "data/CMap_output.dat"
#define OUTPUT_FZMP_FILENAME      "data/FZMP_output.dat"

BIPED_BEGIN_NAMESPACE(QRIO);

typedef map< double, ColumnVector > TimeVectorMap;
typedef map< double, Point_3 > TimePointMap;

// =============================================================================
//  Class: QRIO_Controller
/// \brief C++ definition of a QRIO specific controller which abides with 
/// the generic biped controller interface.
// =============================================================================
class QRIO_Controller : public Controller {
public:							// definitions
	typedef enum { REGULAR, LEARNING, COMPENSATE, ACTUAL } ControlType;

	typedef enum { STAND_FACEUP, STAND_FACEDOWN, TURN_RIGHT, TURN_LEFT,
				   SIDESTEP_RIGHT, SIDESTEP_LEFT, WALK_LEFT, WALK_RIGHT,
				   FIRST_STEP, LAST_STEP_LEFT, LAST_STEP_RIGHT } MotionType;

public:							// methods
	QRIO_Controller(ControlType ctype, uint jNum, int lt, int rt, int range, 
					int gps, DeviceTag * joints, DeviceTag camera, 
					DeviceTag emitter);
	~QRIO_Controller() { finalize(); }
	
	void run();
	void regularRun();
	void jiggleRun();
	uint correctionRun();
	void learnRun();
	void compensateRun();
	void compensateRun2();
	void actualRun();

private:						// methods
	// Balance control related:
	void updateAnklesCompliance();
	ReturnMatrix getNearestNeighbourJacobian(const ColumnVector & q) const;
	void storeJacobianWithJointAngles(const ColumnVector & q,
									  const Matrix & J);
	ReturnMatrix computeJacobian(const ColumnVector & dZMP, 
								 const ColumnVector & dq,
								 const vector< int > & redundants) const;
	void getDynamicsParameters(ColumnVector & masses, Matrix & pos, 
							   Matrix & accel, Matrix & angV, 
							   Matrix & inertias) const;
	void getRobotCurrCOMAndZMP(Point_3 & com, Point_3 & zmp);

	// Redundant joints management:
	void getCurrRedundantJointsOffsets(double t, ColumnVector & dq, 
									   vector< int > & redundants) const;
	ColumnVector getCorrectionJointAngles(const Matrix & J, 
										  const ColumnVector & dZMP) const;
	void adjustCorrectionJointAngles(ColumnVector & dq) const;
	
	// Motion control:
	double robotSimulationStep(float stepNum) 
	{ 
		robot_step(uint(TIME_STEP * stepNum));
		return stepNum * TIME_STEP_SEC;	// the accumulated real-time performed
	}
	void initialStep();
	bool compensateStep(double & t, bool & isZMPOut);
	bool regularStep(bool doCorrect, double & t, bool & isZMPOut);
	void jiggleAllStep(double t);
	void jiggleIndependentStep(double t);
	
	void getJointTrajectories(string motionSeq, double currTime);
	void setMotionTrajectories(MotionType type, double currTime);
	bool updateMotionTrajectories(double t);
	void correctMotionTrajectories();

	// Update the physical information from the physics:
	void readLinkDynamics();
	void readLinkConstDynamics();
	void readContactPoints();
	
	// Shared memory management:
	void initSharedMem();
	void finalizeSharedMem();
	
	// Visualization:
	void plotData(GNUPlot & gplot, const string & prefix, 
				  const TimeVectorMap & map, vector< int > inx) const;
	void plotData(GNUPlot & gplot, const string & prefix, 
				  const TimePointMap & map) const;

	void pollQuit();

	// Read/Write data management:
	void readJacobianData(const string & filename);
	void writeJacobianData(const string & filename) const;
	void readCorrectionMap(const string & filename);
	void writeCorrectionMap(const string & filename) const;
	void writeFzmpData(const string & filename) const;
	void writeZMPDotsAndQDots(const string & filename) const;
	void readPersistentData();
	void writePersistentData();

	// Initialization and finalization methods:
	void initialize();
	void finalize();

private:						// members
	ControlType ctype_;
	Point3Vec contactPts_;
	vector< Link > links_;

	SimpleTrajectory currMotionTraj_; ///< the current trajectories object
// 	InterpTrajectory< BezierCubicSpline > currMotionTraj_;
	MotionType currMotionType_; ///< the current motion type being used

	vector< ColumnVector > qVecs_; ///< storage for joint angles vectors
	vector< Matrix > jacobians_; ///< storage for jacobian matrices
	TimeVectorMap dqCorrectMap_; ///< Maps t --> dq corrections
	TimeVectorMap qMap_;		///< Maps t --> q's of joints
	TimePointMap zmpMap_;		///< Maps t --> ZMP
	TimePointMap comMap_;		///< Maps t --> CoM
	vector< double > fzmpTimes_; ///< stores all time occurences ZMP is outside
};
		
BIPED_END_NAMESPACE;

#endif	// __QRIO_CONTROLLER_HPP__
