/* -*- C++ -*- */
/****************************************************************
 *                                                              *
 *   Project name:      Biped Locomotion                        *
 *                                                              *
 *   Created by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)          *
 ****************************************************************
 * $Id: hoap2_physics.cpp,v 1.5 2004/08/06 18:03:33 s0343981 Exp $
 ****************************************************************/
/**
 * \file: hoap2_physics.cpp
 * \brief Implementations related to ODE and physics.
 */

// CGAL includes:
#include <CGAL/Cartesian.h>
#include <CGAL/convex_hull_2.h>
#include <CGAL/Convex_hull_projective_xz_traits_2.h>
#include <CGAL/functional.h>
// C++ includes:
#include <algorithm>
#include <errno.h>
#include <iostream>
#include <iomanip>
#include <cstdlib>
#include <vector>
// Webots includes:
#include <physics/physics.h>
// ODE includes:
#include <objects.h>
#include <joint.h>
#include <collision_kernel.h>
// OpenGL includes:
#include <GL/gl.h>
// BIPED includes:
#include <biped_exception.hpp>
#include <biped_logging.hpp>
#include <biped_sem.hpp>
#include <biped_shmem.hpp>
// HOAP2 includes:
#include <hoap2/hoap2_joints.hpp>
#include <hoap2/hoap2_shmem.hpp>
#include "hoap2_physics.hpp"

#define PRINT_BODY(def, gid, bid) { cout << def << " IDs: geomID = " << gid	\
										 << "; bodyID = " << bid << endl; }
#define ASSIGN_VECTOR3(lhs, rhs)										\
	{ (lhs)[0] = (rhs)[0]; (lhs)[1] = (rhs)[1]; (lhs)[2] = (rhs)[2]; }
#define ASSIGN_VECTOR4(lhs, rhs) {					\
		(lhs)[0] = (rhs)[0]; (lhs)[1] = (rhs)[1];	\
		(lhs)[2] = (rhs)[2]; (lhs)[3] = (rhs)[3]; }

using namespace std;
using namespace HOAP2;

static dWorldID worldID = NULL;
static dSpaceID spaceID = NULL;
static dJointGroupID contactJointGroupID = NULL;

static dGeomID robotGeomID = NULL;
static dBodyID robotBodyID = NULL;
static dGeomID leftSoleGeomID = NULL;
static dGeomID rightSoleGeomID = NULL;
static dBodyID leftSoleBodyID = NULL;
static dBodyID rightSoleBodyID = NULL;
static dGeomID floorGeomID = NULL;

// Shared memory stuff:
static SharedMemory shmDynamics;
static SharedMemory shmFootGround;
static DynamicsInfo * dynShm;	// dynamicsInfo struct ptr in shared memory
static FootGroundInfo * footShm; // footGroundInfo struct ptr in shared memory
static Semaphore semDynamics;
static Semaphore semFootGround;

static dMass linkMasses[NUM_RIGID_BODIES];
static dVector3 linkAngVel[NUM_RIGID_BODIES];
static dVector3 linkPos[NUM_RIGID_BODIES];
static dVector3 linkAcc[NUM_RIGID_BODIES];
static dVector3 linkLastLinVel[NUM_RIGID_BODIES];

typedef CGAL::Point_3< CGAL::Cartesian< dReal > > Point_3;
static Point_3 soleContacts[2 * CONTACT_NUM];
static HOAP2_JointToStr j2str;

// -----------------------------------------------------------------------------
//  Function: set_pointers_from_DEF
/// \brief Sets the geometry and rigid body pointers from a given DEF string.
/// \param geomID The geometry pointer to set.
/// \param bodyID The rigid body pointer to set.
/// \param DEF The given VRML DEF string.
// -----------------------------------------------------------------------------
void set_pointers_from_DEF(dGeomID & geomID, dBodyID & bodyID, 
						   const char * DEF)
{ 
	TRY(set_pointers_from_DEF);
	geomID = dWebotsGetGeomFromDEF(DEF);
	if (geomID) {
		if (dGeomIsSpace(geomID))
			bodyID = dGeomGetBody(dSpaceGetGeom((dSpaceID) geomID, 0));
		else 
			bodyID = dGeomGetBody(geomID);
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: get_foot_ground_contact_points
/// \brief Access the current foot-ground contact points and store in the static
/// global variable to prepare for the update in the shared memory. 
// -----------------------------------------------------------------------------
void get_foot_ground_contact_points(dGeomID soleGeomID)
{
	TRY(get_foot_ground_contact_points);
	dContact contactPts[CONTACT_NUM];
	uint contactNum = dCollide(floorGeomID, soleGeomID, CONTACT_NUM, 
							   &contactPts[0].geom, sizeof(dContact));
	
	BIPED_ASSERT(contactNum <= CONTACT_NUM); // to be safe...

	// Determine if left or right sole contacts:
	uint offset = (soleGeomID == rightSoleGeomID) ? CONTACT_NUM : 0;

	// Reset portion of soles contact points appropriately:
	Point_3 zeros(0, 0, 0);
	for (uint i = 0; i < CONTACT_NUM; i++)
		soleContacts[i + offset] = zeros;

	if (contactNum == 0)
		return;

	// Update the appropriate portion of the sole contact points:
	for (uint i = 0; i < contactNum; i++) {
		const dVector3 & pos = contactPts[i].geom.pos;
		soleContacts[i + offset] = Point_3(pos[0], pos[1], pos[2]); 
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: get_dynamics_data
/// \brief Access the current dynamics data and store in the static global
/// variables to prepare for the update in the shared memory. 
// -----------------------------------------------------------------------------
void get_dynamics_data()
{
	TRY(get_dynamics_data);
	const float dt = 0.032; // time interval (32 millisec)
	for (uint i = 0; i < NUM_RIGID_BODIES; i++) {
		const char * bodyDEF = j2str.toDEFString(i);
		dGeomID geomID = dWebotsGetGeomFromDEF(bodyDEF);
		if (geomID) {
			dBodyID bodyID;
			if (dGeomIsSpace(geomID))
				bodyID = dGeomGetBody(dSpaceGetGeom((dSpaceID) geomID, 0));
			else 
				bodyID = dGeomGetBody(geomID);
			// Get angular velocities:
			const dReal * angVel = dBodyGetAngularVel(bodyID);
			ASSIGN_VECTOR3(linkAngVel[i], angVel);
			
			// Get absolute positions:
			const dReal * pos = dBodyGetPosition(bodyID);
			ASSIGN_VECTOR3(linkPos[i], pos);

			// Get linear velocities and accelerations:
			const dReal * V = dBodyGetLinearVel(bodyID);
			const dReal * lastV = linkLastLinVel[i];
			linkAcc[i][0] = (V[0] - lastV[0]) / dt; // acc_x = d Vx / dt
			linkAcc[i][1] = (V[1] - lastV[1]) / dt; // acc_y = d Vy / dt
			linkAcc[i][2] = (V[2] - lastV[2]) / dt; // acc_z = d Vz / dt
			ASSIGN_VECTOR3(linkLastLinVel[i], V);
		}
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: get_dynamics_const_data
/// \brief Access the constant dynamics data and store in the static global
/// variables to prepare for the update in the shared memory. 
// -----------------------------------------------------------------------------
void get_dynamics_const_data()
{
	TRY(get_dynamics_const_data);
	for (uint i = 0; i < NUM_RIGID_BODIES; i++) {
		const char * bodyDEF = j2str.toDEFString(i);
		dGeomID geomID = dWebotsGetGeomFromDEF(bodyDEF);
		if (geomID) {
			dBodyID bodyID;
			if (dGeomIsSpace(geomID))
				bodyID = dGeomGetBody(dSpaceGetGeom((dSpaceID) geomID, 0));
			else 
				bodyID = dGeomGetBody(geomID);
			dBodyGetMass(bodyID, &linkMasses[i]);
		}
	}
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: update_contact_points
/// \brief Updates the foot-ground contacts in the shared memory segment. 
// -----------------------------------------------------------------------------
void update_contact_points()
{
	TRY(update_contact_points);
	if (semFootGround.acquire() < 0) // Lock the shared memory semaphore
		THROW_ERROR("Cannot acquire the foot-ground semaphore!");

	// Updating foot-ground contact points:
	for (uint k = 0; k < 2 * CONTACT_NUM; k++) {
		ASSIGN_VECTOR3(footShm->footGroundContacts[k], soleContacts[k]);
	}

	if (semFootGround.release() < 0) // Unlock the shared memory semaphore 
		THROW_ERROR("Cannot release foot-ground semaphore!");
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: update_links_const_dynamics
/// \brief Updates the constant dynamics information in shared memory segment. 
// -----------------------------------------------------------------------------
void update_links_const_dynamics()
{
	TRY(update_links_const_dynamics);
	if (semDynamics.acquire() < 0) // Lock the shared memory semaphore
		THROW_ERROR("Cannot acquire the dynamics semaphore!");

	// Updating links' information:
	for (int i = 0; i < NUM_RIGID_BODIES; i++) {
		// Updating links' masses:
		dynShm->mass[i] = linkMasses[i].mass;

		// Updating links' mass positions:
		memcpy(dynShm->massPos[i], linkMasses[i].c, sizeof(linkMasses[i].c));

		// Updating links' mass positions:
		memcpy(dynShm->inertia[i], linkMasses[i].I, sizeof(linkMasses[i].I));
	}

	if (semDynamics.release() < 0) // Unlock the shared memory semaphore 
		THROW_ERROR("Cannot release dynamics semaphore!");
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: update_links_dynamics
/// \brief Updates the dynamics information in the shared memory segment. 
// -----------------------------------------------------------------------------
void update_links_dynamics()
{
	TRY(update_links_dynamics);
	if (semDynamics.acquire() < 0) // Lock the shared memory semaphore
		THROW_ERROR("Cannot acquire the dynamics semaphore!");

	// Updating links' information:
	for (int i = 0; i < NUM_RIGID_BODIES; i++) {
		// Updating links' angular velocities:
		ASSIGN_VECTOR3(dynShm->angVel[i], linkAngVel[i]);

		// Updating links' body positions:
		ASSIGN_VECTOR3(dynShm->bodyPos[i], linkPos[i]);

		// Updating links' body accelerations:
		ASSIGN_VECTOR3(dynShm->bodyAcc[i], linkAcc[i]);
	}

	if (semDynamics.release() < 0) // Unlock the shared memory semaphore 
		THROW_ERROR("Cannot release dynamics semaphore!");
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: webots_physics_init
/// \brief This function is called upon initialization of the world by Webots.
/// Details: Provides the hoap2 customized physics shared library with ODE 
/// variables used by the simulation. All these parameters are stored in global
/// variables for further use. This is a good place to get pointers to objects
/// on which we want to control the physics (using dWebotsGetGeomFromDEF).
/// \param dWorldID pointer to the ODE world.
/// \param dSpaceID pointer to the ODE geometry space.
/// \param dJointGroupID pointer to contact joint group used by the simulation.
// -----------------------------------------------------------------------------
void webots_physics_init(dWorldID w, dSpaceID s, dJointGroupID j) 
{
	TRY(webots_physics_init);
	cout << "Initializing hoap2 physics:\nWorldID = " << w << ";dSpaceID = " 
		 << s << ";dContactJointGroupID = " << j << endl;

	worldID = w;
	spaceID = s;
	contactJointGroupID = j;

	// Initialize the shared memory segments:
	if (shmDynamics.openAndAttach(DYN_SHM_KEY, sizeof(DynamicsInfo),
								  SharedMemory::SHM_CREATE) == -1) 
		THROW_FATAL("Couldn't open the dynamics shared memory segment!");
	if (shmFootGround.openAndAttach(FOOT_SHM_KEY, sizeof(FootGroundInfo),
									SharedMemory::SHM_CREATE) == -1) 
		THROW_FATAL("Couldn't open the foot-ground shared memory segment!");

	dynShm = (DynamicsInfo *) shmDynamics.getSegmentPtr();
	footShm = (FootGroundInfo *) shmFootGround.getSegmentPtr();

	// Create shared memory semaphores:
	if (semDynamics.open(DYN_SEM_KEY) < 0)
		THROW_FATAL("Can't create dynamics shared memory semaphore");
	if (semFootGround.open(FOOT_SEM_KEY) < 0)
		THROW_FATAL("Can't create foot-ground shared memory semaphore");

	// Get global pointer to the floor geometry:
	floorGeomID = dWebotsGetGeomFromDEF(FLOOR_DEF);

	// Get global pointers to the robot and soles:
	set_pointers_from_DEF(robotGeomID, robotBodyID, ROBOT_DEF);
	set_pointers_from_DEF(leftSoleGeomID, leftSoleBodyID, LEFT_SOLE_DEF);
	set_pointers_from_DEF(rightSoleGeomID, rightSoleBodyID, RIGHT_SOLE_DEF);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: webots_physics_step
/// \brief Called before every physics simulation step.
/// Details: It can be used to add force and or torques to solids. It can also 
/// be used to test the position and orientation of solids.
// -----------------------------------------------------------------------------
void webots_physics_step() 
{
	TRY(webots_physics_step);
	static bool isFirstCall = true;

	// At the first call update the constant dynamics information. No need to
	// perform this at later calls (the values remain the same):
	if (isFirstCall) {
		get_dynamics_const_data();
		update_links_const_dynamics();
		isFirstCall = false;
	}
	// Retrieve current dynamics and update shared memory:
	get_dynamics_data();
	update_links_dynamics();
	
	// Retrieve current foot-ground contacts and update shared memory:
	get_foot_ground_contact_points(leftSoleGeomID);
	get_foot_ground_contact_points(rightSoleGeomID);
	update_contact_points();
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: webots_physics_collide
/// \brief Called whenever a collision occurs between two objects. 
/// Details: May be called several times for a single simulation step with 
/// different parameters corresponding to different objects. We test whether the
/// two colliding objects passed as arguments correspond to the objects we want
/// to control. Then, we can create the contact joints, using the ODE dCollide
/// and dJointCreateContact functions. Finally, we can add this contact joint
/// to the joint group passed as an argument of the webots_physics_init function
/// using ODE's dJointAttach function. 
/// \param g1 a placeable ODE geometry object.
/// \param g2 another placeable ODE geometry object.
/// \return 1 if the collision has been handled or 0 if we wish that Webots 
/// handle this collision using its default collision system.
// -----------------------------------------------------------------------------
int webots_physics_collide(dGeomID g1, dGeomID g2) 
{
	TRY(webots_physics_collide);
  	return 0;					// collision not handled
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: draw_support_polygon
/// \brief Draws the given support polygon using OpenGL primitives.
/// \param polygon a vector of 3D points which form the polygon.
// -----------------------------------------------------------------------------
void draw_support_polygon(vector< Point_3 > polygon)
{
	TRY(draw_support_polygon);
	const float groundY = 0.2;
	const int lineWidth = 2;

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glLineWidth(lineWidth);

	glBegin(GL_LINE_LOOP);
	glColor3f(0.0, 1.0, 1.0);	// cyan color
	for (uint i = 0; i < polygon.size(); i++) {
		glVertex3f(polygon[i][0], groundY, polygon[i][2]);
	}
	glEnd();

	glLineWidth(1);
	glEnable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);
	CATCH;
}
// -----------------------------------------------------------------------------
//  Function: webots_physics_draw
/// \brief A utility function intended to display additional 3D objects in the 
/// main 3D window. 
/// Details: Useful to display some forces as lines with arrows or to add some 
/// objets in the world. It is called immediately after the world is displayed. 
/// It should contain OpenGL calls glEnable, glDisable, glColor4f, glBegin, 
/// glVertex3f, glEnd, etc. The OpenGL state should be restored to default value
/// at the end to avoid subsequent rendering problems in Webots.
// -----------------------------------------------------------------------------
void webots_physics_draw()
{
	TRY(webots_physics_draw);
	vector< Point_3 > points;
	Point_3 zeros(0, 0, 0);
	for (int i = 0; i < 2 * CONTACT_NUM; i++) {
		if (soleContacts[i] != zeros)
			points.push_back(soleContacts[i]);
	}

	// Create the convex hull of the contact points (support polygon):
	vector< Point_3 > convex;
	CGAL::convex_hull_points_2(points.begin(), points.end(), back_inserter(convex),
							   CGAL::Convex_hull_projective_xz_traits_2< Point_3 >());

	// Draw the support polygon:
	draw_support_polygon(convex);
	CATCH;
}

// -----------------------------------------------------------------------------
//  Function: webots_physics_cleanup
/// \brief Counterpart function to the webots_physics_init function. 
/// Details: Called when the world is destroyed and can be used to perform some
/// cleanup, like releasing resources etc.
// -----------------------------------------------------------------------------
void webots_physics_cleanup() 
{
	TRY(webots_physics_cleanup);
	cout << "Cleaning up hoap2 physics:" << endl;

	// Detach the shared memory segments:
	if (shmDynamics.detach() < 0)
		THROW_FATAL("Can't detach dynamics shared memory!");
	if (shmFootGround.detach() < 0)
		THROW_FATAL("Can't detach foot-ground shared memory!");

	// Remove the shared memory segments:
	if (shmDynamics.remove() < 0)
		THROW_FATAL("Can't remove dynamics shared memory!");
	if (shmFootGround.remove() < 0)
		THROW_FATAL("Can't remove foot-ground shared memory!\nerrno = " << errno);

	// Close shared memory semaphores:
	semDynamics.close();
	semFootGround.close();

	cout << "Clean up finished!" << endl;
	CATCH;
}
