/* -*- Java -*- */
/****************************************************************
 *                                                              *
 *   Biped Locomotion                                           *
 *                                                              *
 *   Created by:  Jean-Philippe Eager                           *
 *                Yves-Marie Le Dru                             *
 *   Adapted by:  Yariv Bachar (Y.Bachar@sms.ed.ac.uk)          *
 ****************************************************************
 * $Id: $
 ****************************************************************/
/**
 * @file: FileParser.java
 */

package MotionTool;

import com.cyberbotics.webots.Controller;

import java.io.*;
import java.util.Vector;
import java.util.Iterator;

import MotionTool.Motion;
import MotionTool.GlobalDefinitions;

/**
 *  A Parser for the sequences file based on the the work done by Yves-Marie Le Dru
 *  Reads/Writes the variables, Motions and Seqeunces from/to the Sequences file
 *
 *  usage : <br>
 *  SeqManager MySeqManagerObject= new SeqManager("Sequences.txt",ActiveLog,logMode);  <br>
 *
 *  @author Jean-Philippe Egger (Robot SongoKu)
 *  @author Yves-Marie Le Dru
 *  @see <a href="../Sequences.txt" target=blank>Sequences.txt</a>
 */
public class FileParser extends Thread {

    //The allowed definitions Tags for the Motion definition file

    //The Var tag
    private static final int VARDEFTAGID = 1;
    private static final String VARDEFTAG = "VariableDef";
    private static final String VARTAG = "Var";
    private Vector argListVar = new Vector();
    //The MotionTag
    private static final String MOTIONDEFTAG = "MotionDef";
    private static final int MOTIONDEFTAGID = 2;
    private static final String MOTIONTAG = "Motion";
     private Vector argListMotion = new Vector();
    //The SequenceTag
    private static final String SEQUENCEDEFTAG = "SequenceDef";
    private static final int SEQUENCEDEFTAGID = 3;
    private static final String SEQUENCETAG = "Seq";
    private Vector argListSequence = new Vector();
    //The MotionTool Tag
    private static final int MOTIONTOOLTAGID = 4;
    private static final String MOTIONTOOLTAG = "MotionTool";

    //The ServoTag
    private static final String SERVOTAG = "servo";
    private Vector argListServo = new Vector();
    //The Position
    private static final String POSITIONTAG = "positions";
    //The frames
    private static final String FRAMESTAG = "frames";


    //The name of the parsed file
    String parsedFile="";

    //The Variables Loaded from file
    private Vector variables = new Vector(20);
    //The Motions Loaded from file
    private Vector motions = new Vector(20);
    //The Sequences Loaded from file
    private Vector sequences = new Vector(20);

    //Debug variables
    private boolean logMode = false;
    private boolean logToFileMode = false;
    public boolean hasParsedError = false;
    private int errorsCount = 0;

    //Temporary holders for the parsed attributes name and values
    private Vector attributNames = new Vector();
    private Vector attributValues = new Vector();

    /**
     * Defines the acceptable values for all the diffrent tags
     */
    private void initAcceptableTags()
	{
        // Defining the acceptable attributes for the VAR tag
        argListVar.add(new String("name"));
        argListVar.add(new String("value"));
        logAcceptableTags(VARTAG, argListVar);

        // Defining the acceptable attributes for the MOTION tag
        argListMotion.add(new String("name"));
        argListMotion.add(new String("DefaultLoop"));
        argListMotion.add(new String("Defaultreverse"));
        logAcceptableTags(MOTIONTAG, argListMotion);

        // Defining the acceptable attributes for the SEQUENCE tag
        argListSequence.add(new String("name"));
        argListSequence.add(new String("DefaultLoop"));
        logAcceptableTags(SEQUENCETAG,argListSequence);

        argListServo.add(new String("name"));
        logAcceptableTags(SERVOTAG, argListServo);
    }

    // Logs the defined accptable parameter for a given tag
    private void logAcceptableTags(String tag, Vector list)
	{
        String logText = " Acceptable attributes for '" + tag + "' tag are: (";
        Iterator iter = list.iterator();
        while (iter.hasNext()) {
            logText += (String)iter.next() + ", ";
        }
        logText += ")";
        log(logText);
    }

    // Class for storing the variables defined in the Motion definition File
    private class Variable {
        private String name;
        private Float value;

        public Variable(String nameC, Float valueC)
		{
            name = nameC;
            value = valueC;
        }

        public String toString()
		{
            return "Variable "+ name + " = " + value;
        }
        
		public String getName()
		{
            return name;
        }
        
		public Float getValue()
		{
            return value;
        }
    }

    /**
     * Constructor:
     * Parses the given file and creates the textual defined Objects
     * @param fileName The name of the file to be parsed
     * @param log If true will display information about the parsing process in the terminal
     * @param logToFile If true will write information about the parsing process into a debug file
     */
    public FileParser(String fileName, boolean log, boolean logToFile) 
	{
        logMode = log;
        logToFileMode = logToFile;
        parsedFile = fileName;
        hasParsedError = false;
        // If debug Mode: clean the file or create it
        if (logToFileMode) writeFile("DebugLog.txt", "Debug Log :", false);
        log("");
		
        // Defining the acceptable attributes
        initAcceptableTags();
        loadMotionFile(parsedFile);

        if (hasParsedError) {
            motions = new Vector();
            motions.add(new Motion());
            errorLog ("The parsed File is not Correct please view the errors in Errorlog.txt");
        } else 
			log("The Motion was succesfully loaded");

    }

    // Returns the parsed Motions
    public Vector getMotions() 
	{
        return motions;
    }

    // Return the parsed Sequences
    public Vector getSequences()  
	{
        return sequences;
    }

    /**
     * Parses a Definition Tag and returns it's corresonding ID
     * The StreamTokenizer must be just before a valid definition tag
     * If the Tag is not valide will log the error and stop the programm
     */
    int parseDEFTag(StreamTokenizer st)
	{
        //log("Starting Def Tag");
        String tag = "";
        try {
            // The type of the actual Token
            int type = st.ttype;

            // Get start of Variable Tag
            while (type == ' ' || type == StreamTokenizer.TT_EOL) {
                type = st.nextToken();
            }
            if (type != '<') errorLog("<", st);
            type = st.nextToken();
			
            // Get Name of Def Tag
            if (type == '/'){
                type = st.nextToken();
                if (!st.sval.equals(MOTIONTOOLTAG)) errorLog(MOTIONTOOLTAG, st);
                type = st.nextToken();
                if (type != '>') errorLog(">", st);
                type = st.nextToken();
                return MOTIONTOOLTAGID;
            }

            if (type != StreamTokenizer.TT_WORD) errorLog("DEFTAG", st);
            tag = st.sval;
            type = st.nextToken();
             int result = 0;
            //Check if allowed Definition
            if (tag.equals(VARDEFTAG)) result= VARDEFTAGID;
            else if (tag.equals(SEQUENCEDEFTAG)) result= SEQUENCEDEFTAGID;
            else if (tag.equals(MOTIONDEFTAG)) result= MOTIONDEFTAGID;
            else errorLog("Error in "+ parsedFile + ": Expected <" + VARDEFTAG +
						  ">, <" + SEQUENCEDEFTAG + "> or <" + MOTIONDEFTAG + 
						  "> found <" + tag + "> on line " + st.lineno());
            //type = st.nextToken();
            //Check correct finish of tag
            if (type != '>') errorLog(">", st);
             type = st.nextToken();
            return result;

        } catch (IOException e) {
            System.out.println("Caught access file IOException: " + e.toString());
            errorLog("Caught access file IOException: " + e.toString());
        }
        return 0;
    }

    /**
     * Returns a positive if the variable "varName" has already been defined
     * @param varName The name of the variable we are searching for
     * @return The position in the vector of the variable. If not in vector returns -1
     */
    public int variableAlreadyDefined(String varName){

        Iterator varIt = variables.iterator();
        int pos =0;
        while(varIt.hasNext()){
             if (((Variable)varIt.next()).getName().equals(varName)) return pos;
            pos++;
        }
        return -1;
    }

     /**
     *  Parses the attributes of a tag
     *  The StreamTokenizer must be inside a tag just after the definition
     *  If "argObligatory" is set then the tag must have all the attributes
     *  Example: in argList ther is "name" and "value" the tag to be parsed
     *  must have the form <XXX name=myVar value=12>
     *  The attributes will be stored in the global variables "attributeNames" and "attributeValues"
     */
    boolean parseAttributes(StreamTokenizer st, Vector argList, boolean argObligatory){


        //reset the attribut vectors
        attributNames.removeAllElements();
        attributValues.removeAllElements();
        //To know witch argument has been parsed
        boolean argParsed[] = new boolean[argList.size()];
        //Set to false
        for (int i = 0; i < argList.size(); i++){
            argParsed[i]=false;
        }
        try{
            //The type of the actual Token
            int type = st.ttype;
             //Loop until end of Tag
            while (type != '>' && type !=StreamTokenizer.TT_EOL) {

                String attributeName=st.sval;
                //log("AttributeName = " + attributeName);
                int attributePosition = argList.indexOf(attributeName);
                //Attribute not accepted
                if (attributePosition < 0) errorLog(argList.toString(), st);
                type = st.nextToken();
                if (type != '=') errorLog ("=", st);
                type = st.nextToken();
                //Add attribute Name to vector
                attributNames.add(attributeName);
                //String logMessage = "          Attribute " + attributeName + " ";
                //Add attribute Value to vector
                 String attribute = st.sval.substring(1,st.sval.length()-1);
                if (type == StreamTokenizer.TT_WORD) {


                    //log ("AttributeValue = " + attribute);
                    Float value;
                    int varDefined = variableAlreadyDefined(attribute);

                    //The value is a varibale already defined!!
                    if(varDefined > 0){
                      value = ((Variable)variables.elementAt(varDefined)).getValue();
                      attributValues.add(value);
                        //logMessage += " set to (VariableValue)" + value + " .";
                    }else {
                        attributValues.add(attribute);
                        //logMessage += " set to (word)" + st.sval + " .";
                    }

                }   else if (type == StreamTokenizer.TT_NUMBER){
                    attributValues.add(new Float(attribute));
                    //logMessage += " set to (float)" + st.nval + " .";
                }else errorLog ("a String or a Number", st);
                type = st.nextToken();
                //log(logMessage);
                argParsed[attributePosition]=true;
            }

            //Are all attributes defined??
            if (argObligatory) {
                for (int i = 0; i < argList.size(); i++){
                    if (argParsed[i]==false) errorLog("attribute '" + argList.elementAt(i)+ "'", st);
                }
            }

        } catch (IOException e) {
                   System.out.println("Caught access file IOException: " + e.toString());
                   errorLog("Caught access file IOException: " + e.toString());
        }
        return true;
     }
	
    /**
     *  Parses and reads a tag and its attributes
     *  The parsed tag must have the Tag name = "tagName" and can have the arguments defined in the argList
     *  If "argObligatory" is set then the tag must have all the attributes
     *  Example: tagName="Var" and in argList ther is "name" and "value" the tag to be parsed
     *  must have the form <VAR name=myVar value=12>
     *  The attributes will be stored in the global variables "attributeNames" and "attributeValues"
     */
    boolean parseTag(StreamTokenizer st, String tagName, Vector argList, boolean argObligatory){

    try{
        //The type of the actual Token
        int type = st.ttype;

        //Get start of Tag
        while (type == ' ' || type==StreamTokenizer.TT_EOL) {
            type = st.nextToken();
        }
        if (type != '<') errorLog("<", st);
        type = st.nextToken();

        //End of Defnition
        if (type == '/') {
            return false;
        }

        //Get Name of the Tag
        String tag = st.sval;
        //Is it the right tag name ?
        if (!tag.equals(tagName))errorLog(tagName, st);
        type = st.nextToken();
        parseAttributes(st, argList, argObligatory);
        type=st.ttype;
        //Check correct finish of tag
        if (type != '>') errorLog(">", st);
        type = st.nextToken();

    } catch (IOException e) {
            System.out.println("Caught access file IOException: " + e.toString());
            errorLog("Caught access file IOException: " + e.toString());
        }
     return true;
    }

    private Vector parsePositions(StreamTokenizer st, String name) throws IOException{
         int type = st.nextToken();
        Vector result= new Vector();
               //Get start of Tag
            while (type == ' ' || type==StreamTokenizer.TT_EOL) {
               type = st.nextToken();
            }
            if (type!= '<') errorLog("<", st);
               type = st.nextToken();
            if (type!=StreamTokenizer.TT_WORD || !st.sval.equals(name)) errorLog(name, st);
               type = st.nextToken();
                 if (type!= '>') errorLog(">", st);
            type = st.nextToken();
            while (type!='<'){
                //Has to be a number
                if (type!=StreamTokenizer.TT_NUMBER)errorLog ("A float Value", st);
                //Add the parsed values
                result.add(new Double(st.nval));
                type = st.nextToken();
            }

            if (type!= '<') errorLog("<", st);
                      type = st.nextToken();
            if (type!= '/') errorLog("/", st);
                      type = st.nextToken();

                   if (type!=StreamTokenizer.TT_WORD && !st.sval.equals(name)) errorLog(name, st);
                      type = st.nextToken();
                        if (type!= '>') errorLog(">", st);
                   type = st.nextToken();


        return result;
    }
    private boolean parseKeyTag(StreamTokenizer st, Motion mot){

        try{
            //The type of the actual Token
            int type = st.ttype;

            //Get start of Tag
            while (type == ' ' || type==StreamTokenizer.TT_EOL) {
                type = st.nextToken();
            }
            if (type != '<') errorLog("<", st);
            type = st.nextToken();

            //End of motion Definition
            if (type == '/'){

                return false;
            }
            type = st.nextToken();
            parseAttributes(st,argListServo,true);
            type = st.nextToken();

            //Getthe Name of the Servo
            String servoName = (String)attributValues.elementAt(0);
            log ("Servo Name " + servoName);
            //Does the servo Definition Exist??
            //Find ServoID
            int servoId=-1;
            for (int i = 0; i < GlobalDefinitions.servosNames.length; i++) {
                //log("Searching for servo Id, name = " + servoName + " comparing to " + servosNamesInFile[i] + " result = "+ servoName.equals(servosNamesInFile[i]));
                if (servoName.equals(GlobalDefinitions.servosNames[i])) {
                    servoId = i;
                    //log("found: ServoId = " + servoId + " servoName= " + servoName);
                   // break;
                }
            }
            log("ServoID = " + servoId);
            //Didn't find -> Wrong servo name
            if (servoId < 0) {

                String expected = "Error in "+ parsedFile + ": Expected one of {";
                for (int i=0; i < GlobalDefinitions.servosNames.length; i++){
                    expected += GlobalDefinitions.servosNames[i] + ", ";
                }
                expected+="} found "+ servoName + " on line " + st.lineno();
                errorLog(expected);
            }


            //Find the values
            Vector val = parsePositions(st, POSITIONTAG);
            Vector frameNumber = parsePositions(st, FRAMESTAG);

            //Get start of Tag
            while (type == ' ' || type==StreamTokenizer.TT_EOL) {
                type = st.nextToken();
            }

            if (type!= '<') errorLog("<",st);
                type = st.nextToken();
            if (type!= '/') errorLog("/",st);
            type = st.nextToken();
            if (!st.sval.equals(SERVOTAG)) errorLog(SERVOTAG,st);
            type = st.nextToken();
            if (type!= '>') errorLog(">",st);
            type = st.nextToken();


           //Number of values for positions must be equal to number of values of FrameNumbers
           if (val.size()!=frameNumber.size())errorLog("Error in "+ parsedFile + " on line " + st.lineno() +":  In the definition of Motion '"+ mot.motionName +"' the numbers of servo Values for '"+ servoName + "'("+ val.size() + ") must be equal to the number of FrameNumbers ("+frameNumber.size()+")");
           String logKeyFrames = "          <" + servoName ;
           for (int i=0; i < val.size(); i++){
               int frame = ((Double)frameNumber.elementAt(i)).intValue();
               float value = ((Double)val.elementAt(i)).floatValue();
                mot.addKeyFrame(servoId, frame, value);
                logKeyFrames += " (" + value + "," + frame+ "), ";
           }
           log (logKeyFrames + ">");

        } catch (IOException e) {
                System.out.println("Caught access file IOException: " + e.toString());
                errorLog("Caught access file IOException: " + e.toString());
            }
        return true;
    }


    /**
    * Logs the current loaded attributes with the corresponding tag name
    */
    void logTagWithAttributes(String tagName){

    String logMessage = "     <" + tagName + " ";
    Iterator attrIt = attributNames.iterator();
    while (attrIt.hasNext()){
        String name =(String)attrIt.next();
        logMessage += name;
        logMessage +="=";
        logMessage += attributValues.elementAt(attributNames.indexOf(name));
        logMessage +=" ";
    }
    logMessage+=">";
    log(logMessage);
    }


    void readVariableDefinitions(StreamTokenizer st){

        int type;

        //While encountering var tags parse them
        while (parseTag(st,VARTAG,argListVar,true)){

            //The name and value of the variable parsed
            String varName;
            int namePos=0;
            Float varValue;
            int valuePos=1;
            //Finding name and value from attribute list
            String attrib1= (String)attributNames.elementAt(0);

            //Second attribute is name
            if (!attrib1.equals("name")){
                 namePos=1;
                 valuePos=0;
            }
            varName = (String)attributValues.elementAt(namePos);
            Object varValueO = attributValues.elementAt(valuePos);
            //Check if value is realy a Float
            if (!varValueO.getClass().isInstance(new Float(0))){
                  errorLog("Error in " + parsedFile + ": Expected a Float value found '"+ (String)varValueO + "' on line " + st.lineno());
            }
            varValue= (Float)attributValues.elementAt(valuePos);
            Variable temp = new Variable(varName,varValue);
            variables.add(temp);
            logTagWithAttributes(VARTAG);
        }

        type = st.ttype;
        try{
        //Check for correct ending of DEF Tag
        if (type!= '/') errorLog ("/", st);
        type = st.nextToken();
        if (!st.sval.equals(VARDEFTAG))errorLog(VARDEFTAG, st);
        type = st.nextToken();
        if (type!= '>') errorLog (">", st);
        type = st.nextToken();
        } catch (IOException e) {
                   System.out.println("Caught access file IOException: " + e.toString());
                   errorLog("Caught access file IOException: " + e.toString());
        }
    }

    /**
     * return the position of the attribute 'attribName' in the attribute vector
     * if not found returns -1
     * @param attribName The name of the attribute
     * @return the position of the attribute in the vector
     */
    private int getAttributePos(String attribName){
        return attributNames.indexOf(attribName);
   }

    private void readMotionDefinitions(StreamTokenizer st){

       //While encountering MOTION-TAGS parse them
        while (parseTag(st,MOTIONTAG,argListMotion,false)){
            logTagWithAttributes(MOTIONTAG);
            Motion mot = new Motion();
            int namePos = getAttributePos("name");
            //If no name attribute found ->error
            if (namePos < 0) errorLog("Error in "+ parsedFile + ": Could not find attribute 'name' in Motion-Definition on line " + st.lineno());
            mot.motionName=(String)attributValues.elementAt(namePos);

            //Parses and adds keyframes to the motion mot
            while (parseKeyTag(st,mot)){
            }
            try{
            //Check for correct ending of Motion Tag
            int type = st.nextToken();
            if (!st.sval.equals(MOTIONTAG))errorLog(MOTIONTAG, st);
            type = st.nextToken();
            if (type!= '>') errorLog (">", st);
            type = st.nextToken();
            motions.add(mot);
            log("     </MOTION>");
            log("");

            } catch (IOException e) {
                       System.out.println("Caught access file IOException: " + e.toString());
                       errorLog("Caught access file IOException: " + e.toString());
            }
        }

        int type = st.ttype;
        try{
        //Check for correct ending of MOTION_DEF Tag
        if (type!= '/') errorLog ("/", st);
        type = st.nextToken();
        if (!st.sval.equals(MOTIONDEFTAG))errorLog(MOTIONDEFTAG, st);
        type = st.nextToken();
        if (type!= '>') errorLog (">", st);
        type = st.nextToken();
        } catch (IOException e) {
                   System.out.println("Caught access file IOException: " + e.toString());
                   errorLog("Caught access file IOException: " + e.toString());
        }


    }


    /**
     *  LoadSequenceFile(String Filename);  <br>
     *
     *  Load the Sequence definition file for initialising the SeqManager Object. <br>
     *
     *  You shouldn't call directly this method.    <br>
     *  It is called directly by the constructor of SeqManager.     <br>
     *  If you want Load a new file, its better to create a new SeqManager Object.  <br>
     *  Else you will have to reinitialise all the variables before call to  LoadSequenceFile(String Filename). <br>
     *
     */
    public int loadMotionFile(String Filename) {

        int ret = 1;

        try {

            //Load file and create a corresponding Tokenizer
            FileReader File = new FileReader(Filename);
            BufferedReader Buffer = new BufferedReader(File);
            StreamTokenizer st = new StreamTokenizer(Buffer);

            //Define Tokenizer for specifique File format
            st.eolIsSignificant(true);
            st.lowerCaseMode(false);
            st.ordinaryChar('<');
            st.ordinaryChar('>');
            st.ordinaryChar('=');
            st.ordinaryChar('/');
            //st.commentChar('#');
            st.wordChars('_', '_');
            st.wordChars('?', '?');
            st.wordChars('!', '!');
            st.wordChars('"', '"');
            //st.slashSlashComments(true);
            //st.slashStarComments(true);


            //The type of actual Token
            int type = 0;
             type = st.nextToken();

            //Goto first caracter
            while (type == ' ' || type==StreamTokenizer.TT_EOL) {
                type = st.nextToken();
            }

            if (type!= '<') errorLog("<", st);
            type = st.nextToken();

            //Skip the <?xml> Tag if present
            if (st.sval.equals("?xml")){
                log("<?xml ... >");
                while (type != '>' && type!= StreamTokenizer.TT_EOF){
                    type = st.nextToken();
                }
                if (type == StreamTokenizer.TT_EOF)errorLog(">", st);
                type = st.nextToken();
                 //Get next start of Tag
                while (type == ' ' || type==StreamTokenizer.TT_EOL) {
                type = st.nextToken();
                }
                if (type!= '<') errorLog("<", st);
                type = st.nextToken();
            }
            //Needs to start with the MotionTool Tag
            if (!st.sval.equals(MOTIONTOOLTAG)) errorLog (MOTIONTOOLTAG, st);
            while (type != '>' && type!= StreamTokenizer.TT_EOF){
                    type = st.nextToken();
                }
            if (type == StreamTokenizer.TT_EOF)errorLog(">", st);
            type = st.nextToken();
            log ("  <"+ MOTIONTOOLTAG + ">");
            //Parse file while not the end of File
            while (type != StreamTokenizer.TT_EOF) {

                //Valid definitionTags and there corresponding parse function
                switch (parseDEFTag(st)){

                    //Starting Parsing of Variables
                    case VARDEFTAGID:

                        log("");
                        log("Loading  VARIABLES_DEF...");
                        log("");
                        readVariableDefinitions(st);
                        log("");
                        log("VARIABLES_DEF loaded ");
                        while (type==StreamTokenizer.TT_EOL) type=st.nextToken();
                        break;

                    //Starting Parsing of Variables
                    case MOTIONDEFTAGID:
                        log("");
                        log("Loading  MOTIONS_DEF...");
                        log("");
                        readMotionDefinitions(st);
                        log("");
                        log("MOTIONS_DEF loaded ");
                        type=st.nextToken();
                        while (type==StreamTokenizer.TT_EOL) type=st.nextToken();
                        break;

                    case SEQUENCEDEFTAGID:
                        log("");
                        log("Loading  SEQUENCES_DEF...");
                        log("");
                        log("SEQUENCES_DEF loaded ");
                        type = st.nextToken();
                        while (type==StreamTokenizer.TT_EOL) type=st.nextToken();
                        break;

                    case MOTIONTOOLTAGID:
                        log("  <" + MOTIONTOOLTAG + ">");
                        log("");
                        return 1;

                    default:
                        errorLog("Not a Valid definition ID..");
                        break;
                }

            }

        } catch (IOException e) {
            System.out.println("Caught access file IOException: " + e.toString());
            ret = -1;
        } // end of try / catch for file acces errors

        return ret;
    } // End of  int LoadSequenceFile(String Filename)


    /**
     *  GetVariableValue(String VarTag);     <br>
     *
     *  Return the float value of the Variable Tag in the Sequence file.
     */
    public float GetVariableValue(String VarTag) {

        /*for (int i = 0; i < variablesCount; i++) {
            if (VarTag.toLowerCase().equals(variablesNames[i].toLowerCase())) return variablesValues[i];
        }
        errorLog("WARNING : Variable " + VarTag + " not found in Variables List");
        */
        return 0.0f;

    }

    /**
     *  boolean AddSequence(Sequence TheSequence); <br>
     *
     *  Add a sequence in the Sequence list,
     *  return true if completed.                  <br>
     *
     */
    private boolean AddSequence(Sequence TheSequence) {
        if (TheSequence != null) {
            sequences.add(TheSequence);
            return true;
        }
        return false;
    }


    /**
     *  GetMotionID(String MotionTag)       <br>
     *  Usage :                                 <br>
     *  int MyMotion=GetMotionID("MyMotionTag");    <br>
     *
     *   Return The ID of the motion.              <br>
     *   This function is for SeqManager internal use. <br>
     *   If you want play the motion only (and not a whole sequence),
     *   use MySequence=GetSequenceID("MyMotionTag") instead with the motion tag. <br>
     *
     *
     */
    int GetMotionID(String MotionTag) {

        int id = 0;
        Iterator motionIt = motions.iterator();

        while (motionIt.hasNext())
        {
            String motionName= ((Motion)motionIt.next()).motionName;
            log("Searching for MotionTag '" + MotionTag + "': Actual MotionName is '"+ motionName + "' with id " + id);
            if (MotionTag.equals(motionName)){
                return id;
            }
            id++;
        }
        errorLog("WARNING : motion " + MotionTag + " not found in Motions List");
        return 0;
    }

    /**
     * GetSequenceID(String MotionTag);    <br>
     *  Usage :                                 <br>
     *  int MySequence=GetSequenceID("MySequenceTag");      <br><br>
     *
     *   Return The ID of the Sequence.                       <br>
     *   Use this function to get a Sequence ID for use with NextSeqFrame(SequenceID,DefaultSequenceID). <br>
     *
     *   The tag is the name of the Sequence in Sequences.txt file. <br>
     *   You can also pass a motions Tag as parameter, if so,
     *   it will create a Sequence with only this Motion and return the ID of the newly Sequence. <br>
     *
     *
     */
    public int GetSequenceID(String SeqTag) {
        // Search for the Sequence in the list with the string name : SeqTag
        /*for (int IndexSequence = 0; IndexSequence < sequencesCount; IndexSequence++) {
            if (SeqTag.toLowerCase().equals(sequencesList[IndexSequence].sequenceName.toLowerCase())) return IndexSequence;
        }
// Sequence not found, now check if it's a motion
/*for (int IndexMotion = 0; IndexMotion < motionsCount; IndexMotion++) {
            if (SeqTag.toLowerCase().equals(motionsList[IndexMotion].StringName.toLowerCase())) {
                // The Tag is a Motion, so create a new sequence with only this motion
                if (sequencesCount < MAX_SEQUENCES_ALLOWED) {
                    // Pointer to the Motion
                    Motion M = motionsList[IndexMotion];
                    // Create a new sequence and add the motion into
                    Sequence NewSequence = new Sequence();
                    NewSequence.AddOrder(IndexMotion, M.DefaultLoop, M.DefaultReverse, M.DefaultInversed, M.DefaultAI_MESSAGE);
                    AddSequence(NewSequence);
                    // Return the newly sequence index
                    return NewSequence.ID;
                }
            }


        }*/


        errorLog("WARNING : Sequence or Motion :" + SeqTag + " not found in Sequences List");
        return 0;
    }

// ***************************** Debug Informations ***************************


    /**
     *  Log(String output); <br>
     *
     *  Print a String in the Console if loogActived is set to true
     *  Print it in DebugLog.txt if Debug mode actived
     *  else do nothing .
     *
     */
    public void log(String output) {
        if (logMode&& !hasParsedError) {
            System.out.println(output);
        }
        if (logToFileMode&& !hasParsedError) {
            writeFile("DebugLog.txt", '\n' + output, true);
        }
    }

    /**
     *  errorLog(StreamTokenizer st, String output); <br>
     *
     *  Print an error message with lineNumber of parsed File in the Console and in ErrorLog.txt.
     *  Print the message in debugLog.txt if Debug mode actived
     *
     */
   public void errorLog(String expected, StreamTokenizer st) {
        String output = "Error in"+ parsedFile + ": Expected "+ expected + " found  " + st.toString();
        // Erase/create the errorLog file if its the first error
        if (errorsCount == 0) writeFile("ErrorLog.txt", "Errors found for this session :", false);
        errorsCount++;
        if (!hasParsedError){
            System.out.println(output);
            writeFile("ErrorLog.txt", '\n' + output, true);
            if (logToFileMode) {
                writeFile("DebugLog.txt", '\n' + output, true);
            }
        }
        hasParsedError=true;
    }

    /**
     *  errorLog(String output); <br>
     *
     *  Print an error message in the Console and in ErrorLog.txt.
     *  Print the message in debugLog.txt if Debug mode actived
     *
     */
    public void errorLog(String output) {
        // Erase/create the errorLog file if its the first error
        if (errorsCount == 0) writeFile("ErrorLog.txt", "Errors found for this session :", false);
        errorsCount++;

            System.out.println(output);
            writeFile("ErrorLog.txt", '\n' + output, true);
            if (logToFileMode) {
                writeFile("DebugLog.txt", '\n' + output, true);
            }
        hasParsedError=true;
    }

    /**
     *  writeFile(String Filename, String TmpStr,boolean Append); <br>
     *
     *  Write a String in the specified filename. <br>
     *  If append is set to true, the file is appended. <br>
     *
     */
    public int writeFile(String Filename, String TmpStr, boolean Append) {

        try {
            FileWriter File = new FileWriter(Filename, Append);
            BufferedWriter Buffer = new BufferedWriter(File);

            Buffer.write(TmpStr, 0, TmpStr.length());

            Buffer.close();
            File.close();

        } catch (IOException e) {
            System.out.println("Caught access file IOException: " + e.toString());
            return -1;
        }
        return 0;
    }

// ***************************** End Debug Informations ***************************
}
