package MotionTool.Interface;

import MotionTool.GlobalDefinitions;
import MotionTool.ServoKeyFrames;
import MotionTool.KeyFrame;
import MotionTool.Motion;

import javax.swing.*;
import java.awt.*;
import java.awt.image.*;
import java.lang.*;
import java.util.Iterator;

//Manages a drawing board

public class Board {

    //The dimensions of the drawing board
    public int boardWidth = 500;
    public int boardHeight = 234;
    //The pixels of the board we draw on
    public int[] board = new int[boardWidth * boardHeight];

    //The JLabel Image of the board to be displayed in the JFrame
    BufferedImage boardImage = new BufferedImage(boardWidth, boardHeight, BufferedImage.TYPE_INT_ARGB);
    JLabel boardImageLabel = new JLabel(new ImageIcon(boardImage));

    //The min, max values that can be drawn on the board
    public int maxBoardValue = (int) transformToBoardValue(GlobalDefinitions.maximalPossibleServoValue);
    public int minBoardValue = (int) transformToBoardValue(-GlobalDefinitions.maximalPossibleServoValue);

    //The 22 colors for tracing the diffrent motors
    private int[] servoColors = GlobalDefinitions.servoColors;

    //colors of the Scales on the board
    final int colorPinkLight = 0xffffddff;
    final int colorPink = 0xffffbbff;
    final int colorBlueLight = 0xffddddff;
    final int colorGreyLight = 0xffdddddd;
    final int colorSkyLight1 = 0xffaaaaff;
    final int colorSkyLight2 = 0xffb3ffff;
    final int colorSkyLight3 = 0xffceffff;
    final int colorWhite = 0xffffffff;

    //The KeyFrames defining the mouvements displayed on the board
    VectorOfServoKeyFrames boardKeyFrames = new VectorOfServoKeyFrames();

    /**
     * Constructor
     */
    public Board(MotionInterface.BoardEventHandler eventHandler, 
				 MotionInterface.BoardMotionEventHandler motionHandler, 
				 int position) 
	{
        board = new int[boardWidth * boardHeight];
        boardImageLabel.setBackground(Color.green);
        boardImageLabel.setOpaque(true);
        boardImageLabel.setBounds(new Rectangle(5, position, boardWidth, boardHeight));
        boardImageLabel.addMouseListener(eventHandler);
        boardImageLabel.addMouseMotionListener(motionHandler);
        //Set all servos to visbile
        for (int i=0; i < GlobalDefinitions.nbrJoints; i++){
            isVisible [i] = true;
        }
    }

    //If x,y points to a KeyFrame defined for servoId, this function returns
    //the position of the Key Frame in the KeyFrames Vector corresponding to the servoId.
    //Else returns -1
    public int isOnKeyFrame(int servoId, int x, int y) 
	{
        int pos = 0;
        ServoKeyFrames temp = boardKeyFrames.getServoKeyFrames(servoId);
        Iterator it = temp.keyFrames.iterator();
        while (it.hasNext()) {
            KeyFrame frame = (KeyFrame) it.next();
            if (frame.frameNumber - 5 < x && 
				frame.frameNumber + 5 > x && 
				frame.servoValue - 5 < y && 
				frame.servoValue + 5 > y) return pos;
            pos++;
        }
        return -1;
    }

    public KeyFrame getKeyFrameUnderMouse(int servoId, int x, int y) 
	{
		ServoKeyFrames temp = boardKeyFrames.getServoKeyFrames(servoId);
		Iterator it = temp.keyFrames.iterator();
		while (it.hasNext()) {
			KeyFrame frame = (KeyFrame) it.next();
			if (frame.frameNumber - 5 < x && 
				frame.frameNumber + 5 > x && 
				frame.servoValue - 5 < y && 
				frame.servoValue + 5 > y) 
				return frame;
		}
		return new KeyFrame(-1,-1f);
	}

    //Moves a KeyFrame given it's servoId and it's Position in the KeyFrame Vector
    //The position is the one return by isOnKeyFrame()
    public void moveKeyframe(int servoId, int pos, int x, int y) 
	{
        if (y < maxBoardValue) y = maxBoardValue;
        if (y > minBoardValue) y = minBoardValue;
        ServoKeyFrames toChange = boardKeyFrames.getServoKeyFrames(servoId);
        toChange.keyFrames.removeElementAt(pos);

        toChange.addKeyFrame(x, (float) y);
        System.out.println("TransformedToServoValueInMoveingKey = " + transformToServoValue(y));
        redrawBoard();
    }

    public void removeKeyFrame(int servoId, int pos)
	{
        ServoKeyFrames toChange = boardKeyFrames.getServoKeyFrames(servoId);
        toChange.keyFrames.removeElementAt(pos);
        redrawBoard();
    }
	
    /**
     * Set the min, max values that can be drawn on the board
     */
    public void setMinMaxBoardValuesFromServoLimits(float min, float max) 
	{
        minBoardValue = transformToBoardValue(min);
        maxBoardValue = transformToBoardValue(max);
        redrawBoard();
    }


    public ServoKeyFrames getTransFormedServoKeyFrames(int servoId) 
	{
        Iterator it = boardKeyFrames.vec.iterator();
        int pos = 0;
        while (it.hasNext()) {
            ServoKeyFrames temp = (ServoKeyFrames) it.next();
            if (temp.servoId == servoId) 
				return transformToServoValues(
					(ServoKeyFrames) boardKeyFrames.vec.elementAt(pos));
            pos++;
        }
        //System.out.println("Number Of frames found = " + result.upDateNumberOfFrames());
        return transformToServoValues(new ServoKeyFrames(servoId));

    }

    public void setMotion(Motion actualMotion)
	{
		Iterator it = actualMotion.motionDefinition.iterator();
		removeAlleKeyFrames();

		while (it.hasNext()){
			ServoKeyFrames temp = (ServoKeyFrames) it.next();
			setTransformedServoKeyFrames(temp,true);
		}
    }
	
    public void removeAlleKeyFrames()
	{
		boardKeyFrames = new VectorOfServoKeyFrames();
    }
	
    /**
     *  Transformes the servo KeyFrames to board Key Frames and replaces the
     * board Key Frames by the new frames
	 * @param frames The new Frames to replace old values
     */
    public void setTransformedServoKeyFrames(ServoKeyFrames frames, boolean append)
	{
		//Transfor the values
		ServoKeyFrames transformed = transformToBoardValues(frames);
		//System.out.println("New transformed: " + transformed.toString());
		//Transform the frameNumbers
		ServoKeyFrames recalculated = recalculateFrameNumbers(transformed);
		//System.out.println("New recalculated: " + recalculated.toString());
		boardKeyFrames.setKeyFrames(recalculated,append);
	}

    public ServoKeyFrames recalculateFrameNumbers(ServoKeyFrames frames) 
	{
        Iterator frameIt = frames.keyFrames.iterator();
        KeyFrame actual;
        int evolutionFrameNumber = 0;
        while (frameIt.hasNext()) {
            actual = (KeyFrame) frameIt.next();
            int temp = actual.frameNumber;
            actual.frameNumber += evolutionFrameNumber;
            evolutionFrameNumber+=temp;
        }
        return frames;
    }

    //Transforms the Value of the servos from pixel position to real value
    public ServoKeyFrames transformToBoardValues(ServoKeyFrames frames) 
	{
        ServoKeyFrames result = new ServoKeyFrames(frames.servoId);
        Iterator frameIt = frames.keyFrames.iterator();
        KeyFrame temp,actual;

        while (frameIt.hasNext()) {
            actual = (KeyFrame) frameIt.next();
            temp = new KeyFrame(actual);
            temp.servoValue = transformToBoardValue(temp.servoValue);
            result.appendKeyFrame(temp.frameNumber, temp.servoValue);
        }
        //System.out.println("New Total FrameNumber = " + result.upDateNumberOfFrames());
        return result;
    }

    //Transforms the Value of the servos from pixel position to real value
    public ServoKeyFrames transformToServoValues(ServoKeyFrames frames) 
	{
        ServoKeyFrames result = new ServoKeyFrames(frames.servoId);
        Iterator frameIt = frames.keyFrames.iterator();
        KeyFrame temp,actual,last = new KeyFrame(0, 0);
        int totalFrames = 0;

        while (frameIt.hasNext()) {
            actual = (KeyFrame) frameIt.next();
            temp = new KeyFrame(actual);
            temp.servoValue = transformToServoValue(temp.servoValue);
            if (actual != frames.keyFrames.firstElement()) {
                //System.out.println("Is Not first Element!!!");
                temp.frameNumber = temp.frameNumber - last.frameNumber;
            }
            totalFrames += temp.frameNumber;
            result.appendKeyFrame(temp.frameNumber, temp.servoValue);
            last = new KeyFrame(actual);
        }
        //System.out.println("New Total FrameNumber = " + result.upDateNumberOfFrames());
        return result;
    }

    public float transformToServoValue(float value) 
	{
        float transformedValue = -(value - (boardHeight - 1.0f) / 2.0f) * 
			(GlobalDefinitions.maximalPossibleServoValue * 2 / 
			 (float) (boardHeight - 1.0));
        return transformedValue;
    }

    public int transformToBoardValue(float value) 
	{
        float transformedValue = -(value * (boardHeight - 1.0f) / 
								   (GlobalDefinitions.maximalPossibleServoValue * 2)) + 
			(boardHeight - 1.0f) / 2;
        return (int) transformedValue;
    }

    //Adds a Key Frame to the boardKeyFrame at the right position
    public void addKeyFrame(int servoId, int frameNumber, int value) 
	{
        if (value < maxBoardValue) value = maxBoardValue;
        if (value > minBoardValue) value = minBoardValue;
        boardKeyFrames.addKeyFrame(servoId, frameNumber, (float) value);
        System.out.println("Adding Frame: Frame Number = " + frameNumber + " value = " + value);//+ " Total Number of frames" + boardKeyFrames.getFirstServoKeyFrames().upDateNumberOfFrames());

        Motion test = new Motion();
        test.motionName = "Orig";
        Motion testTrans = new Motion();
        testTrans.motionName = "Transformed";
        test.appendServoKeyFrames(boardKeyFrames.getServoKeyFrames(servoId));
        testTrans.appendServoKeyFrames(getTransFormedServoKeyFrames(servoId));
        System.out.println(test.toString());
        System.out.println(testTrans.toString());
        //System.out.println("minBoardValue = " + minBoardValue);
        //System.out.println("maxBoardValue = " + maxBoardValue);
        redrawBoard();
    }

    //True if corresponding servo is Visible in the drawing board
	boolean[] isVisible = new boolean[GlobalDefinitions.nbrJoints];

	public void setVisibility (int servoID, boolean value)
	{
		if (servoID > 0 && servoID < GlobalDefinitions.nbrJoints)
			isVisible[servoID] = value;
	}


    /**
     * Erases the board and redraws the scale, the KeyFrames and the Motion line
     */
    public void redrawBoard() 
	{
        KeyFrame drawFrame;
        int previousXPos = 0;
        int previousYPos = 0;
        //Erase the board
        eraseBoard();
        //Draw the scale
        drawScale();

        //For each Servo defined draw it's KeyFrames and trajectory
        Iterator servoIt = boardKeyFrames.vec.iterator();
        int servoPos = 0;
        while (servoIt.hasNext()) {

            ServoKeyFrames servo = (ServoKeyFrames) servoIt.next();

            if (isVisible[servo.servoId]){
                //For each KeyFrame draw a point at it's position
                Iterator keyFrameIt = servo.keyFrames.iterator();
                while (keyFrameIt.hasNext()) {

                    drawFrame = (KeyFrame) keyFrameIt.next();
                    int actualColor = servoColors[servo.servoId];
                    int xPos = drawFrame.frameNumber;
                    int yPos = (int) drawFrame.servoValue;
                    drawKeyFrame(xPos, yPos, actualColor);
                    //If Not first element
                    if ((KeyFrame) servo.keyFrames.firstElement() != drawFrame) {
                        //Draw connecting line
                        float dx = xPos - previousXPos;
                        float dy = yPos - previousYPos;
                        float m = dy / dx;
                        //System.out.println("m = " + m + " dx = " + dx + " dy = " + dy);
                        for (int i = previousXPos; i <= xPos; i++) {
                            int y = (int) (m * (i - previousXPos) + previousYPos);
                            //System.out.println("y= " + y);
                            board[i + (y * boardWidth)] = actualColor;
                        }

                    }
                    previousXPos = xPos;
                    previousYPos = yPos;
                }//End While has next keyFrame

            }//End if isVisible

			servoPos++;
        }//End while has next servo

        //Recreate the BoardImage from the board variable and repaint it
        boardImage.setRGB(0, 0, boardWidth, boardHeight, board, 0, boardWidth);
        boardImageLabel.repaint();
    }

    /**
     * Puts all the pixels of the board to the same background color
     */
    public void eraseBoard() 
	{
        //Set Background color to all pixels
        for (int i = 0; i < boardWidth * boardHeight; i++) {
            board[i] = GlobalDefinitions.colorNothing;
        }
    }

    /**
     * Draws the Scale on the oard to make it more readable
     */
    public void drawScale() 
	{
        int colorBords = GlobalDefinitions.colorBords;
        //Draw the horizontal scales
        for (int i = 0; i < boardWidth; i++) {
            //Upper and lower bords
            board[i] = colorBords;
            board[i + (boardHeight - 1) * boardWidth] = colorBords;
            //Middle ligne
            board[i + boardHeight / 2 * boardWidth] = servoColors[16];
            //Quarters
            board[i + boardHeight / 4 * boardWidth] = colorSkyLight1;
            board[i + boardHeight * 3 / 4 * boardWidth] = colorSkyLight1;
            //8ths
            board[i + boardHeight / 8 * boardWidth] = colorSkyLight2;
            board[i + boardHeight * 7 / 8 * boardWidth] = colorSkyLight2;
            board[i + boardHeight * 5 / 8 * boardWidth] = colorSkyLight2;
            board[i + boardHeight * 3 / 8 * boardWidth] = colorSkyLight2;
            //16ths
            board[i + boardHeight / 16 * boardWidth] = colorSkyLight3;
            board[i + boardHeight * 3 / 16 * boardWidth] = colorSkyLight3;
            board[i + boardHeight * 5 / 16 * boardWidth] = colorSkyLight3;
            board[i + boardHeight * 7 / 16 * boardWidth] = colorSkyLight3;
            board[i + boardHeight * 9 / 16 * boardWidth] = colorSkyLight3;
            board[i + boardHeight * 11 / 16 * boardWidth] = colorSkyLight3;
            board[i + boardHeight * 13 / 16 * boardWidth] = colorSkyLight3;
            board[i + boardHeight * 15 / 16 * boardWidth] = colorSkyLight3;
            //The Motor's Limit
            board[i + maxBoardValue * boardWidth] = GlobalDefinitions.colorRed1;
            board[i + minBoardValue * boardWidth] = GlobalDefinitions.colorRed1;
        }

        //Draw the vertical scales
        for (int i = 0; i < boardHeight; i++) {
            //Each 50 Frames draw a vertical scale
            for (int j = 50; j < boardWidth; j += 50) {
                board[j + boardWidth * i] = colorPinkLight;
            }
            //Right and left borders
            board[boardWidth * i] = colorBords;
            board[boardWidth - 1 + boardWidth * i] = colorBords;
        }
    }

    //Draws 11 pixels around the given position on the drawing board
    public void drawKeyFrame(int xPos, int yPos, int drawingColor) 
	{
        int yStretchUp = 2;
        int yStretchDown = 2;
        int xStretchRight = 1;
        int xStretchLeft = 1;

        if (yPos < 2) {
            if (yPos < 1)
                yStretchUp = 0;
            else
                yStretchUp = 1;
        }

        if (yPos > boardHeight - 3) {
            if (yPos > boardHeight - 2)
                yStretchDown = 0;
            else
                yStretchDown = 1;
        }

        if (xPos < 1) xStretchLeft = 0;
        if (xPos > boardWidth - 1) xStretchRight = 0;

        //Top
        board[xPos + ((yPos - yStretchUp) * boardWidth)] = drawingColor;
        //Upper
        board[xPos - xStretchLeft + ((yPos - yStretchUp / 2) * boardWidth)] = drawingColor;
        board[xPos + ((yPos - yStretchUp / 2) * boardWidth)] = drawingColor;
        board[xPos + xStretchRight + ((yPos - yStretchUp / 2) * boardWidth)] = drawingColor;
        //Middle
        board[xPos - xStretchLeft + yPos * boardWidth] = drawingColor;
        board[xPos + yPos * boardWidth] = drawingColor;
        board[xPos + xStretchRight + yPos * boardWidth] = drawingColor;
        //lower
        board[xPos - xStretchLeft + ((yPos + yStretchDown / 2) * boardWidth)] = drawingColor;
        board[xPos + ((yPos + yStretchDown / 2) * boardWidth)] = drawingColor;
        board[xPos + xStretchRight + ((yPos + yStretchDown / 2) * boardWidth)] = drawingColor;
        //Bottom
        board[xPos + ((yPos + yStretchDown) * boardWidth)] = drawingColor;
    }
}
